/***********************************************************************

	File:	art.c
	Rev:	a-1
	Date:	08/09/94

	Copyright (c) 1993, 1994 by The University of Virginia

************************************************************************

	Routines to find Articulation Points.

************************************************************************

	Modification Log:

	a-1:	08/09/94	dmw
		: Created.

************************************************************************/

#include "steiner.h"


/*
 * Global Routines
 */

void			find_articulation_points (bitmap_t *,
						  bitmap_t *,
						  struct cinfo *,
						  bitmap_t *);


/*
 * External References
 */

	/* none */

/*
 * Local Equates
 */

#define	MAX_UFIND		2048


/*
 * Local Routines
 */

static int		find (int);
static void		initialize_union_find (int);
static bool		is_artic (int, bitmap_t *, struct cinfo *);
static void		makeset (int);
static void		unite (int, int);


/*
 * Local Variables
 */


static int		parent [MAX_UFIND];
static int		rank   [MAX_UFIND];
static int		set_size;

/*
 * This routine finds the set of possible articulation points in the given
 * subset of full-sets.
 */

	void
find_articulation_points (

bitmap_t *		tmap,		/* IN - subset of terminals */
bitmap_t *		fset_mask,	/* IN - subset of full-sets */
struct cinfo *		cip,		/* IN - compatibility info */
bitmap_t *		art_mask	/* OUT - set of articulation pts */
)
{
int			i;
bitmap_t *		bp1;

	bp1 = art_mask;
	for (i = 0; i < N_TERM_BMAP; i++) {
		*bp1++ = 0;
	}

	for (i = 0; i < cip -> num_terms; i++) {
		if (BITON (tmap, i) AND is_artic (i, fset_mask, cip)) {
			SETBIT (art_mask, i);
		}
	}
}

/*
 * This routine tests the given terminal to see if it is an articulation
 * point in the sub-problem specified by the full-set mask.  A terminal
 * is an articulation point if the set of all full-sets involving this
 * terminal can be partitioned into TWO OR MORE subsets P1, P2, ... Pn
 * such that every full set in Pi is compatible with every full set in
 * Pj whenever (i <> j).
 */

	static
	bool
is_artic (

int		term,		/* IN - terminal number to test */
bitmap_t *	fset_mask,	/* IN - subset of full-sets to test over */
struct cinfo *	cip		/* IN - compatibility info */
)
{
int		n;
int		i;
int		j;
int		i1;
int		j1;
int		ntrees;
int *		ip0;
int *		ip1;
int *		p;
int *		q;
int8u *		compat;
int		canon [4];

	ntrees = cip -> full_tree_count;
	compat = cip -> carray;

	ip0 = cip -> term_trees [term];
	ip1 = cip -> term_trees [term + 1];

	initialize_union_find (ntrees);

	for (p = ip0; p < ip1; p++) {
		i = *p;
		if (NOT BITON (fset_mask, i)) continue;
		makeset (i);
		for (q = ip0; q < p; q++) {
			j = *q;
			if (NOT BITON (fset_mask, j)) continue;
			if (compat [i*ntrees + j] NE INCOMPATIBLE) continue;
			/* Full sets I and J are incompatible.  This	*/
			/* means they belong in the same Pk -- merge.	*/
			i1 = find (i);
			j1 = find (j);
			if (i1 NE j1) {
				unite (i1, j1);
			}
		}
	}

	/* See if there is more than one set left... */
	n = 0;
	for (p = ip0; p < ip1; p++) {
		if (NOT BITON (fset_mask, *p)) continue;
		i = find (*p);
		j = 0;
		while (TRUE) {
			if (j >= n) {
				if (n >= 4) {
					fatal ("find_artic: Bug 1.");
				}
				canon [n++] = i;
				break;
			}
			if (i EQ canon [j]) break;
			++j;
		}
	}

	if (n >= 2) {
		printf (" %% Terminal %ld is articulation point with %ld components.\n",
			(long) term, (long) n);
	}

	return (n >= 2);
}

/*
 * This routine initializes a collection of N disjoint sets.
 */

	static
	void
initialize_union_find (

int		n		/* IN - number of disjoint sets. */
)
{
int		i;

	set_size = n;

	if (n > MAX_UFIND) {
		/* Asking for too many disjoint sets. */
		fatal ("initialize_union_find: Bug 1.");
	}

	for (i = 0; i < MAX_UFIND; i++) {
		parent [i]	= 0;
		rank [i]	= 0;
	}
}

/*
 * This routine makes a single disjoint set for item "i".
 */

	static
	void
makeset (

int		i		/* IN - item to make into a disjoint set. */
)
{
	if ((i < 0) OR (i >= set_size)) {
		/* Item out of bounds. */
		fatal ("makeset: Bug 1.");
	}
	parent [i]	= i;
	rank [i]	= 0;
}

/*
 * This routine "unites" two sets that were previously disjoint.  I and J
 * must be the "canonical" member of each disjoint set (i.e. they must
 * each be the output of a "find" operation), and must be distinct.
 *
 * We perform the "union by rank" heuristic here.
 */

	static
	void
unite (

int		i,		/* IN - first set to unite. */
int		j		/* IN - second set to unite. */
)
{
int		ri;
int		rj;

	if ((i < 0) OR (i >= set_size)) {
		/* Item I is out of range. */
		fatal ("unite: Bug 1.");
	}
	if ((j < 0) OR (j >= set_size)) {
		/* Item J is out of range. */
		fatal ("unite: Bug 2.");
	}
	if (i EQ j) {
		/* Attempt to unite I with I. */
		fatal ("unite: Bug 3.");
	}

	ri = rank [i];
	rj = rank [j];

	if (ri EQ rj) {
		/* Both subtrees have the same maximum depth.  We	*/
		/* arbitrarily choose I to be underneath J.  The rank	*/
		/* of J must then increase.				*/
		parent [i] = j;
		rank [j]   = rj + 1;
	}
	else if (ri > rj) {
		/* Tree I is (probably) deeper.  Putting J underneath	*/
		/* will not increase I's rank.				*/
		parent [j] = i;
	}
	else {
		/* Tree J is (probably) deeper... */
		parent [i] = j;
	}
}

/*
 * This routine, given a member I of one of the disjoint sets A, will
 * choose a cannonical member J of set A and return it.  Until set A gets
 * united with some other set, find (I) will always return the same J.
 *
 * This routine performs the "path compression" heuristic.
 */

	static
	int
find (

int		i		/* IN - item to find cannonical item for. */
)
{
int		j;
int		k;

	/* Yes, I know this routine is very elegent when coded	*/
	/* recursively...  Here's the iterative version.	*/

	j = parent [i];
	if (i EQ j) {
		/* A cannonical element has itself as parent. */
		return (i);
	}

	/* We must search up the tree -- and compress when done... */
	while (TRUE) {
		k = parent [j];
		if (j EQ k) break;
		j = k;
	}

	/* Now compress the path (make all items in chain point directly */
	/* at the root K) -- we never have to do this search again!	 */
	while (i NE k) {
		j = parent [i];
		parent [i] = k;
		i = j;
	}

	return (k);
}
