/***********************************************************************

	File:	coarse.c
	Rev:	a-1
	Date:	04/18/93

	Copyright (c) 1993, by The University of Virginia

************************************************************************

	Routines to perform Test 2.

************************************************************************

	Modification Log:

	a-1:	04/18/93	dmw
		: Created.  Combined all coarse-grained tests into
		:  a single file.

************************************************************************/

#include "steiner.h"


/*
 * Global Routines
 */

void			coarse_grained_tests (struct pset *,
					      bitmap_t *,
					      int *);
void			compute_mst_longest_edges (struct pset *);
bool			full_topology_edge_too_long (int, struct pset *,
						     struct edge *, int);


/*
 * External References
 */

	/* none */


/*
 * Local Routines
 */

static void	backbone_test (struct pset *, bitmap_t *, int *);
static void	build_adjacency_list (int, int, struct edge *,
				      struct adj **, struct adj *);
static bool	has_empty_rectangle (struct pset *,
				     struct point *,
				     struct point *,
				     bitmap_t *);
static bool	is_empty_rectangle (struct pset *,
				    struct point *,
				    struct point *,
				    struct point *,
				    bitmap_t *);
static bool	is_valid_backbone (struct pset *,
				   struct point *,
				   struct point *);
static void	mst_adjacency_list (struct pset *,
				    struct adj **, struct adj *);
static bool	okay_pair (struct pset *, struct point *, struct point *);
static void	print_bounds (int *, int, char *);
static void	test2 (struct pset *, int *);
static void	test3 (struct pset *, bitmap_t *, int *);
static void	test6 (struct pset *, bitmap_t *, int *);
static void	walk (pnum_t, dist_t, bool *, struct adj **, struct adj *,
		      dist_t *);


/*
 * Local Variables
 */

static dist_t		mst_longest_edges [MAX_TERMINALS] [MAX_TERMINALS];

/*
 * This routine pre-computes a matrix that contains for each pair (A, B)
 * of terminals, the length of the longest edge along the (unique) path
 * from A to B in an arbitrary Minimum Spanning Tree of the given
 * terminal set.  This matrix is used during both the coarse and the
 * fine tests -- therefore, we compute it once and save the results.
 */

	void
compute_mst_longest_edges (

struct pset *		pts	/* IN - the original point set. */
)
{
int			i;
int			n;
dist_t *		rowp;
struct adj *		adj_nodes [MAX_TERMINALS+1];
struct adj		adj_list [2 * MAX_TERMINALS];
bool			mark [MAX_TERMINALS];

	n = pts -> n;

	mst_adjacency_list (pts, &adj_nodes [0], &adj_list [0]);

	(void) memset (mark, FALSE, n);	/* clear out the mark flags. */

	/* Fill in the matrix, one row at a time... */
	for (i = 0; i < n; i++) {
		rowp = &mst_longest_edges [i] [0];
		walk (i, 0, mark, &adj_nodes [0], &adj_list [0], rowp);
	}
}

/*
 * This routine performs all of the coarse-grained tests to eliminate
 * certain backbones and/or subsets of the terminals from consideration
 * as full-sets.
 *
 * This involves computation of a lower bound (for each backbone) on the
 * number of points that can be a member of a full-set with that backbone.
 */

	void
coarse_grained_tests (

struct pset *	pts,		/* IN - the input set of terminals. */
bitmap_t *	tmap,		/* IN - relevent subset of terminals. */
int *		lb		/* IN - lower bounds for each backbone. */
)
{
int		n;

	n = pts -> n;

	test2 (pts, lb);
	print_bounds (lb, n, "Lower bounds after test 2:");

	test6 (pts, tmap, lb);
	print_bounds (lb, n, "Lower bounds after test 6:");

	test3 (pts, tmap, lb);
	print_bounds (lb, n, "Lower bounds after test 3:");

	backbone_test (pts, tmap, lb);
	print_bounds (lb, n, "Lower bounds after backbone_test:");
}

/*
 * This routine computes lower bounds on the sizes of Full Sets for each
 * prospective backbone (i.e. pair of points).  For each pair (A, B) of
 * points, we need to know the length of the longest edge on the path
 * from A to B in the Minimal Spanning Tree of the set.  We then compute
 * the lower bound using Corollary 2 of [1].
 */

	static
	void
test2 (

struct pset *		pts,		/* IN - point set. */
int *			lb		/* OUT - NxN array of lower bounds. */
)
{
int			i;
int			j;
int			n;
struct point *		p1;
struct point *		p2;
dist_t			d;
dist_t			elen;

	n = pts -> n;

	p1 = &(pts -> a [0]);
	for (i = 0; i < n; p1++, i++) {
		p2 = &(pts -> a [0]);
		for (j = 0; j < n; p2++, j++) {
			if (i EQ j) {
				*lb++ = -1;
			}
			else {
				elen = mst_longest_edges [i] [j];
				if (elen EQ 0) {
					*lb++ = 2;
				}
				else {
					d = DISTANCE (p1, p2);
					*lb++ = ((d + elen - 1) / elen) + 1;
				}
			}
		}
	}
}

/*
 * This routine recursively walks through the given Minimum Spanning Tree
 * starting at the given node, and determines the longest edge along the
 * path from the original given root node, to every other node in the tree.
 */

	static
	void
walk (

pnum_t		node,		/* IN - current node in tree. */
dist_t		longest,	/* IN - longest edge in current path. */
bool *		mark,		/* IN - TRUE for nodes on current path. */
struct adj **	adj_nodes,	/* IN - nodes ptrs into adjacency list. */
struct adj *	adj_list,	/* IN - adjacency list.	*/
dist_t *	rowp		/* OUT - array of longest edge lengths. */
)
{
struct adj *		ap;
struct adj *		endp;

	/* We now know the longest edge along the path from the root */
	/* to this current node!  Remember it in the matrix! */
	rowp [node] = longest;

	mark [node] = TRUE;
	ap	= adj_nodes [node];
	endp	= adj_nodes [node + 1];
	while (ap < endp) {
		if (NOT mark [ap -> pnum]) {
			walk (ap -> pnum,
			      max (longest, ap -> len),
			      mark,
			      adj_nodes,
			      adj_list,
			      rowp);
		}
		++ap;
	}
	mark [node] = FALSE;
}

/*
 * This routine computes a Minimum Spanning Tree and converts the result
 * into adjacency list form.
 */

	static
	void
mst_adjacency_list (

struct pset *		pts,		/* IN - point set. */
struct adj **		adj_nodes,	/* OUT - per-node ptrs into list. */
struct adj *		adj_list	/* OUT - adjacency list. */
)
{
int			n;
int			nedges;
struct edge		edges [MAX_TERMINALS - 1];

	n = pts -> n;

	nedges = mst (pts, &edges [0]);

	build_adjacency_list (n, nedges, &edges [0],
			      adj_nodes, adj_list);
}

/*
 * This routine converts a list of edges into a full graph structure
 * represented in adjacency list form.  The adjacency list is in two parts:
 * an array of pointers indexed by node number, and an array of "adj"
 * structures indexed by these pointers.  To find all neighbors of node K,
 * look at every "adj" structure between adj_nodes [K] and adj_nodes [K+1].
 */

	static
	void
build_adjacency_list (

int			n,		/* IN - number of nodes. */
int			nedges,		/* IN - number of edges. */
struct edge *		edges,		/* IN - array of edges. */
struct adj **		adj_nodes,	/* OUT - per-node ptrs into list. */
struct adj *		adj_list	/* OUT - adjacency list. */
)
{
int			i;
struct edge *		ep;
struct adj *		ap;
int			count [MAX_TERMINALS];
struct adj *		tmp_ap [MAX_TERMINALS];

	for (i = 0; i < n; i++) {
		count [i] = 0;
	}

	/* Count neighbors for each node. */
	for (ep = edges, i = 0; i < nedges; ep++, i++) {
		++(count [ep -> p1]);
		++(count [ep -> p2]);
	}

	/* Set up the pointers for each node. */
	ap = adj_list;
	for (i = 0; i < n; i++) {
		adj_nodes [i]	= ap;
		tmp_ap [i]	= ap;
		ap += count [i];
	}
	adj_nodes [i] = ap;		/* set overall end of list. */

	/* Deposit neighbors into properly allocated lists. */
	for (ep = edges, i = 0; i < nedges; ep++, i++) {
		ap = tmp_ap [ep -> p1]++;
		ap -> len	= ep -> len;
		ap -> pnum	= ep -> p2;

		ap = tmp_ap [ep -> p2]++;
		ap -> len	= ep -> len;
		ap -> pnum	= ep -> p1;
	}
}

/*
 * This routine tests the given Full Topology (a graph in edge-list
 * form) to see if the longest edge along the path between any pair of
 * terminals exceeds the equivalent longest edge of the overall Minimum
 * Spanning Tree of all terminals.  If there is such an edge, then this
 * full-topology cannot exist in a Steiner Minimal Tree.
 */

	bool
full_topology_edge_too_long (

int		nterms,		/* IN - number of terminal points. */
struct pset *	comb,		/* IN - terminal/steiner point set. */
struct edge *	edges,		/* IN - edge list. */
int		nedges		/* IN - number of edges. */
)
{
int			i;
int			j;
int			n;
dist_t *		mst_rowp;
dist_t *		smt_rowp;
struct point *		p1;
struct point *		p2;
struct adj *		adj_nodes [MAX_POINTS+1];
struct adj		adj_list [2 * MAX_POINTS];
bool			mark [MAX_POINTS];
dist_t			full_top_longest [MAX_POINTS];

	n = comb -> n;		/* # of terminals/steiners combined. */

	/* Translate into an adjacency list. */
	build_adjacency_list (n,
			      nedges,
			      edges,
			      &adj_nodes [0],
			      &adj_list [0]);

	(void) memset (mark, FALSE, n);	/* clear out the mark flags. */

	/* For each terminal, compute longest edge along paths to */
	/* all other nodes.					  */
	p1 = &(comb -> a [0]);
	for (i = 0; i < nterms; p1++, i++) {
		walk (i,
		      0,
		      mark,
		      &adj_nodes [0],
		      &adj_list [0],
		      &full_top_longest [0]);

		/* Now see if any edges in this row are too long... */
		mst_rowp = &mst_longest_edges [p1 -> other_pnum] [0];
		p2 = &(comb -> a [0]);
		smt_rowp = &full_top_longest [0];
		for (j = 0; j < nterms; p2++, smt_rowp++, j++) {
			if (*smt_rowp > mst_rowp [p2 -> other_pnum]) {
				return (TRUE);
			}
		}
	}

	return (FALSE);
}

/*
 * This routine implements test 6.
 */

	static
	void
test6 (

struct pset *		pts,		/* IN - the point set. */
bitmap_t *		tmap,		/* IN - relevent set of terminals */
int *			lb		/* IN/OUT - backbone lower bounds */
)
{
int			i;
int			j;
int			n;
struct point *		p1;
struct point *		p2;

	n = pts -> n;

	p1 = &(pts -> a [0]);
	for (i = 0; i < n; i++, p1++) {
		if (NOT BITON (tmap, i)) {
			lb += n;
			continue;
		}
		p2 = &(pts -> a [0]);
		for (j = 0; j < n; j++, p2++, lb++) {
			/* We get P1 NE P2 for every call to okay_pair, */
			/* because P1 EQ P2 ==> *lb EQ -1...		*/
			if ((*lb EQ 2) AND
			    (BITON (tmap, j)) AND
			    (NOT okay_pair (pts, p1, p2))) {
				/* This backbone must have at least one */
				/* other point in any fullset! */
				*lb = 3;
			}
		}
	}
}

/*
 * This routine discovers certain backbones that cannot have only 2 points
 * in any full set.
 */

	static
	bool
okay_pair (

struct pset *		pts,		/* IN - the point set. */
struct point *		p1,		/* IN - first backbone point. */
struct point *		p2		/* IN - second backbone point. */
)
{
int		i;
int		n;
struct point *	p3;
dist_t		d12;

	n = pts -> n;

	d12 = DISTANCE (p1, p2);

	p3 = &(pts -> a [0]);
	for (i = 0; i < n; i++, p3++) {
		if (p3 EQ p1) continue;
		if (p3 EQ p2) continue;
		if ((DISTANCE (p3, p1) < d12) AND (DISTANCE (p3, p2) < d12)) {
			return (FALSE);
		}
		if (BETWEEN_X (p1, p3, p2) AND BETWEEN_Y (p1, p3, p2)) {
			return (FALSE);
		}
	}
	return (TRUE);
}

/*
 * This routine implements test 3.  This test is justified by Theorem 4
 * of [1], which states that if {A, B, C} is a full set with respect to
 * V, then the circumrectangle of {A, B, C} must be empty.  If this isn't
 * the case, then any of the prospective backbones AB, AC, and BC that
 * currently have a lower bound of 3 must be raised to 4, because the
 * full set cannot consist of just {A, B, C}.
 */

	static
	void
test3 (

struct pset *		pts,		/* IN - the point set. */
bitmap_t *		tmap,		/* IN - subset of terminals to use */
int *			lb		/* IN/OUT - backbone lower bounds */
)
{
int			i;
int			j;
int			n;
struct point *		p1;
struct point *		p2;

	n = pts -> n;

	p1 = &(pts -> a [0]);
	for (i = 0; i < n; i++, p1++) {
		if (NOT BITON (tmap, i)) {
			lb += n;
			continue;
		}
		p2 = &(pts -> a [0]);
		for (j = 0; j < n; j++, p2++, lb++) {
			if (*lb NE 3) continue;
			if (NOT BITON (tmap, j)) continue;

			/* We have P1 NE P2 here because P1 EQ P2 ==>	*/
			/* *lb EQ -1.					*/

			if (has_empty_rectangle (pts, p1, p2, tmap)) continue;

			/* The circumrectangle of p1 and p2 is not	*/
			/* empty!  Full sets must contain at least 4!	*/
			*lb = 4;
		}
	}
}

/*
 * This routine determines whether or not there is a third terminal p3 such
 * that the circumrectangle of {P1, P2, P3} has no other terminals in it.
 */

	static
	bool
has_empty_rectangle (

struct pset *		pts,		/* IN - the point set. */
struct point *		p1,		/* IN - first backbone point. */
struct point *		p2,		/* IN - second backbone point. */
bitmap_t *		tmap		/* IN - subset of points to use. */
)
{
int		i;
int		n;
struct point *	p3;

	n = pts -> n;

	p3 = &(pts -> a [0]);
	for (i = 0; i < n; i++, p3++) {
		if (p3 EQ p1) continue;
		if (p3 EQ p2) continue;
		if (NOT BITON (tmap, i)) continue;
		if (is_empty_rectangle (pts, p1, p2, p3, tmap)) {
			return (TRUE);
		}
	}
	return (FALSE);
}

/*
 * This routine determines whether or not the (closed) circumrectangle of
 * P1, P2, and P3 contains any other terminals.
 */

	static
	bool
is_empty_rectangle (

struct pset *		pts,		/* IN - the point set. */
struct point *		p1,		/* IN - first point. */
struct point *		p2,		/* IN - second point. */
struct point *		p3,		/* IN - third point. */
bitmap_t *		tmap		/* IN - subset of points to use. */
)
{
int		i;
int		n;
struct point *	p4;
coord_t		max_x;
coord_t		max_y;
coord_t		min_x;
coord_t		min_y;

#define	MIN_IN(var, val)	if (val < var) var = val
#define	MAX_IN(var, val)	if (val > var) var = val

	max_x = p1 -> x;
	MAX_IN (max_x, p2 -> x);
	MAX_IN (max_x, p3 -> x);

	max_y = p1 -> y;
	MAX_IN (max_y, p2 -> y);
	MAX_IN (max_y, p3 -> y);

	min_x = p1 -> x;
	MIN_IN (min_x, p2 -> x);
	MIN_IN (min_x, p3 -> x);

	min_y = p1 -> y;
	MIN_IN (min_y, p2 -> y);
	MIN_IN (min_y, p3 -> y);

	n = pts -> n;

	p4 = &(pts -> a [0]);
	for (i = 0; i < n; i++, p4++) {
		if (p4 -> x < min_x) continue;
		if (p4 -> x > max_x) continue;
		if (p4 -> y < min_y) continue;
		if (p4 -> y > max_y) continue;
		if (p4 EQ p1) continue;
		if (p4 EQ p2) continue;
		if (p4 EQ p3) continue;
		if (NOT BITON (tmap, i)) continue;

		/* Point P4 is somewhere inside the rectangle. */
		return (FALSE);
	}
	return (TRUE);
}

/*
 * This routine kills off any backbone that has a terminal that is too
 * close to where the corner is.  We find the pair of *potential* Steiner
 * points that are closest to the corner (one on each leg).  If there is
 * a terminal that is closer to each of these than they are to each other,
 * then we kill the backbone by Theorem 8 of [1].
 */

	static
	void
backbone_test (

struct pset *		inpts,		/* IN - the point set. */
bitmap_t *		tmap,		/* IN - subset of points to use */
int *			lb		/* IN/OUT - backbone lower bounds */
)
{
int			i;
int			j;
int			n;
struct point *		p1;
struct point *		p2;
struct pset		pts;

	/* Get a copy of just the relevent points... */
	copy_terminal_subset (&pts, inpts, tmap);

	n = inpts -> n;

	p1 = &(inpts -> a [0]);
	for (i = 0; i < n; p1++, i++) {
		if (NOT BITON (tmap, i)) {
			lb += n;
			continue;
		}
		p2 = &(inpts -> a [0]);
		for (j = 0; j < n; lb++, p2++, j++) {
			if (*lb < 0) continue;
			if (NOT BITON (tmap, j)) continue;
			if (NOT is_valid_backbone (&pts, p1, p2)) {
				*lb = -1;
			}
		}
	}
}

/*
 * This routine applies Theorem 8 of [1] to the backbone from point A to
 * point B.  (The vertical line is incident to point A.)  We find the
 * pair of *potential* Steiner points that are closest to the corner on
 * each leg of the backbone.  If there is a terminal that is closer to
 * each of these Steiner points than they are to each other, then by
 * Theorem 8, this backbone cannot form a Steiner Minimal Tree.
 */

	static
	bool
is_valid_backbone (

struct pset *		pts,		/* IN - the point set. */
struct point *		ap,		/* IN - point A of the backbone. */
struct point *		bp		/* IN - point B of the backbone. */
)
{
int			i;
int			n;
struct point *		p1;
struct point *		first_off_horizontal;
struct point *		first_off_vertical;
struct point		s1;
struct point		s2;

	n = pts -> n;

	first_off_horizontal	= bp;
	first_off_vertical	= ap;

	p1 = &(pts -> a [0]);
	for (i = 0; i < n; p1++, i++) {
		if (BETWEEN_Y (ap, bp, p1) AND
		    BETWEEN_X (ap, p1, first_off_horizontal)) {
			first_off_horizontal = p1;
		}
		if (BETWEEN_X (bp, ap, p1) AND
		    BETWEEN_Y (bp, p1, first_off_vertical)) {
			first_off_vertical = p1;
		}
	}

	s1.x	= first_off_horizontal -> x;
	s1.y	= bp -> y;

	s2.x	= ap -> x;
	s2.y	= first_off_vertical -> y;

	if (is_point_close_to (pts, &s1, &s2)) {
		/* This backbone cannot be valid! */
		return (FALSE);
	}

	/* Backbone is valid. */
	return (TRUE);
}

/*
 * This routine prints out an upper/lower bound array.
 */

	static
	void
print_bounds (

int *		a,		/* IN - array. */
int		n,		/* IN - dimension. */
char *		msg		/* IN - message. */
)
{
int		i, j;

	if (Print_Bounds) {
		(void) printf ("\n%s\n", msg);
		for (i = 0; i < n; i++) {
			for (j = 0; j < n; j++) {
				(void) printf (" %2d", *a++);
			}
			(void) printf ("\n");
		}
	}
}
