/***********************************************************************

	File:	compat.c
	Rev:	a-1
	Date:	03/20/93

	Copyright (c) 1993, by The University of Virginia

************************************************************************

	This file contains routines for checking the relative
	compatibility of pairs of full-sets.

************************************************************************

	Modification Log:

	a-1:	03/20/93	dmw
		: Created.

************************************************************************/

#include "steiner.h"


/*
 * Global Routines
 */

void			build_compatibility_data (struct pset *,
						  bitmap_t *,
						  struct full_set *,
						  bitmap_t *,
						  struct cinfo *);
void			free_compatibility_data (struct cinfo *);
void			init_tables (void);
void			initial_pruning (bitmap_t *,
					 bitmap_t *,
					 struct cinfo *);
struct full_set **	put_trees_in_array (struct full_set *, int *);


/*
 * External References
 */

	/* none */

/*
 * Local Types
 */

enum direction {
	NORTH,
	SOUTH,
	EAST,
	WEST
};


/*
 * Local Routines
 */

static enum compat		basic_compatibility (struct full_set *,
						     struct full_set *);
static int8u *			build_compatibility_matrix (bitmap_t *,
							    struct cinfo *);
static dist_t			closest_distance (struct point *,
						  struct pset *,
						  struct pset *);
static char *			compat_name (enum compat);
static bool			compat_test_1 (bitmap_t *,
					       bitmap_t *,
					       struct cinfo *);
static bool			compat_test_2 (bitmap_t *,
					       bitmap_t *,
					       struct cinfo *);
static bool			compat_test_3 (bitmap_t *,
					       bitmap_t *,
					       struct cinfo *);
static bool			compat_test_4 (bitmap_t *,
					       bitmap_t *,
					       bitmap_t *,
					       struct cinfo *);
static bool			compat_test_5 (bitmap_t *,
					       bitmap_t *,
					       bitmap_t *,
					       struct cinfo *);
static int **			decompose_tree_list (int, struct full_set *);
static bool			directions_incompatible (struct full_set *,
							 struct full_set *);
static dist_t			distance_from_point_to_line (struct point *,
							     struct point *,
							     struct point *);
static pnum_t			find_common_point_number (struct full_set *,
							  struct full_set *);
static int			find_first_fake_term (struct pset *);
static int			first_bit_set (bitmap_t);
static bool			have_intersections (struct full_set *,
						    struct full_set *);
static void			init_compat_bitmasks (struct cinfo *);
static void			init_nbits (void);
static void			init_one_bits_in_byte_table (void);
static void			init_term_tree_masks (struct cinfo *);
static bool			lines_intersect (struct point *,
						 struct point *,
						 struct point *,
						 struct point *);
static bool			okay_test_1 (int,
					     int,
					     bitmap_t *,
					     bitmap_t *,
					     struct cinfo *);
static bool			okay_test_2 (int,
					     int,
					     bitmap_t *,
					     bitmap_t *,
					     struct cinfo *);
static bool			okay_test_3 (int,
					     bitmap_t *,
					     bitmap_t *,
					     struct cinfo *);
static bool			okay_test_4 (int,
					     bitmap_t *,
					     bitmap_t *,
					     bitmap_t *,
					     struct cinfo *);
static bool			okay_test_5 (int,
					     int,
					     bitmap_t *,
					     bitmap_t *,
					     bitmap_t *,
					     struct cinfo *);
static void			print_compatibility_matrix (struct cinfo *);
static enum direction		segment_direction (struct full_set *, pnum_t);
static dist_t			segment_distance (struct full_set *,
						  pnum_t,
						  struct pset *,
						  int *);
static void			set_bit_maps (struct full_set *);
static bool			still_viable (int,
					      bitmap_t *,
					      bitmap_t *,
					      struct cinfo *);
static bool			too_close (struct full_set *,
					   struct full_set *);
static void			ip_walk (int,
					 bitmap_t *,
					 bitmap_t *,
					 struct full_set **,
					 int **);

/*
 * This routine initializes various global tables.
 */

	void
init_tables (void)

{
	init_nbits ();
	init_one_bits_in_byte_table ();
}

/*
 * This routine initializes a translation table that converts an 8-bit
 * byte into the count of ONE bits in the byte.
 */

	static
	void
init_nbits (void)

{
int		i;
int32u		bits;
int		count;

	for (i = 0; i < 256; i++) {
		count = 0;
		for (bits = ((int32u) i); bits > 0; bits >>= 1) {
			if ((bits & 0x01) NE 0) {
				++count;
			}
		}
		nbits [i] = ((int8u) count);
	}
}

/*
 * This routine initializes a table that is used to rapidly convert an
 * 8-bit byte into a sequence of bit-positions that are on within that
 * byte.  For example, it translates the byte 0xB7 into {0, 1, 2, 4, 5, 7}.
 */

	static
	void
init_one_bits_in_byte_table (void)

{
int		i;
int		j;
int		byte;
int8u *		p;

	p = &one_bits_in_byte_array [0];
	for (i = 0; i < 256; i++) {
		one_bits_in_byte [i] = p;
		byte = i;
		j = 0;
		while (byte > 0) {
			if ((byte & 0x01) NE 0) {
				*p++ = j;
			}
			++j;
			byte >>= 1;
		}
	}

	/* Terminate last table entry. */
	one_bits_in_byte [i] = p;
}

/*
 * This routine sets up several important compatibility data-bases.
 */

	void
build_compatibility_data (

struct pset *		pts,		/* IN - set of terminals */
bitmap_t *		tmap,		/* IN - subset of terminals to use */
struct full_set *	fsp,		/* IN - list of full-sets */
bitmap_t *		fset_mask,	/* IN - subset of full-sets to use */
struct cinfo *		cip		/* OUT - compatibility data-bases */
)
{
int			i;
int			n;
struct point *		p1;
struct full_set *	tmp;
int			nft;
struct full_set **	full_trees;

	/* Make certain each full-sets terminal bit-map is set. */
	set_bit_maps (fsp);

#if 0
 for (tmp = fsp; tmp NE NULL; tmp = tmp -> next) {
	(void) printf (" %% full set %d:", tmp -> tree_num);
	n = tmp -> terminals -> n;
	p1 = &(tmp -> terminals -> a [0]);
	for (i = 0; i < n; p1++, i++) {
		(void) printf (" %d", p1 -> pnum);
	}
	(void) printf ("\n");
 }
#endif

	full_trees = put_trees_in_array (fsp, &nft);

	cip -> full_tree_count		= nft;
	cip -> full_trees		= full_trees;
	cip -> num_terms		= pts -> n;
	cip -> num_term_masks		= BMAP_ELTS (pts -> n);
	cip -> num_fset_masks		= BMAP_ELTS (nft);
	cip -> first_fake_term		= find_first_fake_term (pts);
	cip -> pts			= pts;
	cip -> carray			= NULL;
	cip -> cmasks			= NULL;
	cip -> incmasks			= NULL;
	cip -> term_trees		= NULL;
	cip -> term_tree_masks		= NULL;
	cip -> sinfo_ptr		= NULL;

	cip -> carray = build_compatibility_matrix (fset_mask, cip);
	print_compatibility_matrix (cip);

	cip -> term_trees = decompose_tree_list (pts -> n, fsp);

	init_compat_bitmasks (cip);
	init_term_tree_masks (cip);
}

/*
 * This routine frees all of the compatibility data-bases.
 */

	void
free_compatibility_data (

struct cinfo *		cip		/* IN - compat info. */
)
{
	free ((char *) cip -> full_trees);
	free ((char *) cip -> carray);
	free ((char *) cip -> term_trees [0]);
	free ((char *) cip -> term_trees);
	free ((char *) cip -> cmasks);
	free ((char *) cip -> incmasks);
	free ((char *) cip -> term_tree_masks);

	cip -> full_trees	= NULL;
	cip -> full_tree_count	= 0;
	cip -> carray		= NULL;
	cip -> term_trees	= NULL;
	cip -> cmasks		= NULL;
	cip -> incmasks		= NULL;
	cip -> term_tree_masks	= NULL;
}

/*
 * This routine iterates through each of the given full-sets, and initializes
 * its Terminal bit-map to be the set of terminals in the full-set.
 */

	static
	void
set_bit_maps (

struct full_set *	fsp		/* IN - the list of full-sets. */
)
{
int			i;
int			n;
bitmap_t *		bp;
struct point *		p1;

	while (fsp NE NULL) {
		bp = &(fsp -> tmap [0]);
		for (i = 0; i < N_TERM_BMAP; bp++, i++) {
			*bp = 0;
		}

		bp = &(fsp -> tmap [0]);

		n = fsp -> terminals -> n;
		p1 = &(fsp -> terminals -> a [0]);
		for (i = 0; i < n; p1++, i++) {
			SETBIT (bp, p1 -> pnum);
		}
		fsp = fsp -> next;
	}
}

/*
 * This routine allocates an array of pointers that lets us access a particular
 * tree number directly.
 */

	struct full_set **
put_trees_in_array (

struct full_set *	fsp,		/* IN - list of full-trees. */
int *			acount		/* OUT - count of array elements. */
)
{
struct full_set **	ap;
struct full_set *	p;
int			count;
int			num;
struct full_set **	array;

	count = 0;
	for (p = fsp; p NE NULL; p = p -> next) {
		++count;
	}

	array = NEWA (count, struct full_set *);

	num = 0;
	ap = array;
	for (p = fsp; p NE NULL; p = p -> next) {
		p -> tree_num		= num++;
		*ap++ = p;
	}

	*acount = count;

	return (array);
}

/*
 * This routine determines the index of the first "fake" terminal
 * in the given set of terminals.
 */

	static
	int
find_first_fake_term (

struct pset *		pts	/* IN - set of terminals */
)
{
int		i;
int		n;
struct point *	p1;

	/* The "fake" terminals are all at the end.  Scan back	*/
	/* until we find the first non-fake one...		*/

	n = pts -> n;
	p1 = &(pts -> a [n - 1]);
	for (i = n; i > 0; p1--, i--) {
		if (p1 -> other_pnum >= 0) return (i);
	}

	return (0);
}

/*
 * This routine builds the N by N matrix that indicates the relative
 * compatibility of full-sets I and J.  (Here N is the number of full-sets.)
 */

	static
	int8u *
build_compatibility_matrix (

bitmap_t *		fset_mask,	/* IN - subset of full-sets to use */
struct cinfo *		cip		/* IN - compat info. */
)
{
int			i;
int			j;
int			nft;
struct full_set *	ip1;
struct full_set *	jp1;
enum compat		cval;
int8u *			carray;
struct pset		termi2;
struct pset		steini2;
struct full_set		fsi2;
struct pset		termj2;
struct pset		steinj2;
struct full_set		fsj2;

	nft = cip -> full_tree_count;

	carray = NEWA (nft * nft, int8u);

	fsi2.terminals	= &termi2;
	fsi2.steiners	= &steini2;

	fsj2.terminals	= &termj2;
	fsj2.steiners	= &steinj2;

	for (i = 0; i < nft; i++) {
		if (NOT BITON (fset_mask, i)) {
			/* Full-set not being considered -- make it	*/
			/* incompatible with everything else.		*/
			for (j = 0; j < nft; j++) {
				carray [i*nft + j] = (int8u) INCOMPATIBLE;
				carray [j*nft + i] = (int8u) INCOMPATIBLE;
			}
			continue;
		}

		ip1 = cip -> full_trees [i];

		carray [i*nft + i] = INCOMPATIBLE;

		if (ip1 -> style NE FAKE) {
			make_second_topology (ip1, &fsi2);
		}

		for (j = i+1; j < nft; j++) {
			if (NOT BITON (fset_mask, j)) continue;
			jp1 = cip -> full_trees [j];

			cval = basic_compatibility (ip1, jp1);
			carray [i*nft + j] = ((int8u) cval);
			carray [j*nft + i] = ((int8u) cval);

			if (cval NE COMPATIBLE) continue;

			/* A "fake" full-set is compatible or not	*/
			/* strictly on the basis of # terminals shared.	*/
			if (ip1 -> style EQ FAKE) continue;
			if (jp1 -> style EQ FAKE) continue;

			make_second_topology (jp1, &fsj2);

			if (directions_incompatible (ip1, jp1) OR
			    directions_incompatible (ip1, &fsj2) OR
			    directions_incompatible (&fsi2, jp1) OR
			    directions_incompatible (&fsi2, &fsj2) OR
			    have_intersections (ip1, jp1) OR
			    have_intersections (ip1, &fsj2) OR
			    have_intersections (&fsi2, jp1) OR
			    have_intersections (&fsi2, &fsj2) OR
			    too_close (ip1, jp1) OR
			    too_close (ip1, &fsj2) OR
			    too_close (&fsi2, jp1) OR
			    too_close (&fsi2, &fsj2)) {
				carray [i*nft + j] = INCOMPATIBLE;
				carray [j*nft + i] = INCOMPATIBLE;
			}
		}
	}

	return (carray);
}

/*
 * This routine determines the basic relative compatibility of two
 * full-sets.  This is determined solely by the number of points they
 * have in common.
 */

	static
	enum compat
basic_compatibility (

struct full_set *	fsp1,		/* IN - first full-set. */
struct full_set *	fsp2		/* IN - second full-set. */
)
{
int		i;
bitmap_t *	bp1;
bitmap_t *	bp2;
bitmap_t	mask;
int		common_count;

	bp1 = &(fsp1 -> tmap [0]);
	bp2 = &(fsp2 -> tmap [0]);
	common_count = 0;

	for (i = 0; i < N_TERM_BMAP; bp1++, bp2++, i++) {
		mask = *bp1 & *bp2;
		common_count += NBITSON (mask);
	}

	switch (common_count) {
	case 0:			return (DISJOINT);
	case 1:			return (COMPATIBLE);
	default:		return (INCOMPATIBLE);
	}
}

/*
 * this routine takes two full-topologies that are known to share a single
 * common terminal.  If the two trees connect to this terminal from the
 * same direction, then they are definitely incompatible!
 */

	static
	bool
directions_incompatible (

struct full_set *	fsp1,		/* IN - first full-topology. */
struct full_set *	fsp2		/* IN - second full-topology. */
)
{
pnum_t			pnum;
enum direction		dir1;
enum direction		dir2;

	pnum = find_common_point_number (fsp1, fsp2);

	dir1 = segment_direction (fsp1, pnum);
	dir2 = segment_direction (fsp2, pnum);

	return (dir1 EQ dir2);
}

/*
 * This routine scans the given bit-mask to find the bit number of the
 * least-significant bit that is ON.
 */

	static
	int
first_bit_set (

bitmap_t	mask		/* IN - mask to search. */
)
{
int		n;

	if (mask EQ 0) return (-1);

#if (BPW > 32)
	#error "This code assumes at most 32 bits per word!"
#endif

	n = 0;
#if (BPW > 16)
	if ((mask & 0x0000FFFF) EQ 0) {
		mask >>= 16;
		n += 16;
	}
#endif
#if (BPW > 8)
	if ((mask & 0x000000FF) EQ 0) {
		mask >>= 8;
		n += 8;
	}
#endif
#if (BPW > 4)
	if ((mask & 0x0000000F) EQ 0) {
		mask >>= 4;
		n += 4;
	}
#endif
#if (BPW > 2)
	if ((mask & 0x00000003) EQ 0) {
		mask >>= 2;
		n += 2;
	}
#endif
#if (BPW > 1)
	if ((mask & 0x00000001) EQ 0) {
		++n;
	}
#endif

	return (n);
}

/*
 * This routine determines the direction by which the line segment leaves
 * point number "pnum" within the given full-topology.
 */

	static
	enum direction
segment_direction (

struct full_set *	fsp,		/* IN - the full-topology. */
pnum_t			pnum		/* IN - the point number. */
)
{
int			i;
int			n;
struct point *		p1;
struct point *		p2;
struct pset *		terms;
struct pset *		steins;
enum direction		dir;

	terms	= fsp -> terminals;
	steins	= fsp -> steiners;

	n = terms -> n;

	p1 = &(terms -> a [0]);
	i = 0;
	while (TRUE) {
		if (i >= n) {
			fatal ("segment_direction: Bug 1.");
		}
		if (p1 -> pnum EQ pnum) break;
		++p1;
		++i;
	}

	switch (i) {
	case 0:
		p2 = &(terms -> a [1]);
		if (p2 -> y > p1 -> y) {
			dir = NORTH;
		}
		else if (p2 -> y < p1 -> y) {
			dir = SOUTH;
		}
		else if (p2 -> x > p1 -> x) {
			dir = EAST;
		}
		else {
			dir = WEST;
		}
		break;

	case 1:
		p2 = &(terms -> a [0]);
		if (p2 -> x > p1 -> x) {
			dir = EAST;
		}
		else if (p2 -> x < p1 -> x) {
			dir = WEST;
		}
		else if (p2 -> y > p1 -> y) {
			dir = NORTH;
		}
		else {
			dir = SOUTH;
		}
		break;

	default:
		p2 = &(steins -> a [i-2]);
		if (p2 -> x > p1 -> x) {
			dir = EAST;
		}
		else if (p2 -> x < p1 -> x) {
			dir = WEST;
		}
		else if (p2 -> y > p1 -> y) {
			dir = NORTH;
		}
		else {
			dir = SOUTH;
		}
		break;
	}

	return (dir);
}

/*
 * this routine takes two full-topologies that are known to share a single
 * common terminal.  If any of the individual line-segments of the two
 * trees intersect, then they are definitely incompatible!
 */

	static
	bool
have_intersections (

struct full_set *	fsp1,		/* IN - first full-topology. */
struct full_set *	fsp2		/* IN - second full-topology. */
)
{
int			i;
int			j;
int			n1;
int			n2;
struct point *		ap1;
struct point *		bp1;
struct point *		ap2;
struct point *		bp2;
struct point *		tp1;
struct point *		sp1;
struct point *		tp2;
struct point *		sp2;
struct pset *		terms1;
struct pset *		terms2;
struct pset *		steins1;
struct pset *		steins2;
struct point		c1;		/* corner of tree 1. */
struct point		c2;		/* corner of tree 2. */

	terms1	= fsp1 -> terminals;
	steins1	= fsp1 -> steiners;
	terms2	= fsp2 -> terminals;
	steins2	= fsp2 -> steiners;

	ap1 = &(terms1 -> a [0]);
	bp1 = &(terms1 -> a [1]);
	ap2 = &(terms2 -> a [0]);
	bp2 = &(terms2 -> a [1]);

	c1.x	= ap1 -> x;
	c1.y	= bp1 -> y;
	c2.x	= ap2 -> x;
	c2.y	= bp2 -> y;

	/* See if any of the backbone segments intersect... */
	if (lines_intersect (ap1, &c1, ap2, &c2) OR
	    lines_intersect (ap1, &c1, &c2, bp2) OR
	    lines_intersect (&c1, bp1, ap2, &c2) OR
	    lines_intersect (&c1, bp1, &c2, bp2)) {
		return (TRUE);
	}

	n1 = terms1 -> n;
	n2 = terms2 -> n;

	tp1 = &(terms1 -> a [2]);
	sp1 = &(steins1 -> a [0]);
	for (i = 2; i < n1; tp1++, sp1++, i++) {
		if (lines_intersect (tp1, sp1, ap2, &c2) OR
		    lines_intersect (tp1, sp1, &c2, bp2)) {
			return (TRUE);
		}
	}

	tp2 = &(terms2 -> a [2]);
	sp2 = &(steins2 -> a [0]);
	for (j = 2; j < n2; tp2++, sp2++, j++) {
		if (lines_intersect (ap1, &c1, tp2, sp2) OR
		    lines_intersect (&c1, bp1, tp2, sp2)) {
			return (TRUE);
		}
	}

	tp1 = &(terms1 -> a [2]);
	sp1 = &(steins1 -> a [0]);
	for (i = 2; i < n1; tp1++, sp1++, i++) {
		tp2 = &(terms2 -> a [2]);
		sp2 = &(steins2 -> a [0]);
		for (j = 2; j < n2; tp2++, sp2++, j++) {
			if (lines_intersect (tp1, sp1, tp2, sp2)) {
				return (TRUE);
			}
		}
	}

	return (FALSE);
}

/*
 * This routine returns true if the given pair of line segments (each denoted
 * by a pair of end-points) intersect anywhere OTHER than at their end-
 * points.
 */

	static
	bool
lines_intersect (

struct point *		p1,		/* IN - point 1. */
struct point *		p2,		/* IN - point 2. */
struct point *		p3,		/* IN - point 3. */
struct point *		p4		/* IN - point 4. */
)
{
bool		line1_vert;
bool		line2_vert;
bool		result;

#define	XMAX(p1, p2)	((coord_t) max (p1 -> x, p2 -> x))
#define	XMIN(p1, p2)	((coord_t) min (p1 -> x, p2 -> x))
#define	YMAX(p1, p2)	((coord_t) max (p1 -> y, p2 -> y))
#define	YMIN(p1, p2)	((coord_t) min (p1 -> y, p2 -> y))

	if ((p1 -> x EQ p2 -> x) AND (p1 -> y EQ p2 -> y)) {
		/* first line is a single point! */
		return (FALSE);
	}
	if ((p3 -> x EQ p4 -> x) AND (p3 -> y EQ p4 -> y)) {
		/* second line is a single point! */
		return (FALSE);
	}

	line1_vert = (p1 -> x EQ p2 -> x);
	line2_vert = (p3 -> x EQ p4 -> x);

	if (line1_vert) {
		if (line2_vert) {
			if (p1 -> x NE p3 -> x) {
				result = FALSE;
			}
			else if (YMIN (p1, p2) <= YMIN (p3, p4)) {
				result = (YMAX (p1, p2) > YMIN (p3, p4));
			}
			else {
				result = (YMAX (p3, p4) > YMIN (p1, p2));
			}
		}
		else {
			result = (XMIN (p3, p4) < p1 -> x) AND
				 (p1 -> x < XMAX (p3, p4)) AND
				 (YMIN (p1, p2) < p3 -> y) AND
				 (p3 -> y < YMAX (p1, p2));
		}
	}
	else {
		if (line2_vert) {
			result = (XMIN (p1, p2) < p3 -> x) AND
				 (p3 -> x < XMAX (p1, p2)) AND
				 (YMIN (p3, p4) < p1 -> y) AND
				 (p1 -> y < YMAX (p3, p4));
		}
		else {
			if (p1 -> y NE p3 -> y) {
				result = FALSE;
			}
			else if (XMIN (p1, p2) <= XMIN (p3, p4)) {
				result = (XMAX (p1, p2) > XMIN (p3, p4));
			}
			else {
				result = (XMAX (p3, p4) > XMIN (p1, p2));
			}
		}
	}

	return (result);

#undef	XMAX
#undef	XMIN
#undef	YMAX
#undef	YMIN
}

/*
 * This routine takes two full-topologies that are known to share a single
 * common terminal.  If it turns out that there is a way to hook the two
 * trees together (other than at the common point) that results in a shorter
 * tree, then these full-sets are definitely incompatible!
 */

	static
	bool
too_close (

struct full_set *	fsp1,		/* IN - first full-topology. */
struct full_set *	fsp2		/* IN - second full-topolgy. */
)
{
int			i;
int			n;
struct point *		p1;
dist_t			bound1;
dist_t			bound2;
int			pos1;
int			pos2;
pnum_t			pnum;
struct pset *		terms1;
struct pset *		terms2;
struct pset *		steins1;
struct pset *		steins2;
struct pset		new_terms1;
struct pset		new_terms2;

	pnum = find_common_point_number (fsp1, fsp2);

	bound1 = segment_distance (fsp1, pnum, &new_terms1, &pos1);
	bound2 = segment_distance (fsp2, pnum, &new_terms2, &pos2);

	terms1	= fsp1 -> terminals;
	steins1	= fsp1 -> steiners;
	terms2	= fsp2 -> terminals;
	steins2	= fsp2 -> steiners;

	n = terms1 -> n;
	p1 = &(terms1 -> a [0]);
	for (i = 0; i < n; p1++, i++) {
		if (i EQ pos1) continue;
		if (closest_distance (p1, &new_terms2, steins2) < bound2) {
			return (TRUE);
		}
	}

	n = terms2 -> n;
	p1 = &(terms2 -> a [0]);
	for (i = 0; i < n; p1++, i++) {
		if (i EQ pos2) continue;
		if (closest_distance (p1, &new_terms1, steins1) < bound1) {
			return (TRUE);
		}
	}

	return (FALSE);
}

/*
 * This routine finds the terminal whose original point-number is "pnum"
 * within the given full-topology.  Its index within the particular full-set
 * is returned via "pos".  A new terminal set is then constructed in
 * "new_terms", and this terminal is replaced by the Steiner point that is
 * closest to it along the backbone.  The distance to this point is returned.
 */

	static
	dist_t
segment_distance (

struct full_set *	fsp,		/* IN - the full-topology. */
pnum_t			pnum,		/* IN - the particular point number. */
struct pset *		new_terms,	/* OUT - modified terminal set. */
int *			pos		/* OUT - index of pnum in terminals. */
)
{
int			i;
int			j;
int			n;
struct point *		p1;
struct point *		sp;
struct point *		ap;
struct point *		bp;
struct pset *		terms;
struct pset *		steins;
dist_t			bound;
dist_t			d;

	terms	= fsp -> terminals;
	steins	= fsp -> steiners;

	/* Make a full copy of the terminals... */
	COPY_PSET (new_terms, terms);

	n = terms -> n;

	p1 = &(terms -> a [0]);
	i = 0;
	while (TRUE) {
		if (i >= n) {
			fatal ("segment_distance: Bug 1.");
		}
		if (p1 -> pnum EQ pnum) break;
		++p1;
		++i;
	}

	*pos = i;

	ap = &(terms -> a [0]);
	bp = &(terms -> a [1]);

	switch (i) {
	case 0:
		bound = DISTANCE (ap, bp);
		new_terms -> a [0] = *bp;
		sp = &(steins -> a [0]);
		for (j = 2; j < n; sp++, j++) {
			d = DISTANCE (ap, sp);
			if (d < bound) {
				bound = d;
				new_terms -> a [0] = *sp;
			}
		}
		break;

	case 1:
		bound = DISTANCE (ap, bp);
		new_terms -> a [1] = *ap;
		sp = &(steins -> a [0]);
		for (j = 2; j < n; sp++, j++) {
			d = DISTANCE (bp, sp);
			if (d < bound) {
				bound = d;
				new_terms -> a [1] = *sp;
			}
		}
		break;

	default:
		sp = &(steins -> a [i-2]);
		bound = DISTANCE (p1, sp);
		new_terms -> a [i] = *sp;
		break;
	}

	return (bound);
}

/*
 * This routine finds the shortest distance between the given point and
 * any of the line segments in the full-topology described by the given
 * sets of terminals and Steiner points.
 */

	static
	dist_t
closest_distance (

struct point *		p1,		/* IN - the point to measure from. */
struct pset *		terms,		/* IN - the set of terminals. */
struct pset *		steins		/* IN - the set of Steiner points. */
)
{
int			i;
int			n;
struct point *		tp;
struct point *		sp;
dist_t			d;
dist_t			best;
struct point *		ap;
struct point *		bp;
struct point		c;		/* the backbone corner point. */

	n = terms -> n;

	ap = &(terms -> a [0]);
	bp = &(terms -> a [1]);

	c.x	= ap -> x;
	c.y	= bp -> y;

	best = distance_from_point_to_line (p1, ap, &c);

	d = distance_from_point_to_line (p1, &c, bp);
	if (d < best) {
		best = d;
	}

	tp = &(terms -> a [2]);
	sp = &(steins -> a [0]);
	for (i = 2; i < n; tp++, sp++, i++) {
		d = distance_from_point_to_line (p1, tp, sp);
		if (d < best) {
			best = d;
		}
	}

	return (best);
}

/*
 * This routine computes the shortest distance from point P1 to the line
 * segment defined by P2 and P3.
 */

	static
	dist_t
distance_from_point_to_line (

struct point *		p1,		/* IN - the point. */
struct point *		p2,		/* IN - first line segment endpt. */
struct point *		p3		/* IN - second line segment endpt. */
)
{
dist_t		d;
dist_t		best;

	best = DISTANCE (p1, p2);
	d    = DISTANCE (p1, p3);

	if (d < best) {
		best = d;
	}
	if (p2 -> x EQ p3 -> x) {
		/* Line is vertical. */
		if (STRICTLY_BETWEEN_Y (p2, p1, p3)) {
			d = DIFF (p1 -> x, p2 -> x);
			if (d < best) {
				best = d;
			}
		}
	}
	else if (p2 -> y EQ p3 -> y) {
		/* Line is horizontal. */
		if (STRICTLY_BETWEEN_X (p2, p1, p3)) {
			d = DIFF (p1 -> y, p2 -> y);
			if (d < best) {
				best = d;
			}
		}
	}
	else {
		fatal ("distance_from_point_to_line: Bug 1.");
	}

	return (best);
}

/*
 * This routine determines the point-number of the terminal that the two
 * given full-topologies have in common.
 */

	static
	pnum_t
find_common_point_number (

struct full_set *	fsp1,		/* IN - first full-topology. */
struct full_set *	fsp2		/* IN - second full-topology. */
)
{
int			i;
bitmap_t *		bp1;
bitmap_t *		bp2;
bitmap_t		mask;
pnum_t			pnum;

	/* Find the point-number of the point that is shared... */
	pnum = 0;
	bp1 = &(fsp1 -> tmap [0]);
	bp2 = &(fsp2 -> tmap [0]);
	for (i = 0; i < N_TERM_BMAP; i++) {
		mask = (*bp1++ & *bp2++);
		if (mask NE 0) {
			pnum += first_bit_set (mask);
			return (pnum);
		}
		pnum += BPW;
	}

	fatal ("find_common_point_number: Bug 1.");

	return (0);
}

/*
 * This routine creates an array that is indexed by point number.  It
 * gives a list of all tree-numbers that involve that point.
 */

	static
	int **
decompose_tree_list (

int			nterms,		/* IN - total number of terminals. */
struct full_set *	fsp		/* IN - list of all full-sets. */
)
{
int			i;
int			n;
struct full_set *	fp;
struct pset *		terms;
struct point *		p1;
int			total;
int *			ip;
int **			buckets;
int			counts [MAX_TERMINALS];
int *			ptrs [MAX_TERMINALS];

	total = 0;
	for (i = 0; i < nterms; i++) {
		counts [i] = 0;
	}

	for (fp = fsp; fp NE NULL; fp = fp -> next) {
		terms = fp -> terminals;
		n = terms -> n;
		p1 = &(terms -> a [0]);
		for (i = 0; i < n; p1++, i++) {
			++counts [p1 -> pnum];
			++total;
		}
	}

	ip	= NEWA (total, int);
	buckets	= NEWA (nterms + 1, int *);

	for (i = 0; i < nterms; i++) {
		buckets [i]	= ip;
		ptrs [i]	= ip;
		ip += counts [i];
	}
	buckets [i] = ip;

	for (fp = fsp; fp NE NULL; fp = fp -> next) {
		terms = fp -> terminals;
		n = terms -> n;
		p1 = &(terms -> a [0]);
		for (i = 0; i < n; p1++, i++) {
			*(ptrs [p1 -> pnum])++ = fp -> tree_num;
		}
	}

	return (buckets);
}

/*
 * This routine pre-initializes the two tables containing compatibility/
 * incompatibility information in bit-mask form.
 */

	static
	void
init_compat_bitmasks (

struct cinfo *		cip	/* IN - compatibility info. */
)
{
int			i;
int			j;
int			n;
int			nft;
bitmap_t *		bp1;
bitmap_t *		bp2;
bitmap_t *		cp;
bitmap_t *		incp;
int			nmasks;
int8u *			carray;
enum compat		cval;

	nft	= cip -> full_tree_count;
	nmasks	= BMAP_ELTS (nft);

	n = (nft * nmasks);

	cip -> cmasks	= NEWA (n, bitmap_t);
	cip -> incmasks	= NEWA (n, bitmap_t);

	bp1 = cip -> cmasks;
	for (j = 0; j < n; j++) {
		*bp1++ = 0;
	}

	bp2 = cip -> incmasks;
	for (j = 0; j < n; j++) {
		*bp2++ = 0;
	}

	cp	= cip -> cmasks;
	incp	= cip -> incmasks;
	carray	= cip -> carray;

	for (i = 0; i < nft; cp += nmasks, incp += nmasks, i++) {
		for (j = 0; j < nft; j++) {
			cval = *carray++;
			if (cval EQ COMPATIBLE) {
				SETBIT (cp, j);
			}
			else if (cval EQ INCOMPATIBLE) {
				SETBIT (incp, j);
			}
		}
	}
}

/*
 * This routine pre-initializes a table of bit-masks, indexed by terminal
 * number.  For each terminal, it lists the set of full-trees that involve
 * it.
 */

	static
	void
init_term_tree_masks (

struct cinfo *		cip	/* IN - compatibility info. */
)
{
int			i;
int			n;
bitmap_t *		bp1;
int *			tp1;
int *			endp;
int			t;
int			nmasks;
int			nterms;

	nterms = cip -> num_terms;
	nmasks	= BMAP_ELTS (cip -> full_tree_count);

	n = (nterms * nmasks);

	bp1 = NEWA (n, bitmap_t);

	cip -> term_tree_masks = bp1;

	for (i = 0; i < n; i++) {
		*bp1++ = 0;
	}

	bp1 = cip -> term_tree_masks;
	for (i = 0; i < nterms; i++) {
		tp1	= cip -> term_trees [i];
		endp	= cip -> term_trees [i+1];
		while (tp1 < endp) {
			t = *tp1++;
			SETBIT (bp1, t);
		}
		bp1 += nmasks;
	}
}

/*
 * This routine uses compatibility criteria to further reduce the number
 * of full-sets.  These tests can only be applied after we have all potential
 * full-sets enumerated.
 */

#define	DEBUG_INITIAL_PRUNING	0

	void
initial_pruning (

bitmap_t *		tmap,		/* IN - subset of terminals to use */
bitmap_t *		fset_mask,	/* IN - subset of full-sets to use */
struct cinfo *		cip		/* IN - compatibility info */
)
{
int			i;
int			j;
int			n;
struct point *		p1;
bool			changed;
struct full_set *	fsp;
int32u			count;
bitmap_t *		tmp_fsmask;

#if DEBUG_INITIAL_PRUNING
 { int *		ip1;
   int *		ip2;

	(void) printf ("\n%% Dump of term trees:\n");
	for (i = 0; i < cip -> num_terms; i++) {
		(void) printf (" %% Term: %3d,  Trees:", i);
		ip1 = cip -> term_trees [i];
		ip2 = cip -> term_trees [i+1];
		while (ip1 < ip2) {
			(void) printf (" %d", *ip1++);
		}
		(void) printf ("\n");
	}
 }
#endif

	tmp_fsmask = NEWA (cip -> num_fset_masks, bitmap_t);

	if (Use_Thorough_Pruning) {
		/* Use the two more powerful rules... */
		do {
			changed = compat_test_4 (tmap,
						 fset_mask,
						 tmp_fsmask,
						 cip);
			if (compat_test_5 (tmap, fset_mask, tmp_fsmask, cip)) {
				changed = TRUE;
			}
		} while (changed);
	}
	else {
		/* Before we try the expensive ones, do a quick check	*/
		/* (compat test 3) for any OBVIOUSLY bogus full-sets.	*/
		(void) compat_test_3 (tmap, fset_mask, cip);

		do {
			changed = compat_test_1 (tmap, fset_mask, cip);
			if (compat_test_2 (tmap, fset_mask, cip)) {
				changed = TRUE;
			}
			if (compat_test_3 (tmap, fset_mask, cip)) {
				changed = TRUE;
			}
		} while (changed);
	}

	print_compatibility_matrix (cip);

	count = 0;
	for (j = 0; j < cip -> full_tree_count; j++) {
		if (NOT BITON (fset_mask, j)) continue;
		if (still_viable (j, tmap, fset_mask, cip)) {
			++count;
		}
		else {
			CLRBIT (fset_mask, j);
		}
	}

	(void) printf ("  %%  ip = %lu\n", count);
}

/*
 * This routine detects incompatible full-sets with the following rule:
 *
 *	Two trees I and J are incompatible if there exists a terminal
 * 	X not in (I union J) such that every full-set containing X is
 *	incompatible with either I or J.
 *
 * If I and J were ever to be included in the same solution, there would
 * be no possible way for the solution to contain terminal X.
 */

	static
	bool
compat_test_1 (

bitmap_t *		tmap,		/* IN - subset of terminals to use */
bitmap_t *		fset_mask,	/* IN - subset of full-sets to use */
struct cinfo *		cip		/* IN - compat info. */
)
{
int			I;
int			J;
int			nft;
int8u *			carray;
bool			changed;

	changed = FALSE;

	nft		= cip -> full_tree_count;
	carray		= cip -> carray;

	for (I = 0; I < nft; I++) {
		if (NOT BITON (fset_mask, I)) continue;
		for (J = I+1; J < nft; J++) {
			if (NOT BITON (fset_mask, J)) continue;
			if (carray [I*nft + J] EQ INCOMPATIBLE) continue;
			if (NOT okay_test_1 (I, J, tmap, fset_mask, cip)) {
				carray [I*nft + J] = INCOMPATIBLE;
				carray [J*nft + I] = INCOMPATIBLE;
				changed = TRUE;
			}
		}
	}

	return (changed);
}

/*
 * This routine returns TRUE if-and-only-if for every terminal X not in
 * (I union J) there is a tree involving X that is simultaneously
 * compatible with IP and JP.
 */

	static
	bool
okay_test_1 (

int			I,		/* IN - full-set I. */
int			J,		/* IN - full-set J. */
bitmap_t *		tmap,		/* IN - subset of terminals to use */
bitmap_t *		fset_mask,	/* IN - subset of full-sets to use */
struct cinfo *		cip		/* IN - compat info. */
)
{
int			X;
int			T;
int			n;
int			nft;
int *			p1;
int *			endp;
struct full_set *	ip;
struct full_set *	jp;
int8u *			carray;
int **			term_trees;

	n = cip -> num_terms;

	ip	= cip -> full_trees [I];
	jp	= cip -> full_trees [J];

	nft		= cip -> full_tree_count;
	carray		= cip -> carray;
	term_trees	= cip -> term_trees;

	for (X = 0; X < n; X++) {
		if (NOT BITON (tmap, X)) continue;
		if (BITON (ip -> tmap, X)) continue;
		if (BITON (jp -> tmap, X)) continue;
		p1	= term_trees [X];
		endp	= term_trees [X+1];
		while (TRUE) {
			if (p1 >= endp) {
				/* Every tree involving this point is */
				/* incompatible with I or J. */
#if DEBUG_INITIAL_PRUNING
  (void) printf ("%% Full sets %lu and %lu (which were %s) become\n",
		 I, J, compat_name (carray [I*nft + J]));
  (void) printf ("%%   INCOMPATIBLE because of point %lu\n", X);
#endif

				return (FALSE);
			}
			T = *p1++;
			if (BITON (fset_mask, T) AND
			    (carray [T*nft + I] NE INCOMPATIBLE) AND
			    (carray [T*nft + J] NE INCOMPATIBLE)) {
				/* There is a tree involving this point */
				/* that works simultaneously with I and J. */
				/* On to the next point... */
				break;
			}
		}
	}

	/* Every point is part of a tree that can work with I and J. */

	return (TRUE);
}

/*
 * This routine converts an "enum compat" into a "char *" for debugging.
 */

	static
	char *
compat_name (

enum compat		c		/* IN - compatibility value */
)
{
	switch (c) {
	case COMPATIBLE:		return ("COMPATIBLE");
	case INCOMPATIBLE:		return ("INCOMPATIBLE");
	case DISJOINT:			return ("DISJOINT");
	default:			return ("???");
	}
}

/*
 * This routine detects incompatible full-sets with the following rule:
 *
 *	Two trees I and J that share a common terminal are incompatible
 *	if there exists a terminal X not in (I union J) such that every
 *	full-set containing X is either incompatible with (I union J) or
 *	shares 2 or more terminals with (I union J).
 *
 * If I and J were ever to be included in the same solution, there would
 * be no possible way for the solution to contain terminal X.
 */

	static
	bool
compat_test_2 (

bitmap_t *		tmap,		/* IN - subset of terminals to use */
bitmap_t *		fset_mask,	/* IN - subset of full-sets to use */
struct cinfo *		cip		/* IN - compat info. */
)
{
int			I;
int			J;
int			nft;
int8u *			carray;
bool			changed;

	changed = FALSE;

	nft		= cip -> full_tree_count;
	carray		= cip -> carray;

	for (I = 0; I < nft; I++) {
		if (NOT BITON (fset_mask, I)) continue;
		for (J = I+1; J < nft; J++) {
			if (NOT BITON (fset_mask, J)) continue;
			if (carray [I*nft + J] NE COMPATIBLE) continue;
			if (NOT okay_test_2 (I, J, tmap, fset_mask, cip)) {
				carray [I*nft + J] = INCOMPATIBLE;
				carray [J*nft + I] = INCOMPATIBLE;
				changed = TRUE;
			}
		}
	}

	return (changed);
}

/*
 * This routine returns TRUE if-and-only-if for every terminal X not in
 * (I union J) there is a tree T involving X that is simultaneously
 * compatible with I and J.
 */

	static
	bool
okay_test_2 (

int			I,		/* IN - full-set I. */
int			J,		/* IN - full-set J. */
bitmap_t *		tmap,		/* IN - subset of terminals to use */
bitmap_t *		fset_mask,	/* IN - subset of full-sets to use */
struct cinfo *		cip		/* IN - compat info. */
)
{
int			X;
int			T;
int			k;
int			n;
int			nft;
int *			p1;
int *			endp;
struct full_set *	ip;
struct full_set *	jp;
struct full_set *	tp;
int			mask;
int			count;
struct full_set **	full_trees;
int8u *			carray;
int **			term_trees;
bitmap_t		iandj [N_TERM_BMAP];
bitmap_t		other_terms [N_TERM_BMAP];

	n = cip -> num_terms;

	nft		= cip -> full_tree_count;
	full_trees	= cip -> full_trees;
	carray		= cip -> carray;
	term_trees	= cip -> term_trees;

	ip = full_trees [I];
	jp = full_trees [J];

	/* Compute the union of I and J. */
	for (k = 0; k < N_TERM_BMAP; k++) {
		iandj [k] = (ip -> tmap [k] | jp -> tmap [k]);
	}

	/* Compute set of terminals that do not include I and J. */
	for (k = 0; k < N_TERM_BMAP; k++) {
		other_terms [k] = tmap [k] & ~iandj [k];
	}

	for (X = 0; X < n; X++) {
		if (NOT BITON (other_terms, X)) continue;
		p1	= term_trees [X];
		endp	= term_trees [X+1];
		while (TRUE) {
			if (p1 >= endp) {
				/* Every tree involving this point is */
				/* incompatible with I or J. */

#if DEBUG_INITIAL_PRUNING
  (void) printf ("%% Full sets %lu and %lu (which were COMPATIBLE) become\n",
		 I, J);
  (void) printf ("%%   INCOMPATIBLE because of point %lu\n", X);
#endif

				return (FALSE);
			}
			T = *p1++;
			if (NOT BITON (fset_mask, T)) continue;
			tp = full_trees [T];
			count = 0;
			for (k = 0; k < N_TERM_BMAP; k++) {
				mask = (tp -> tmap [k] & iandj [k]);
				count += NBITSON (mask);
			}
			if ((count <= 1) AND
			    (carray [T*nft + I] NE INCOMPATIBLE) AND
			    (carray [T*nft + J] NE INCOMPATIBLE)) {
				/* There is a tree involving this point */
				/* that works simultaneously with I and J. */
				/* On to the next point... */
				break;
			}
		}
	}

	/* Every point is part of a tree that can work with I and J. */

	return (TRUE);
}

/*
 * This routine detects incompatible full-sets with the following rule:
 *
 *	A tree J is incompatible with all other trees if there exists a
 *	terminal X not in J such that every full-set containing X is
 *	incompatible with J.
 *
 * If J were ever to be included in a solution, there would be no possible
 * way for the solution to contain terminal X.
 */

	static
	bool
compat_test_3 (

bitmap_t *		tmap,		/* IN - subset of terminals to use */
bitmap_t *		fset_mask,	/* IN - subset of full-sets to use */
struct cinfo *		cip		/* IN - compat info. */
)
{
int			J;
int			k;
int			nft;
int8u *			carray;
bool			changed;

	changed = FALSE;

	nft		= cip -> full_tree_count;
	carray		= cip -> carray;

	for (J = 0; J < nft; J++) {
		if (NOT BITON (fset_mask, J)) continue;
		if (NOT still_viable (J, tmap, fset_mask, cip)) continue;
		if (NOT okay_test_3 (J, tmap, fset_mask, cip)) {
			for (k = 0; k < nft; k++) {
				carray [k*nft + J] = INCOMPATIBLE;
				carray [J*nft + k] = INCOMPATIBLE;
			}
			changed = TRUE;
		}
	}

	return (changed);
}

/*
 * This routine returns TRUE if-and-only-if for every terminal X not in
 * J there is a tree T involving X that is compatible with J.
 */

	static
	bool
okay_test_3 (

int			J,		/* IN - full-set J. */
bitmap_t *		tmap,		/* IN - subset of terminals to use */
bitmap_t *		fset_mask,	/* IN - subset of full-sets to use */
struct cinfo *		cip		/* IN - compat info. */
)
{
int			X;
int			T;
int			n;
int			nft;
struct full_set *	jp;
int *			p1;
int *			endp;
int8u *			carray;
int **			term_trees;

	n = cip -> num_terms;

	jp		= cip -> full_trees [J];
	nft		= cip -> full_tree_count;
	carray		= cip -> carray;
	term_trees	= cip -> term_trees;

	for (X = 0; X < n; X++) {
		if (NOT BITON (tmap, X)) continue;
		if (BITON (jp -> tmap, X)) continue;
		p1	= term_trees [X];
		endp	= term_trees [X+1];
		while (TRUE) {
			if (p1 >= endp) {
				/* Every tree involving this point is */
				/* incompatible with J. */

#if DEBUG_INITIAL_PRUNING
  (void) printf ("%% Full set %lu is ELIMINATED because of point %lu!\n",
		 J, X);
#endif

				return (FALSE);
			}
			T = *p1++;
			if (NOT BITON (fset_mask, T)) continue;
			if (carray [T*nft + J] NE INCOMPATIBLE) {
				/* There is a tree involving this point */
				/* that works with J. */
				/* On to the next point... */
				break;
			}
		}
	}

	/* Every point is part of at least one tree that can work with J. */

	return (TRUE);
}

/*
 * This routine checks to see if the given full-set number is still viable.
 * A set is viable if there is ANY other full-set that it is compatible with.
 *
 * As a special detail, it is not possible for a full-set that contains ALL
 * of the original terminals to become non-viable.
 */

	static
	bool
still_viable (

int		j,		/* IN - full-set number to test viability of */
bitmap_t *	tmap,		/* IN - subset of terminals to use */
bitmap_t *	fset_mask,	/* IN - subset of full-sets to use */
struct cinfo *	cip		/* IN - compat info. */
)
{
int			i;
int			n;
int			nft;
struct full_set *	jp;
int8u *			carray;
bitmap_t *		bp1;
bitmap_t *		bp2;

	n		= cip -> num_terms;
	nft		= cip -> full_tree_count;
	carray		= cip -> carray;

	jp = cip -> full_trees [j];

	i = 0;
	bp1 = &tmap [0];
	bp2 = &(jp -> tmap [0]);
	while (TRUE) {
		if (i >= cip -> num_term_masks) {
			/* Special case -- entire terminal set is	*/
			/* always viable!				*/
			return (TRUE);
		}
		if (*bp1++ NE *bp2++) break;
		++i;
	}

	carray = &carray [j*nft];
	for (i = 0; i < nft; carray++, i++) {
		if (NOT BITON (fset_mask, i)) continue;
		if (*carray EQ COMPATIBLE) {
			return (TRUE);
		}
	}

	/* This full set is not compatible with anything else!  NOT VIABLE! */
	return (FALSE);
}

/*
 * This routine detects incompatible full-sets with the following rule:
 *
 *	A tree J is incompatible with all other trees if it disconnects
 *	the graph when included in any solution.
 */

	static
	bool
compat_test_4 (

bitmap_t *		tmap,		/* IN - subset of terminals to use */
bitmap_t *		fset_mask,	/* IN - subset of full-sets to use */
bitmap_t *		tmp_fsmask,	/* IN - temporary full-set mask */
struct cinfo *		cip		/* IN - compat info. */
)
{
int			J;
int			k;
int			nft;
int			nmasks;
int8u *			carray;
bool			changed;

#if DEBUG_INITIAL_PRUNING
	(void) printf (" %% Entering compat_test_4\n");
#endif

	changed = FALSE;

	nft		= cip -> full_tree_count;
	nmasks		= cip -> num_fset_masks;
	carray		= cip -> carray;

	for (J = 0; J < nft; J++) {
		if (NOT BITON (fset_mask, J)) continue;
		if (NOT okay_test_4 (J, tmap, fset_mask, tmp_fsmask, cip)) {
			for (k = 0; k < nft; k++) {
				carray [k*nft + J] = INCOMPATIBLE;
				carray [J*nft + k] = INCOMPATIBLE;
				SETBIT (&(cip -> incmasks [k*nmasks]), J);
				CLRBIT (&(cip -> cmasks [k*nmasks]), J);
				SETBIT (&(cip -> incmasks [J*nmasks]), k);
				CLRBIT (&(cip -> cmasks [J*nmasks]), k);
			}
			CLRBIT (fset_mask, J);
#if DEBUG_INITIAL_PRUNING
			(void) printf (" %% Piece %d removed because of split!\n",
				       J);
#endif
			changed = TRUE;
		}
	}

#if DEBUG_INITIAL_PRUNING
	(void) printf (" %% Exiting compat_test_4\n");
#endif

	return (changed);
}

/*
 * This routine returns TRUE if-and-only-if the graph of all remaining
 * full-sets is connected when tree J is included and all trees that are
 * incompatible with J are excluded.
 */

	static
	bool
okay_test_4 (

int			J,		/* IN - full-set J. */
bitmap_t *		tmap,		/* IN - subset of terminals to use */
bitmap_t *		fset_mask,	/* IN - subset of full-sets to use */
bitmap_t *		tmp_fsmask,	/* IN - temporary full-set mask */
struct cinfo *		cip		/* IN - compat info. */
)
{
int			i;
int			nmasks;
int			kmasks;
struct full_set *	jp;
bitmap_t *		bp1;
bitmap_t *		bp2;
bitmap_t *		bp3;
bitmap_t		tmp_tmask [N_TERM_BMAP];

	kmasks = cip -> num_term_masks;
	nmasks = cip -> num_fset_masks;

	/* Create a new set of full-sets that exclude all that are	*/
	/* incompatible with full-set J.				*/
	bp1 = tmp_fsmask;
	bp2 = fset_mask;
	bp3 = &(cip -> incmasks [J * nmasks]);
	for (i = 0; i < nmasks; i++) {
		*bp1++ = *bp2++ & ~(*bp3++);
	}
	SETBIT (tmp_fsmask, J);

	/* Create a MODIFIABLE mask of the terminals... */
	bp1 = &tmp_tmask [0];
	bp2 = tmap;
	for (i = 0; i < kmasks; i++) {
		*bp1++ = *bp2++;
	}

	jp		= cip -> full_trees [J];

	/* Recursively walk to all reachable terminals... */
	ip_walk (jp -> terminals -> a [0].pnum,
		 tmp_tmask,
		 tmp_fsmask,
		 cip -> full_trees,
		 cip -> term_trees);

	bp1 = &tmp_tmask [0];
	for (i = 0; i < kmasks; i++) {
		if (*bp1 NE 0) {
			/* Some terminals cannot be reached! */
			return (FALSE);
		}
		++bp1;
	}

	/* All terminals reachable -- Tree J does NOT split the graph! */

	return (TRUE);
}

/*
 * This routine detects incompatible full-sets with the following rule:
 *
 *	Two trees I and J are incompatible with each other if they cause
 *	the graph to become disconnected when included in any solution.
 */

	static
	bool
compat_test_5 (

bitmap_t *		tmap,		/* IN - subset of terminals to use */
bitmap_t *		fset_mask,	/* IN - subset of full-sets to use */
bitmap_t *		tmp_fsmask,	/* IN - temporary full-set mask */
struct cinfo *		cip		/* IN - compat info. */
)
{
int			I;
int			J;
int			k;
int			nft;
int			nmasks;
int8u *			carray;
bool			changed;

#if DEBUG_INITIAL_PRUNING
	(void) printf (" %% Entering compat_test_5\n");
#endif

	changed = FALSE;

	nft		= cip -> full_tree_count;
	nmasks		= cip -> num_fset_masks;

	for (I = 0; I < nft; I++) {
		if (NOT BITON (fset_mask, I)) continue;
		carray = &(cip -> carray [I * nft]);
		for (J = I+1; J < nft; J++) {
			if (NOT BITON (fset_mask, J)) continue;
			if (carray [J] EQ INCOMPATIBLE) continue;
			if (NOT okay_test_5 (I, J, tmap, fset_mask, tmp_fsmask, cip)) {
				carray [J]		  = INCOMPATIBLE;
				cip -> carray [J*nft + I] = INCOMPATIBLE;
				SETBIT (&(cip -> incmasks [I*nmasks]), J);
				CLRBIT (&(cip -> cmasks [I*nmasks]), J);
				SETBIT (&(cip -> incmasks [J*nmasks]), I);
				CLRBIT (&(cip -> cmasks [J*nmasks]), I);
#if DEBUG_INITIAL_PRUNING
				(void) printf (" %% Pieces %d and %d become incompatible!\n",
					       I, J);
 #if 0
				plot_full_set_split_pair (I, J, fset_mask, cip);
 #endif
#endif
				changed = TRUE;
			}
		}
	}

#if DEBUG_INITIAL_PRUNING
	(void) printf (" %% Exiting compat_test_5\n");
#endif

	return (changed);
}

/*
 * This routine returns TRUE if-and-only-if the graph of all full-sets
 * that remain when trees I and J are included (and all incompatible
 * trees excluded) remains connected.
 */

	static
	bool
okay_test_5 (

int			I,		/* IN - full-set I. */
int			J,		/* IN - full-set J. */
bitmap_t *		tmap,		/* IN - subset of terminals to use */
bitmap_t *		fset_mask,	/* IN - subset of full-sets to use */
bitmap_t *		tmp_fsmask,	/* IN - temporary full-set mask */
struct cinfo *		cip		/* IN - compat info. */
)
{
int			k;
int			nmasks;
int			kmasks;
struct full_set *	jp;
bitmap_t *		bp1;
bitmap_t *		bp2;
bitmap_t *		bp3;
bitmap_t *		bp4;
bitmap_t		tmp_tmask [N_TERM_BMAP];

	kmasks = cip -> num_term_masks;
	nmasks = cip -> num_fset_masks;

	/* Create a new set of full-sets that exclude all that are	*/
	/* incompatible with full-sets I or J.				*/
	bp1 = tmp_fsmask;
	bp2 = fset_mask;
	bp3 = &(cip -> incmasks [I * nmasks]);
	bp4 = &(cip -> incmasks [J * nmasks]);
	for (k = 0; k < nmasks; k++) {
		*bp1++ = *bp2++ & ~(*bp3++ | *bp4++);
	}
	SETBIT (tmp_fsmask, I);
	SETBIT (tmp_fsmask, J);

	/* Create a MODIFIABLE mask of the terminals... */
	bp1 = &tmp_tmask [0];
	bp2 = tmap;
	for (k = 0; k < kmasks; k++) {
		*bp1++ = *bp2++;
	}

	jp		= cip -> full_trees [J];

	/* Recursively walk to all reachable terminals... */
	ip_walk (jp -> terminals -> a [0].pnum,
		 tmp_tmask,
		 tmp_fsmask,
		 cip -> full_trees,
		 cip -> term_trees);

	bp1 = &tmp_tmask [0];
	for (k = 0; k < kmasks; k++) {
		if (*bp1 NE 0) {
			/* Some terminals cannot be reached! */
			return (FALSE);
		}
		++bp1;
	}

	/* All terminals are reachable -- Trees I and J do NOT	*/
	/* mutually cause the graph to split!			*/

	return (TRUE);
}

	static
	void
ip_walk (

int		pnum,		/* IN - point to walk from */
bitmap_t *	terms_left,	/* IN/OUT - terminals not yet visted */
bitmap_t *	fs_mask,	/* IN/OUT - full-trees not yet visited */
struct full_set ** full_trees,	/* IN - array of full-trees */
int **		term_trees	/* IN - trees involving each terminal */
)
{
int			i;
int			j;
int			n;
struct point *		p1;
struct pset *		terms;
struct full_set *	fsp;
int *			tp;
int *			endp;
int			t;

	/* Unmark this terminal so that we don't visit it again. */
	CLRBIT (terms_left, pnum);

	/* Walk every full-tree involving this terminal... */
	tp	= term_trees [pnum];
	endp	= term_trees [pnum + 1];
	while (tp < endp) {
		t = *tp++;
		if (BITON (fs_mask, t)) {
			/* This full-tree has now been visited... */
			CLRBIT (fs_mask, t);
			terms = full_trees [t] -> terminals;
			n = terms -> n;
			p1 = &(terms -> a [0]);
			for (i = 0; i < n; p1++, i++) {
				if (NOT BITON (terms_left, p1 -> pnum)) continue;
				ip_walk (p1 -> pnum,
					 terms_left,
					 fs_mask,
					 full_trees,
					 term_trees);
			}
		}
	}
}

/*
 * This routine prints out the compatibility matrix.
 */

	static
	void
print_compatibility_matrix (

struct cinfo *	cip		/* IN - compat info. */
)
{
int		i;
int		j;
int		nft;
int8u *		carray;
enum compat	compat;
char		c;

	if (NOT Print_Compatibility_Matrix) {
		return;
	}

	nft	= cip -> full_tree_count;
	carray	= cip -> carray;

	(void) printf ("    0         1         2         3         4         5         6         7\n");
	(void) printf ("    01234567890123456789012345678901234567890123456789012345678901234567890\n");
	(void) printf ("   +-----------------------------------------------------------------------\n");
	for (i = 0; i < nft; i++) {
		(void) printf ("%2d |", i);
		for (j = 0; j < nft; j++) {
			compat = *carray++;
			switch (compat) {
			case COMPATIBLE:	c = 'C';	break;
			case INCOMPATIBLE:	c = '.';	break;
			case DISJOINT:		c = ' ';	break;
			default:		c = '?';	break;
			}
			(void) putchar (c);
		}
		(void) printf ("\n");
	}
}
