/***********************************************************************

	File:	cputime.c
	Rev:	a-1
	Date:	04/16/93

	Copyright (c) 1993 by The University of Virginia

************************************************************************

	This file contains routines for computing CPU time
	consumption for the program.

************************************************************************

	Modification Log:

	a-1:	04/16/93	dmw
		: Created.

************************************************************************/

#include "steiner.h"

#ifdef UNIX_CPU_TIME
#include <sys/types.h>
#include <sys/times.h>
 #ifndef CLK_TCK
 #define CLK_TCK	60	/* for SunOS... */
 #endif
#else
#include <time.h>
#endif


/*
 * Global Routines
 */

cpu_time_t	get_cpu_time (void);
void		convert_cpu_time (cpu_time_t, char *);


/*
 * Local Macros
 */

	/* Convert ticks from one set of units to another, rouding */
	/* properly, and carefully avoiding overflow.		   */

#define CV_TICK_UNITS(ticks, from_per_sec, to_per_sec)			\
	if (from_per_sec < to_per_sec) {				\
		if ((to_per_sec % from_per_sec) EQ 0) {			\
			ticks *= (to_per_sec / from_per_sec);		\
		}							\
		else {							\
			ticks = (2 * to_per_sec * ticks + from_per_sec);\
			ticks /= (2 * from_per_sec);			\
		}							\
	}								\
	else if (from_per_sec > to_per_sec) {				\
		if ((from_per_sec % to_per_sec) EQ 0) {			\
			ticks *= (from_per_sec / to_per_sec);		\
		}							\
		else {							\
			ticks = (2 * from_per_sec * ticks + to_per_sec);\
			ticks /= (2 * to_per_sec);			\
		}							\
	}

/*
 * This routine gets the current amount of CPU time that the program has
 * used.  It is returned in a format we define that is independent of the
 * operating system (i.e. hundreth's of a second).
 *
 * Because the times we are measuring can be quite long (several days),
 * we can run into overflow problems if the operating system gives us
 * a CPU timer with too high a resolution (i.e. microseconds).  This
 * code can be compiled to use either the standard interfaces defined
 * by ANSI C, or the native UNIX interface.  Choose the one that has
 * a resolution closest to our own TICKS_PER_SEC units.
 */

	cpu_time_t
get_cpu_time (void)

{
clock_t		total;
int32u		seconds;
int32u		ticks;
cpu_time_t	cpu_time;

#ifdef UNIX_CPU_TIME
	{ struct tms	t;

	  times (&t);

	  total = t.tms_utime + t.tms_stime + t.tms_cutime + t.tms_cstime;

	  seconds = total / CLK_TCK;
	  ticks   = total % CLK_TCK;

	  CV_TICK_UNITS (ticks, CLK_TCK, TICKS_PER_SEC);
	}
#else
	/* Using the ANSI C defined interface... */

	total = clock ();

	seconds = total / CLOCKS_PER_SEC;
	ticks	= total % CLOCKS_PER_SEC;

	CV_TICK_UNITS (ticks, CLOCKS_PER_SEC, TICKS_PER_SEC);
#endif

	cpu_time = seconds * TICKS_PER_SEC + ticks;

	return (cpu_time);
}

/*
 * This routine will convert the given CPU time into a printable
 * null-terminated ASCII string.  The answer contains two decimal places.
 */

	void
convert_cpu_time (

cpu_time_t	time,
char *		out_buf
)
{
cpu_time_t	secs;
cpu_time_t	ticks;
char *		p;
char *		endp;
char		buf [20];

#define	ZERO	((cpu_time_t) 0)
#define	TEN	((cpu_time_t) 10)

	secs	= time / TICKS_PER_SEC;
	ticks	= time % TICKS_PER_SEC;

	p = &buf [0];
	if (secs <= ZERO) {
		*p++ = '0';
	}
	else {
		while (secs > ZERO) {
			*p++ = (secs % TEN) + '0';
			secs /= TEN;
		}
	}
	while (p > &buf [0]) {
		*out_buf++ = *--p;
	}
	*out_buf++ = '.';
	ticks *= TEN;
	*out_buf++ = (ticks / TICKS_PER_SEC) + '0';
	ticks %= TICKS_PER_SEC;
	ticks *= TEN;
	*out_buf++ = (ticks / TICKS_PER_SEC) + '0';
	*out_buf++ = '\0';

#undef ZERO
#undef TEN
}
