/***********************************************************************

	File:	decomp.c
	Rev:	a-1
	Date:	04/03/93

	Copyright (c) 1993, by The University of Virginia

************************************************************************

	Routines to decompose the search problem into smaller
	sub-problems.

************************************************************************

	Modification Log:

	a-1:	04/03/93	dmw
		: Created.

************************************************************************/

#include "steiner.h"


/*
 * Global Routines
 */

struct full_set *	copy_full_set (struct full_set *);
void			print_mask (char *, bitmap_t *, int);
void			print_tree (char *, bitmap_t *, struct cinfo *);
bool			split_and_solve (bitmap_t *,
					 bitmap_t *,
					 struct cinfo *,
					 bitmap_t *);
dist_t			total_tree_length (bitmap_t *, struct cinfo *);


/*
 * External References
 */

	/* none */

/*
 * Local Types
 */

struct bicon {
	int		n;		/* number of vertices. */
	int		rowsize;	/* graph's row-size. */
	bitmap_t *	g;		/* the graph. */
	int *		dfs_num;	/* DFS number of each node. */
	int *		back;		/* lowest DFS num in component. */
	pnum_t *	stack;		/* base-address of stack. */
	int *		comps;		/* output components. */
	pnum_t *	sp;		/* current stack pointer. */
	int		counter;	/* DFS number generator. */
	int		ncomps;		/* count of BCC's generated. */
};


/*
 * Local Routines
 */

static bool		backtrack_search_forest (int,
						 int,
						 bitmap_t *,
						 bitmap_t *,
						 struct cinfo *,
						 bitmap_t *);
static void		bcc (struct bicon *, int);
static int		biconnected_components (bitmap_t *,
						bitmap_t *,
						struct cinfo *,
						int *);
static void		compute_tmap (struct pset *, bitmap_t *);
static void		compute_fset_subset (bitmap_t *,
					     bitmap_t *,
					     struct cinfo *,
					     bitmap_t *);
static bitmap_t *	construct_graph (bitmap_t *,
					 bitmap_t *,
					 struct cinfo *);
static bool		fix_comps (struct bicon *,
				   int,
				   bitmap_t *,
				   bitmap_t *,
				   bitmap_t *,
				   struct cinfo *,
				   int *,
				   int *,
				   int **);
static bool		is_terminal_subset (bitmap_t *, bitmap_t *);
static int		more_connected_components (bitmap_t *,
						   bitmap_t *,
						   struct cinfo *,
						   int *,
						   int **);
static void		omit_and_bicon (bitmap_t *,
					struct bicon *,
					int,
					bitmap_t *);
static bool		only_one_fake_full_set (bitmap_t *,
						bitmap_t *,
						struct cinfo *);
static bool		perform_cycle_busting (bitmap_t *,
					       bitmap_t *,
					       struct cinfo *,
					       bitmap_t *);
static struct full_set * remap (struct pset *, struct full_set *, struct pset *);

/*
 * This routine first splits the problem into smaller sub-problems, if
 * possible.  It then uses the brute-force backtrack search to find an
 * optimal solution for each sub-problem.
 *
 * The way we break up the problem is to examine the graph G in which nodes
 * A and B are connected if point A and point B both belong to a full-set.
 * In this graph, the bi-connected components are separated by articulation
 * points such that all paths between the two components must pass through
 * this articulation point boundary.  The sub-problems corresponding to each
 * component are separately solved, and their solutions are trivially
 * concatenated to form the final solution.
 *
 * If a graph G is K-connected, it is possible to split the graph by removing
 * K vertices, although as K gets larger it becomes more difficult to put the
 * separate solutions back together into an optimal final solution.
 *
 * Note:  this routine assumes that the input point-set is numbered
 *	  sequentially.  (e.g. pts -> a [I].pnum EQ I for 0 < I < pts -> n).
 */

	bool
split_and_solve (

bitmap_t *		tmap,		/* IN - subset of terminals to use */
bitmap_t *		fset_mask,	/* IN - subset of full-sets to use */
struct cinfo *		cip,		/* IN - compatibility info */
bitmap_t *		smt_mask	/* OUT - SMT as a set of full-sets */
)
{
int			i;
int			j;
int			k;
int			n;
int *			comps;
int			ncomps;
bitmap_t *		bp1;
bitmap_t *		bp2;
bitmap_t *		new_fset_mask;
bitmap_t *		new_smt_mask;
bool			failed;
bitmap_t		new_tmap [N_TERM_BMAP];
int			vertices [MAX_TERMINALS];

	n = cip -> num_terms;

	comps = NEWA (n * n, int);

	ncomps = biconnected_components (tmap, fset_mask, cip, comps);

	if (ncomps <= 1) {
		(void) free ((char *) comps);
		failed = perform_cycle_busting (tmap,
						fset_mask,
						cip,
						smt_mask);
		return (failed);
	}

	/* First, zero out the final answer bit-mask. */
	bp1 = smt_mask;
	for (i = 0; i < cip -> num_fset_masks; i++) {
		*bp1++ = 0;
	}

	/* Create bitmasks for a sub-problem and its answer... */
	new_fset_mask = NEWA (cip -> num_fset_masks, bitmap_t);
	new_smt_mask  = NEWA (cip -> num_fset_masks, bitmap_t);

	for (i = 0; i < n; i++) {
		vertices [i] = 0;
	}

	failed = FALSE;

	/* Solve each component, combining results. */
	for (k = 1; k <= ncomps; k++) {
		for (i = 0; i < n; i++) {
			for (j = 0; j < n; j++) {
				if (comps [i*n + j] EQ k) {
					vertices [i] = k;
					vertices [j] = k;
				}
			}
		}
		/* Create mask of terminals for this component. */
		for (i = 0; i < N_TERM_BMAP; i++) {
			new_tmap [i] = 0;
		}
		(void) printf (" %% Component %d:", k);
		for (i = 0; i < n; i++) {
			if (vertices [i] NE k) continue;
			(void) printf (" %d", i);
			SETBIT (new_tmap, i);
		}
		(void) printf ("\n");

		compute_fset_subset (new_tmap, fset_mask, cip, new_fset_mask);

		/* Solve component, busting cyclic components if possible. */
		failed = perform_cycle_busting (new_tmap,
						new_fset_mask,
						cip,
						new_smt_mask);

		/* No solution is possible if ANY sub-problem fails! */
		if (failed) break;

		bp1 = new_smt_mask;
		bp2 = smt_mask;
		for (i = 0; i < cip -> num_fset_masks; i++) {
			*bp2++ |= *bp1++;
		}
	}

	free ((char *) new_smt_mask);
	free ((char *) new_fset_mask);
	free ((char *) comps);

	return (failed);
}

/*
 * This routine constructs the graph G, that relates points co-resident in
 * the same full-set, and then computes the set of bi-connected components
 * of this graph.
 * The output is an N by N matrix giving the COMPONENT NUMBER that the
 * edge from I to J belongs to.  Note that component numbers begin at 1,
 * and that component 0 implies that the edge does not exist.
 */

	static
	int
biconnected_components (

bitmap_t *		tmap,		/* IN - subset of terminals to use */
bitmap_t *		fset_mask,	/* IN - subset of full-sets to use */
struct cinfo *		cip,		/* IN - compatibility info */
int *			comps		/* OUT - the components. */
)
{
int			i;
int			j;
int			n;
int *			ip;
bitmap_t *		graph;
struct bicon		bicon;
int			dfs [MAX_TERMINALS];
int			back [MAX_TERMINALS];

	n = cip -> num_terms;
	ip = comps;
	for (i = 0; i < n; i++) {
		for (j = 0; j < n; j++) {
			*ip++ = 0;
		}
	}

	bicon.n			= n;
	bicon.rowsize		= cip -> num_term_masks;
	bicon.g			= construct_graph (tmap, fset_mask, cip);
	bicon.dfs_num		= &dfs [0];
	bicon.back		= &back [0];
	bicon.stack		= NEWA (2 * n * n, pnum_t);
	bicon.comps		= comps;
	bicon.sp		= bicon.stack;
	bicon.counter		= 0;
	bicon.ncomps		= 0;

	ip = bicon.dfs_num;
	for (i = 0; i < n; i++) {
		*ip++ = 0;
	}

	ip = bicon.back;
	for (i = 0; i < n; i++) {
		*ip++ = 0;
	}

	for (i = 0; i < n; i++) {
		if (BITON (tmap, i)) {
			bcc (&bicon, i);
			break;
		}
	}

	free ((char *) bicon.stack);
	free ((char *) bicon.g);

	return (bicon.ncomps);
}

/*
 * This routine performs the recursive Depth-First-Search of the Graph
 * computing the Bi-Connected-Components along the way.  The algorithm is
 * standard, yet slightly subtle.
 */

	static
	void
bcc (

struct bicon *		bcp,		/* IN - recursion info. */
int			v		/* IN - current node in graph. */
)
{
int			i;
int			n;
int			w;
int			c;
bitmap_t *		g;
int			v1;
int			w1;

	n = bcp -> n;

	++(bcp -> counter);
	bcp -> dfs_num [v] = bcp -> counter;
	bcp -> back [v]	   = bcp -> counter;

	g = &(bcp -> g [v * bcp -> rowsize]);

	for (w = 0; w < bcp -> n; w++) {
		if (NOT BITON (g, w)) continue;
		if (bcp -> dfs_num [w] < bcp -> dfs_num [v]) {
			*(bcp -> sp)++ = v;
			*(bcp -> sp)++ = w;
		}
		if (bcp -> dfs_num [w] EQ 0) {
			bcc (bcp, w);
			if (bcp -> back [w] >= bcp -> dfs_num [v]) {
				c = ++(bcp -> ncomps);
				do {
					w1 = *--(bcp -> sp);
					v1 = *--(bcp -> sp);
					bcp -> comps [n*v1 + w1] = c;
					bcp -> comps [n*w1 + v1] = c;
				} while ((v1 NE v) OR (w1 NE w));
			}
			else if (bcp -> back [w] < bcp -> back [v]) {
				bcp -> back [v] = bcp -> back [w];
			}
		}
		else if (bcp -> dfs_num [w] < bcp -> back [v]) {
			bcp -> back [v] = bcp -> dfs_num [w];
		}
	}
}

/*
 * This routine builds the graph G (represented as a bit-matrix) wherein
 * nodes I and J are connected if-and-only-if points I and J are distinct
 * and are members of the same full-set.
 */

	static
	bitmap_t *
construct_graph (

bitmap_t *		tmap,		/* IN - subset of terminals to use */
bitmap_t *		fset_mask,	/* IN - subset of full-sets to use */
struct cinfo *		cip		/* IN - compatibility info */
)
{
int			i;
int			j;
int			k;
int			n;
int			nmasks;
bitmap_t *		graph;
bitmap_t *		bp1;
bitmap_t *		bp2;
struct full_set *	fsp;
struct pset *		terms;
struct point *		p1;
struct point *		p2;

	n	= cip -> num_terms;
	nmasks	= cip -> num_term_masks;

	n = (n * nmasks);

	graph = NEWA (n, bitmap_t);

	for (i = 0, bp1 = graph; i < n; bp1++, i++) {
		*bp1 = 0;
	}

	for (k = 0; k < cip -> full_tree_count; k++) {
		if (NOT BITON (fset_mask, k)) continue;
		fsp = cip -> full_trees [k];
		terms = fsp -> terminals;
		n = terms -> n;
		p1 = &(terms -> a [0]);
		for (i = 0; i < n; p1++, i++) {
			bp1 = &graph [p1 -> pnum * nmasks];
			p2 = p1 + 1;
			for (j = i + 1; j < n; p2++, j++) {
				bp2 = &graph [p2 -> pnum * nmasks];
				SETBIT (bp1, p2 -> pnum);
				SETBIT (bp2, p1 -> pnum);
			}
		}
	}

	return (graph);
}

/*
 * This routine renumbers the given set of points.  It also creates a COPY
 * of the SUBSET of the given full-sets that pertain to the given point set.
 * These full-sets are then also renumbered.
 */

	static
	struct full_set *
remap (

struct pset *		oldpts,		/* IN - input point set. */
struct full_set *	fsp,		/* IN - input full-sets. */
struct pset *		newpts		/* OUT - renumbered point set. */
)
{
int			i;
int			j;
int			n;
struct point *		p1;
struct full_set *	new_trees;
struct full_set *	newp;
struct full_set *	answer;
struct pset *		terms;
bitmap_t		tmap [N_TERM_BMAP];
int			forward_map [MAX_TERMINALS];

	/* Copy point set if necessary... */
	if (oldpts NE newpts) {
		COPY_PSET (newpts, oldpts);
	}

	compute_tmap (newpts, tmap);

	/* Construct new forward mappings. */
	for (i = 0; i < MAX_TERMINALS; i++) {
		forward_map [i] = -1;
	}

	n = newpts -> n;
	if (n > MAX_TERMINALS) {
		fatal ("remap: Bug 1.");
	}

	/* Create forward-mapping array, and renumber points. */
	p1 = &(newpts -> a [0]);
	for (i = 0; i < n; p1++, i++) {
		j = p1 -> pnum;
		if (j >= MAX_TERMINALS) {
			fatal ("remap: Bug 2.");
		}
		forward_map [j] = i;
		p1 -> pnum = i;
	}

	/* Get a list of the (renumbered) full-sets that involve only	*/
	/* points from this set.  This list is a COPY.			*/
	new_trees = NULL;
	while (fsp NE NULL) {
		if (is_terminal_subset (fsp -> tmap, tmap)) {
			newp = copy_full_set (fsp);
			terms = newp -> terminals;
			n = terms -> n;
			p1 = &(terms -> a [0]);
			for (i = 0; i < n; p1++, i++) {
				j = p1 -> pnum;
				if ((j < 0) OR (j >= MAX_TERMINALS)) {
					fatal ("remap: Bug 3.");
				}
				j = forward_map [j];
				if (j < 0) {
					fatal ("remap: Bug 4.");
				}
				p1 -> pnum = j;
			}
			compute_tmap (terms, newp -> tmap);
			newp -> next = new_trees;
			new_trees = newp;
		}
		fsp = fsp -> next;
	}

	return (new_trees);
}

/*
 * This routine computes a bit-mask of all of the point numbers present
 * in the given set of terminals.
 */

	static
	void
compute_tmap (

struct pset *		pts,		/* IN - set of terminals. */
bitmap_t *		tmap		/* OUT - bitmap of terminals. */
)
{
int			i;
int			n;
struct point *		p1;

	for (i = 0; i < N_TERM_BMAP; i++) {
		tmap [i] = 0;
	}

	n = pts -> n;
	p1 = &(pts -> a [0]);
	for (i = 0; i < n; p1++, i++) {
		SETBIT (tmap, p1 -> pnum);
	}
}

/*
 * This routine determines the subset of the given set of full-trees
 * that involves only terminals found in the given terminal set.
 */

	static
	void
compute_fset_subset (

bitmap_t *		tmap,		/* IN - subset of terminals to use */
bitmap_t *		fset_mask,	/* IN - subset of full-sets to use */
struct cinfo *		cip,		/* IN - compatibility info */
bitmap_t *		fset_subset	/* OUT - fsets that only use tmap */
)
{
int			i;
int			nmasks;
int			ntrees;
bitmap_t *		bp1;
bitmap_t *		bp2;
struct full_set *	fsp;

	nmasks = cip -> num_fset_masks;
	ntrees = cip -> full_tree_count;

	/* First, zero out the result... */
	bp1 = fset_subset;
	for (i = 0; i < nmasks; i++) {
		*bp1++ = 0;
	}

	for (i = 0; i < ntrees; i++) {
		if (NOT BITON (fset_mask, i)) continue;
		fsp = cip -> full_trees [i];
		if (NOT is_terminal_subset (fsp -> tmap, tmap)) continue;
		SETBIT (fset_subset, i);
	}
}

/*
 * This routine determines if terminal set X is a subset of Y.
 */

	static
	bool
is_terminal_subset (

bitmap_t *	xp,		/* IN - terminal set X. */
bitmap_t *	yp		/* IN - terminal set Y. */
)
{
int		i;

	for (i = 0; i < N_TERM_BMAP; i++) {
		if ((*xp & ~*yp) NE 0) return (FALSE);
		++xp;
		++yp;
	}

	return (TRUE);
}
/*
 * This routine creates a dynamically allocated COPY of the given full-set.
 * It also contains a COPY of the terminals.
 */

	struct full_set *
copy_full_set (

struct full_set *	fsp		/* IN - full-set to copy. */
)
{
struct full_set *	newp;

	newp = NEW (struct full_set);

	*newp = *fsp;
	newp -> terminals = save_pset (newp -> terminals);

	return (newp);
}

/*
 * This routine tries to do some "cycle-busting" by seeing if any new
 * bi-connected components appear when we omit a single terminal.  If
 * so, and if the resulting components form a single cycle, then we
 * can solve each sub-problem twice (once as a tree, and once as a
 * forest) and combine the sub-problem answers for the final answer.
 */

	static
	bool
perform_cycle_busting (

bitmap_t *		tmap,		/* IN - subset of terminals to use */
bitmap_t *		fset_mask,	/* IN - subset of full-sets to use */
struct cinfo *		cip,		/* IN - compatibility info */
bitmap_t *		smt_mask	/* OUT - SMT as a set of full-sets */
)
{
int			i;
int			n;
int			k;
int *			p1;
int *			p2;
int			ncomps;
dist_t			len;
dist_t			total;
dist_t			best_len;
int			best_forest;
bitmap_t *		bp1;
bitmap_t *		bp2;
bitmap_t		word;
bitmap_t *		new_fset_mask;
bool			failed;
int			failed_tree_count;
bitmap_t		new_tmap [N_TERM_BMAP];
int			members [2 * MAX_TERMINALS];
int *			comps [MAX_TERMINALS + 1];
bitmap_t *		trees [MAX_TERMINALS];
bitmap_t *		forests [MAX_TERMINALS];
dist_t			tlen [MAX_TERMINALS];
dist_t			flen [MAX_TERMINALS];
bool			tfailed [MAX_TERMINALS];
bool			ffailed [MAX_TERMINALS];

	n = 0;
	bp1 = fset_mask;
	for (i = 0; i < cip -> num_fset_masks; i++) {
		word = *bp1++;
		n += NBITSON (word);
	}
	if (n <= 16) {
		/* Problem small enough to brute-force... */
		failed = backtrack_search (tmap, fset_mask, cip, smt_mask);
		return (failed);
	}

	/* Find best components with 1 terminal omitted... */
	ncomps = more_connected_components (tmap,
					    fset_mask,
					    cip,
					    members,
					    comps);
	if (ncomps <= 1) {
		/* Cycle-busting doesn't work -- just search. */
		failed = backtrack_search (tmap, fset_mask, cip, smt_mask);
		return (failed);
	}

#if 1
	print_mask (" % Component: ", tmap, cip -> num_terms);
	(void) printf (" %% Busted into cyclic components:\n");
	for (k = 0; k < ncomps; k++) {
		(void) printf ("  %% cyclic component: ");
		p1 = comps [k];
		p2 = comps [k + 1];
		while (p1 < p2) {
			(void) printf (" %d", *p1++);
		}
		(void) printf ("\n");
	}
#endif

	new_fset_mask = NEWA (cip -> num_fset_masks, bitmap_t);

	/* Allocate bit-masks to hold the tree and forest	*/
	/* solution for each component.				*/
	for (k = 0; k < ncomps; k++) {
		trees [k]	= NEWA (cip -> num_fset_masks, bitmap_t);
		forests [k]	= NEWA (cip -> num_fset_masks, bitmap_t);
		tlen [k]	= INF_DISTANCE;
		flen [k]	= INF_DISTANCE;
		tfailed [k]	= TRUE;
		ffailed [k]	= TRUE;
	}

	failed_tree_count = 0;
	for (k = 0; k < ncomps; k++) {
		bp1 = &new_tmap [0];
		for (i = 0; i < N_TERM_BMAP; i++) {
			*bp1++ = 0;
		}
		p1 = comps [k];
		p2 = comps [k + 1];
		while (p1 < p2) {
			i = *p1++;
			SETBIT (new_tmap, i);
		}

		/* Compute proper subset of full-sets for this component. */
		compute_fset_subset (new_tmap, fset_mask, cip, new_fset_mask);

		/* Solve tree sub-problem... */
		if (Disable_Recursive_Decomposition) {
			failed = backtrack_search (new_tmap, new_fset_mask, cip, trees [k]);
		}
		else {
			failed = perform_cycle_busting (new_tmap,
							new_fset_mask,
							cip,
							trees [k]);
		}
		tfailed [k] = failed;
		if (failed) {
			if (++failed_tree_count > 1) {
				/* We can never permit more than 1 tree	*/
				/* sub-problem to fail.  We need tree	*/
				/* solutions for all but one of them!	*/
				break;
			}
		}
		tlen [k] = total_tree_length (trees [k], cip);

		/* Solve forest sub-problem. */
		failed = backtrack_search_forest (comps [k] [0],
						  comps [k] [1],
						  new_tmap,
						  new_fset_mask,
						  cip,
						  forests [k]);
		ffailed [k] = failed;
		flen [k] = total_tree_length (forests [k], cip);

		(void) printf ("  %% cyc-comp done: ");
		p1 = comps [k];
		p2 = comps [k+1];
		while (p1 < p2) {
			(void) printf (" %d", *p1++);
		}
		(void) printf ("\n");
		if (tfailed [k]) {
			(void) printf ("  %% Tree:  FAILED!\n");
		}
		else {
			print_tree ("Tree", trees [k], cip);
		}
		if (ffailed [k]) {
			(void) printf ("  %% Forest:  FAILED!\n");
		}
		else {
			print_tree ("Forest", forests [k], cip);
		}
		(void) printf ("  %% tlen=%lu  flen=%lu\n", tlen [k], flen [k]);
		if ((NOT tfailed [k]) AND (NOT ffailed [k]) AND
		    (flen [k] >= tlen [k])) {
			/* A *valid* forest should always be	*/
			/* shorter than the tree!  (But there	*/
			/* might be no valid way to build such	*/
			/* a forest...  This is indicated by	*/
			/* one of infinite distance...		*/
			fatal ("perform_cycle_busting: Bug 1.");
		}
	}

	if (failed_tree_count > 1) {
		failed = TRUE;
	}
	else {
		best_forest = -1;
		if (failed_tree_count > 0) {
			/* Must use the forest for the component whose	*/
			/* tree sub-problem failed!			*/
			for (k = 0; k < ncomps; k++) {
				if (tfailed [k] AND NOT ffailed [k]) {
					best_forest = k;
					break;
				}
			}
		}
		else {
			best_len = 0;
			for (k = 0; k < ncomps; k++) {
				if (tfailed [k] OR ffailed [k]) continue;
				/* Compute the advantage gained by using this	*/
				/* components forest, rather than its tree.	*/
				len = (tlen [k] - flen [k]);
				if ((best_forest < 0) OR (len > best_len)) {
					/* This component achieves greater	*/
					/* savings than the previous best!	*/
					best_len = len;
					best_forest = k;
				}
			}
		}

		failed = FALSE;
		if (best_forest < 0) {
			/* None of the components has a valid forest!	*/
			/* Therefore, no solution can be found!		*/
			failed = TRUE;
		}
		else {
			/* Final answer is best forest, plus all other trees. */
			bp1 = smt_mask;
			bp2 = forests [best_forest];
			for (i = 0; i < cip -> num_fset_masks; i++) {
				*bp1++ = *bp2++;
			}
			for (k = 0; k < ncomps; k++) {
				if (k EQ best_forest) continue;
				bp1 = smt_mask;
				bp2 = trees [k];
				for (i = 0; i < cip -> num_fset_masks; i++) {
					*bp1++ |= *bp2++;
				}
			}
		}
	}

	if (failed) {
		print_mask (" % Component FAILED! : ", tmap, cip -> num_terms);
	}
	else {
		print_mask (" % Component: ", tmap, cip -> num_terms);
		(void) printf ("  %% used forest #%d\n", best_forest);
		print_tree ("Resulting tree", smt_mask, cip);
	}

	for (k = 0; k < ncomps; k++) {
		(void) free ((char *) trees [k]);
		(void) free ((char *) forests [k]);
	}

	(void) free ((char *) new_fset_mask);

	return (failed);
}

/*
 * This routine prints out a given subset.
 */

	void
print_mask (

char *		msg,		/* IN - header message to display */
bitmap_t *	bp1,		/* IN - subset to display */
int		n		/* IN - number of bits in mask */
)
{
int		i;

	(void) printf ("%s", msg);
	for (i = 0; i < n; i++) {
		if (BITON (bp1, i)) {
			printf (" %d", i);
		}
	}
	printf ("\n");
}

/*
 * This routine displays a one-line trace of each full-set whose bit
 * is set in the given mask.
 */

	void
print_tree (

char *		msg,		/* IN - message. */
bitmap_t *	fset_mask,	/* IN - subset of full-sets to use */
struct cinfo *	cip		/* IN - compatibility info */
)
{
struct full_set *	fsp;
struct pset *		terms;
int			i;
int			j;
int			n;
struct point *		p;

	(void) printf ("  %% %s:\n", msg);
	for (i = 0; i < cip -> full_tree_count; i++) {
		if (NOT BITON (fset_mask, i)) continue;
		(void) printf ("   %%");
		fsp = cip -> full_trees [i];
		terms = fsp -> terminals;
		n = terms -> n;
		p = &(terms -> a [0]);
		for (j = 0; j < n; p++, j++) {
			(void) printf (" %d", p -> pnum);
		}
		(void) printf ("\n");
	}
}

/*
 * This routine determines whether or not the given sub-problem can be
 * broken into a simple cycle of components that would each be a
 * bi-connected component, except for the fact that they are all connected
 * into a simple cycle.
 *
 * We detect this by omitting a point and computing bi-connected components.
 * Because the input point set is already a single BCC, deleting the
 * appropriate point will break the ring and expose the other components as
 * BCC's.
 */

	static
	int
more_connected_components (

bitmap_t *		tmap,		/* IN - subset of terminals to use */
bitmap_t *		fset_mask,	/* IN - subset of full-sets to use */
struct cinfo *		cip,		/* IN - compatibility info */
int *			best_members,	/* OUT - members of best components */
int **			best_comps	/* OUT - start/end of best components */
)
{
int			i;
int			j;
int			n;
int *			ip;
int *			comps;
int			rowsize;
bitmap_t *		g;
bitmap_t *		newg;
int			omit;
int			best_ncomps;
int			best_largest;
int			largest_comp;
int *			memp;
int *			p1;
int *			p2;
bool			fixed;
struct bicon		bicon;
int			dfs [MAX_TERMINALS];
int			back [MAX_TERMINALS];
int			members [2 * MAX_TERMINALS];
int *			groups [MAX_TERMINALS + 1];

	n = cip -> num_terms;

	comps = NEWA (n * n, int);

	rowsize	= cip -> num_term_masks;
	g = construct_graph (tmap, fset_mask, cip);

	newg = NEWA (n * rowsize, bitmap_t);

	bicon.n			= n;
	bicon.rowsize		= rowsize;
	bicon.g			= newg;
	bicon.dfs_num		= &dfs [0];
	bicon.back		= &back [0];
	bicon.stack		= NEWA (2 * n * n, pnum_t);
	bicon.comps		= comps;

	best_ncomps = 0;
	best_largest = n + 1;

	for (omit = 0; omit < n; omit++) {
		if (NOT BITON (tmap, omit)) continue;
		omit_and_bicon (tmap, &bicon, omit, g);

		/* Put omitted point back into proper components, */
		/* determine all components and articulation points. */
		/* Make sure it satisfies all necessary constraints... */
		fixed = fix_comps (&bicon,
				   omit,
				   g,
				   tmap,
				   fset_mask,
				   cip,
				   &largest_comp,
				   members,
				   groups);
		if (NOT fixed) {
			/* Unable to use this decomposition... */
			continue;
		}
		/* See if best split so far... */
		if (largest_comp < best_largest) {
			best_largest = largest_comp;
			best_ncomps = bicon.ncomps;
			memp = best_members;
			for (i = 0; i < best_ncomps; i++) {
				/* Set start of this component... */
				best_comps [i] = memp;
				p1 = groups [i];
				p2 = groups [i + 1];
				while (p1 < p2) {
					*memp++ = *p1++;
				}
			}
			best_comps [i] = memp;	/* terminate final group. */
		}
	}

	free ((char *) bicon.stack);
	free ((char *) newg);
	free ((char *) g);
	free ((char *) comps);

	return (best_ncomps);
}

/*
 * This routine omits the given point and then computes the Bi-Connected
 * Components.
 */

	static
	void
omit_and_bicon (

bitmap_t *		tmap,		/* IN - subset of terminals to use */
struct bicon *		bcp,		/* IN - Bi-Connected Component info */
int			omit,		/* IN - point to omit. */
bitmap_t *		g		/* IN - graph before point omitted */
)
{
int		i;
int		j;
int		n;
int *		ip;
int		rowsize;
size_t		gsize;
bitmap_t *	newg;

	n = bcp -> n;
	rowsize	= bcp -> rowsize;
	gsize = n * rowsize * sizeof (bitmap_t);
	newg = bcp -> g;

	/* Make a copy of the graph with the one point removed. */
	memcpy (newg, g, gsize);
	for (i = 0; i < n; i++) {
		CLRBIT (&newg [i * rowsize], omit);
		CLRBIT (&newg [omit * rowsize], i);
	}
	ip = bcp -> comps;
	for (i = 0; i < n; i++) {
		for (j = 0; j < n; j++) {
			*ip++ = 0;
		}
	}
	ip = bcp -> dfs_num;
	for (i = 0; i < n; i++) {
		*ip++ = 0;
	}
	ip = bcp -> back;
	for (i = 0; i < n; i++) {
		*ip++ = 0;
	}
	bcp -> sp		= bcp -> stack;
	bcp -> counter		= 0;
	bcp -> ncomps		= 0;

	/* Compute Bi-Connected Components with point omitted... */
	for (i = 0; i < n; i++) {
		if ((i NE omit) AND BITON (tmap, i)) {
			/* Start BCC walk on valid non-omitted terminal. */
			bcc (bcp, i);
			break;
		}
	}
}

	static
	bool
fix_comps (

struct bicon *		bcp,		/* IN - BCC info */
int			omit,		/* IN - omitted point */
bitmap_t *		g,		/* IN - graph before omission */
bitmap_t *		tmap,		/* IN - subset of terminals to use */
bitmap_t *		fset_mask,	/* IN - subset of full-sets to use */
struct cinfo *		cip,		/* IN - compatibility info */
int *			largest_comp,	/* OUT - largest component */
int *			ip,		/* OUT - component members */
int **			groups		/* OUT - component begin/end ptrs */
)
{
int			i;
int			j;
int			k;
int			n;
int			rowsize;
int			ncomps;
int *			rowp;
bitmap_t *		g_omit_rowp;
bitmap_t *		bp1;
int			nmasks;
int			narts;
bitmap_t		arts [N_TERM_BMAP];
int			incomps [MAX_TERMINALS];
bitmap_t		comp_masks [MAX_TERMINALS * N_TERM_BMAP];

	n	= bcp -> n;
	rowsize	= bcp -> rowsize;
	ncomps	= bcp -> ncomps;

	nmasks = BMAP_ELTS (n);

	j = ncomps * nmasks;
	for (i = 0; i < j; i++) {
		comp_masks [i] = 0;
	}

	rowp = bcp -> comps;
	for (i = 0; i < n; i++) {
		for (j = i + 1; j < n; j++) {
			k = rowp [j];
			if (k > 0) {
				/* I and J are both members of component K. */
				--k;
				bp1 = &comp_masks [nmasks * k];
				SETBIT (bp1, i);
				SETBIT (bp1, j);
			}
		}
		rowp += n;
	}

	/* Put the omitted point back into the proper components. */
	g_omit_rowp = g + (omit * rowsize);
	for (i = 0; i < n; i++) {
		if (BITON (g_omit_rowp, i)) {
			/* The omitted point is a member of each component */
			/* that point I is a member of...		   */
			bp1 = comp_masks;
			for (k = 0; k < ncomps; k++) {
				if (BITON (bp1, i)) {
					SETBIT (bp1, omit);
				}
				bp1 += nmasks;
			}
		}
	}

	/* Determine how many components each point is a member of... */
	for (i = 0; i < n; i++) {
		incomps [i] = 0;
	}
	bp1 = comp_masks;
	for (k = 0; k < ncomps; k++) {
		for (i = 0; i < n; i++) {
			if (BITON (bp1, i)) {
				++incomps [i];
			}
		}
		bp1 += nmasks;
	}

	/* Compute set of articulation points... */
	for (i = 0; i < N_TERM_BMAP; i++) {
		arts [i] = 0;
	}
	for (i = 0; i < n; i++) {
		j = incomps [i];
		if (j > 2) {
			/* Articulation point I is a member of MORE than */
			/* two components -- too complicated...		 */
			return (FALSE);
		}
		if (j EQ 2) {
			SETBIT (arts, i);
		}
	}

	/* Produce output set of components. */
	*largest_comp = 0;
	bp1 = comp_masks;
	for (k = 0; k < ncomps; k++) {
		/* Set start of current component... */
		*groups++ = ip;
		/* Emit all articulations points for this component... */
		narts = 0;
		for (i = 0; i < n; i++) {
			if (BITON (bp1, i) AND BITON (arts, i)) {
				*ip++ = i;
				++narts;
			}
		}
		if (narts NE 2) {
			/* This component connects to MORE than 2 others! */
			/* Too hairy to bother with right now... */
			return (FALSE);
		}
		/* Generate the non-articulation points into component. */
		for (i = 0; i < n; i++) {
			if (BITON (bp1, i) AND NOT BITON (arts, i)) {
				*ip++ = i;
				++narts;
			}
		}
		if ((narts EQ 3) AND only_one_fake_full_set (bp1, fset_mask, cip)) {
			/* This component consists ENTIRELY of a fake	*/
			/* full-set.  We cannot consider this to be a	*/
			/* valid decomposition, because it would lead	*/
			/* to an infinite-recursion of cycle-busting.	*/
			return (FALSE);
		}
		if (narts > *largest_comp) {
			*largest_comp = narts;
		}
		bp1 += nmasks;
	}
	*groups = ip;		/* Terminate last component */

	for (i = 0; i < n; i++) {
		if (BITON (arts, i) AND BITON (g_omit_rowp, i)) {
			/* Edge exists between "omit" and one	*/
			/* of the other articulation points.	*/
			/* This indicates that a full-set of	*/
			/* size 3 has been split apart by the	*/
			/* removal of point "omit".  Thus omit	*/
			/* is not a true articulation point,	*/
			/* and we must punt this possibility.	*/
			return (FALSE);
		}
	}

	return (TRUE);
}

/*
 * This routine checks every full-set to see if it is contained entirely
 * within the given component (set of terminals).  We return TRUE if-and-
 * only-if there is exactly ONE such full-set, and it is a FAKE!
 */

	static
	bool
only_one_fake_full_set (

bitmap_t *		tmap,		/* IN - subset of terminals to use */
bitmap_t *		fset_mask,	/* IN - subset of full-sets to use */
struct cinfo *		cip		/* IN - compatibility info */
)
{
int			i;
int			j;
int			ntrees;
int			kmasks;
int			num_fake;
struct full_set *	fsp;
struct full_set *	p;
bitmap_t *		bp1;
bitmap_t *		bp2;

	ntrees	= cip -> full_tree_count;
	kmasks	= cip -> num_term_masks;

	num_fake = 0;
	for (i = 0; i < ntrees; i++) {
		if (NOT BITON (fset_mask, i)) continue;
		fsp = cip -> full_trees [i];
		bp1 = tmap;
		bp2 = &(fsp -> tmap [0]);
		j = 0;
		while (TRUE) {
			if (j >= kmasks) {
				if (fsp -> style NE FAKE) return (FALSE);
				++num_fake;
				if (num_fake > 1) return (FALSE);
				break;
			}
			if ((*bp1 & *bp2) NE *bp2) break;
			++j;
			++bp1;
			++bp2;
		}
	}

	/* Component contained exactly ONE full-set, which was a FAKE! */

	return (TRUE);
}

/*
 * This routine performs a backtrack search for a Steiner Minimal Forest
 * of the given point set.  The first two points in the set are the
 * articulation points that all of the rest of the points must connect
 * to in the shortest way possible.  However, there must be NO PATH within
 * the forest from one articulation point to the other.
 *
 * We accomplish this by creating a "fake" full-set containing the two
 * articulation points, and a new "fake" terminal.  Conceptually, this
 * full-set has zero length, and can intersect the "real" full-sets only
 * at the two articulation points.  Thus, we determine the relative
 * compatibility/incompatibility/disjointness of the "fake" full-set
 * with others solely by the number of terminals in common.  This full-set
 * MUST be present in every solution, because it provides the ONLY way to
 * connect the "fake" terminal.  Because this "fake" full-set connects the
 * two articulation points, no other path between them will be permitted.
 * Thus, the forest solution we seek is found by simply deleting the
 * "fake" full-set from the solution after a normal search for a "tree"
 * style of solution.
 */

	static
	bool
backtrack_search_forest (

int			art1,		/* IN - first articulation point */
int			art2,		/* IN - second articulation point */
bitmap_t *		tmap,		/* IN - subset of terminals to use */
bitmap_t *		fset_mask,	/* IN - subset of full-sets to use */
struct cinfo *		cip,		/* IN - compatibility info */
bitmap_t *		smt_mask	/* OUT - SMT as a set of full-sets */
)
{
int			i;
int			n;
int			new_ntrees;
int			new_nmasks;
bitmap_t *		bp1;
bitmap_t *		bp2;
bitmap_t		word;
struct point *		p1;
struct full_set *	fsp;
bitmap_t *		new_fset_mask;
bitmap_t *		new_smt_mask;
bitmap_t		new_tmap [N_TERM_BMAP];
struct full_set *	tmp;
bool			failed;
struct pset		newpts;
struct pset		fake_pts;
struct cinfo		new_cinfo;

	n = 0;
	bp1 = &new_tmap [0];
	for (i = 0; i < N_TERM_BMAP; i++) {
		*bp1++ = 0;
	}

	bp1 = tmap;
	bp2 = &new_tmap [0];
	for (i = 0; i < cip -> num_term_masks; i++) {
		word = *bp1++;
		*bp2++ = word;
		n += NBITSON (word);
	}

	if (n <= 2) {
		/* Just the two articulation points!  Forest is empty! */
		bp1 = smt_mask;
		for (i = 0; i < cip -> num_fset_masks; i++) {
			*bp1++ = 0;
		}
		return (FALSE);
	}

	/* Copy the input point set, and add a "fake" point to it. */
	COPY_PSET (&newpts, cip -> pts);

	n = newpts.n;

	SETBIT (new_tmap, n);

	p1 = &newpts.a [n];
	p1 -> x			= 0;
	p1 -> y			= 0;
	p1 -> pnum		= n;
	p1 -> other_pnum	= -1;	/* marks this point as "fake"... */
	newpts.n = n + 1;

	/* Construct "fake" full-set, containing both articulation	*/
	/* points and the new "fake" point.				*/

	fake_pts.n = 3;
	fake_pts.a [0]	= newpts.a [art1];
	fake_pts.a [1]	= newpts.a [art2];
	fake_pts.a [2]	= newpts.a [n];

	fsp = NEW (struct full_set);

	fsp -> next		= NULL;
	fsp -> style		= FAKE;
	fsp -> tree_num		= cip -> full_tree_count;
	fsp -> tree_len		= 0;
	fsp -> terminals	= save_pset (&fake_pts);
	fake_pts.n = 0;
	fsp -> steiners		= save_pset (&fake_pts);

	compute_tmap (fsp -> terminals, fsp -> tmap);

	/* Compute entirely new problem instance, including	*/
	/* new compatibility info.				*/
	for (i = cip -> full_tree_count - 1; i >= 0; i--) {
		tmp = copy_full_set (cip -> full_trees [i]);
		tmp -> next = fsp;
		fsp = tmp;
	}

	new_ntrees = cip -> full_tree_count + 1;
	new_nmasks = BMAP_ELTS (new_ntrees);

	new_fset_mask = NEWA (new_nmasks, bitmap_t);
	new_smt_mask = NEWA (new_nmasks, bitmap_t);

	/* Pre-zero last bit-mask element, in case new_nmasks > nmasks.	*/
	/* Whenever this happens, we have new_nmasks EQ nmasks + 1.	*/
	new_fset_mask [new_nmasks - 1] = 0;

	memcpy (new_fset_mask,
		fset_mask,
		cip -> num_fset_masks * sizeof (bitmap_t));
	SETBIT (new_fset_mask, cip -> full_tree_count);

	/* This is gross, disgusting, and extremely wasteful, but... */
	build_compatibility_data (&newpts,
				  new_tmap,
				  fsp,
				  new_fset_mask,
				  &new_cinfo);

#if 0
	failed = backtrack_search (new_tmap,
				   new_fset_mask,
				   &new_cinfo,
				   new_smt_mask);
#else
	failed = perform_cycle_busting (new_tmap,
					new_fset_mask,
					&new_cinfo,
					new_smt_mask);
#endif

	/* Remove the "fake" piece from the solution, and copy result	*/
	/* back to caller.						*/
	CLRBIT (new_smt_mask, cip -> full_tree_count);
	bp1 = smt_mask;
	bp2 = new_smt_mask;
	for (i = 0; i < cip -> num_fset_masks; i++) {
		*bp1++ = *bp2++;
	}

	/* Now clean up... */
	free_compatibility_data (&new_cinfo);
	(void) free ((char *) new_smt_mask);
	(void) free ((char *) new_fset_mask);
	while (fsp NE NULL) {
		tmp = fsp -> next;
		(void) free ((char *) fsp -> terminals);
		(void) free ((char *) fsp);
		fsp = tmp;
	}

	return (failed);
}

/*
 * This routine adds up the total length of the given set of full-trees.
 */

	dist_t
total_tree_length (

bitmap_t *		fset_mask,	/* IN - set of full-trees to tally */
struct cinfo *		cip		/* IN - compatibility info */
)
{
int		i;
dist_t		total;

	total = 0;
	for (i = 0; i < cip -> full_tree_count; i++) {
		if (BITON (fset_mask, i)) {
			total += cip -> full_trees [i] -> tree_len;
		}
	}

	return (total);
}
