/***********************************************************************

	File:	fine.c
	Rev:	a-1
	Date:	03/13/93

	Copyright (c) 1993, by The University of Virginia

************************************************************************

	Routines to perform fine-grained tests on particular
	subsets.

************************************************************************

	Modification Log:

	a-1:	03/13/93	dmw
		: Created.

************************************************************************/

#include "steiner.h"


/*
 * Global Routines
 */

void		make_second_topology (struct full_set *, struct full_set *);
bool		passes_fine_tests (struct pset *, struct full_set *);


/*
 * External References
 */

	/* none */


/*
 * Local Routines
 */

static bool		alt_0HRV (struct pset *);
static bool		alt_0VRH (struct pset *);
static bool		alt_1HRV (struct pset *);
static bool		alt_1VRH (struct pset *);
static bool		alternation (struct full_set *);
static int		full_topology_graph (struct full_set *,
					     struct pset *, struct edge *);
static bool		longest_edge_too_long (struct full_set *);
static bool		terminal_on_backbone (struct pset *, struct pset *);
static bool		terminal_too_close_to_segment (struct pset *,
						       struct full_set *);
static bool		within_dist_of_point_and_backbone (struct pset *,
							   struct point *,
							   dist_t,
							   struct point *,
							   struct point *);

/*
 * This routine performs the fine-grained tests on the given full-topology
 * to determine whether it can be a full set.
 */

	bool
passes_fine_tests (

struct pset *		pts,		/* IN - original point set. */
struct full_set *	fsp		/* IN - potential full-set to test. */
)
{
dist_t			smt_len;
dist_t			kr_len;
struct pset		term2;
struct pset		stein2;
struct full_set		fset2;

	fset2.terminals	= &term2;
	fset2.steiners	= &stein2;

	make_second_topology (fsp, &fset2);

	if (terminal_on_backbone (pts, fsp -> terminals)) return (FALSE);
	if (terminal_on_backbone (pts, &term2)) return (FALSE);

	if (terminal_too_close_to_segment (pts, fsp)) return (FALSE);
	if (terminal_too_close_to_segment (pts, &fset2)) return (FALSE);

	if (NOT alternation (fsp)) return (FALSE);
	if (NOT alternation (&fset2)) return (FALSE);

	if (longest_edge_too_long (fsp)) return (FALSE);
	if (longest_edge_too_long (&fset2)) return (FALSE);

	smt_len = min (fsp -> tree_len, fset2.tree_len);

	kr_len = kahng_robins_length (fsp -> terminals, smt_len);

	if (kr_len < smt_len) return (FALSE);

	return (TRUE);
}

/*
 * This routine returns TRUE if there exists a terminal that lies
 * precisely along the given backbone from point A to point B.  If
 * this is the case, we have an invalid backbone because it would
 * have to be split into at least two sub-backbones...
 */

	static
	bool
terminal_on_backbone (

struct pset *		pts,	/* IN - original point set. */
struct pset *		terms	/* IN - terminals of full-set. */
)
{
int			i;
int			n;
int			m;
struct point *		ap;
struct point *		bp;
struct point *		p1;

	n = terms -> n;
	ap = &(terms -> a [0]);
	bp = &(terms -> a [1]);
	m = pts -> n;

	p1 = &(pts -> a [0]);
	for (i = 0; i < m; p1++, i++) {
		if (STRICTLY_BETWEEN_X (ap, p1, bp) AND
		    (p1 -> y EQ bp -> y)) {
			return (TRUE);
		}
		if (STRICTLY_BETWEEN_Y (ap, p1, bp) AND
		    (p1 -> x EQ ap -> x)) {
			return (TRUE);
		}
		if ((p1 -> x EQ ap -> x) AND (p1 -> y EQ bp -> y) AND
		    (ap -> x NE bp -> x) AND (ap -> y NE bp -> y)) {
			return (TRUE);	/* p1 is at the corner! */
		}
	}
	if (n <= 4) {
		return (FALSE);
	}
	/* Additional tests for large full-sets... */
	p1 = &(terms -> a [2]);
	for (i = 2; i < n; p1++, i++) {
		if (p1 -> x EQ ap -> x) return (TRUE);
		if (p1 -> x EQ bp -> x) return (TRUE);
		if (p1 -> y EQ ap -> y) return (TRUE);
		if (p1 -> y EQ bp -> y) return (TRUE);
	}

	return (FALSE);
}

/*
 * This routine returns TRUE if there exists a terminal that is
 * closer to both point X and point Y than point X and Y are to each
 * other, where points X and Y refer to every point (terminal or steiner)
 * connected by a segment in the given full topology.  By Theorem 7,
 * this full topology cannot be a component of a Steiner Minimal Tree.
 */

	static
	bool
terminal_too_close_to_segment (

struct pset *		pts,	/* IN - original point set. */
struct full_set *	fsp	/* IN - full-set to test. */
)
{
int			i;
int			n;
struct point *		ap;
struct point *		bp;
struct point *		firsts;
struct point *		lasts;
struct point *		p1;
struct point *		p2;
struct pset *		terms;
struct pset *		steins;
dist_t			d;

	terms	= fsp -> terminals;
	steins	= fsp -> steiners;

	n = terms -> n;

	ap = &(terms -> a [0]);
	bp = &(terms -> a [1]);

	if (n <= 2) {
		if (is_point_close_to (pts, ap, bp)) {
			return (TRUE);
		}
		return (FALSE);
	}

	firsts	= &(steins -> a [0]);
	lasts	= &(steins -> a [n-3]);

	switch (fsp -> style) {
	case ZERO_H_REST_V:
	case ONE_H_REST_V:
	case ONLY_V:
	case CROSS:
		if (is_point_close_to (pts, ap, lasts)) return (TRUE);
		if (is_point_close_to (pts, bp, firsts)) return (TRUE);
		break;

	case ZERO_V_REST_H:
	case ONE_V_REST_H:
	case ONLY_H:
		if (is_point_close_to (pts, ap, firsts)) return (TRUE);
		if (is_point_close_to (pts, bp, lasts)) return (TRUE);
		break;

	default:
		fatal ("terminal_too_close_to_segment: Bug 1.");
		break;
	}

	n = steins -> n;
	p1 = &(terms -> a [2]);
	p2 = &(steins -> a [0]);
	for (i = 1; i < n; p1++, p2++, i++) {
		d = DISTANCE (p1, p2);
		if (within_dist_of_point_and_backbone (pts, p1, d, ap, bp)) {
			return (TRUE);
		}
		if (is_point_close_to (pts, p2, p2+1)) return (TRUE);
	}
	d = DISTANCE (p1, p2);
	if (within_dist_of_point_and_backbone (pts, p1, d, ap, bp)) {
		return (TRUE);
	}

	return (FALSE);
}

/*
 * This routine returns TRUE if-and-only-if there exists a terminal
 * Y that is less than DELTA distance away from both the backbone from
 * A to B, and from point X.
 */

	static
	bool
within_dist_of_point_and_backbone (

struct pset *		pts,	/* IN - the complete set of terminals. */
struct point *		xp,	/* IN - point X. */
dist_t			delta,	/* IN - the DELTA distance value. */
struct point *		ap,	/* IN - backbone point A. */
struct point *		bp	/* IN - backbone point B. */
)
{
int			i;
int			n;
struct point *		p1;

	n = pts -> n;
	p1 = &(pts -> a [0]);
	for (i = 0; i < n; p1++, i++) {
		if (p1 EQ xp) continue;
		if ((DISTANCE (p1, xp) < delta) AND
		    (backbone_distance (ap, bp, p1) < delta)) {
			return (TRUE);
		}
	}

	return (FALSE);
}

/*
 * This routine returns TRUE if-and-only-if the given set of points
 * form a backbone of the given style that properly alternates down
 * the "rest" leg of the backbone.
 */

	static
	bool
alternation (

struct full_set *	fsp		/* IN - full-set to test. */
)
{
bool			result;

	switch (fsp -> style) {
	case ZERO_H_REST_V:
		result = alt_0HRV (fsp -> terminals);
		break;

	case ONE_H_REST_V:
		result = alt_1HRV (fsp -> terminals);
		break;

	case ZERO_V_REST_H:
		result = alt_0VRH (fsp -> terminals);
		break;

	case ONE_V_REST_H:
		result = alt_1VRH (fsp -> terminals);
		break;

	case ZERO_H_ZERO_V:
		result = TRUE;		/* vacuously alternates! */
		break;

	case ONLY_H:
	case ONLY_V:
	case CROSS:
		/* These full sets always alternate properly	*/
		/* when generated.  Since make-second-topology	*/
		/* does nothing to these, they are always OK.	*/
		result = TRUE;
		break;

	default:
		fatal ("alternation: Bug 1.");
		result = FALSE;
	}

	return (result);
}

/*
 * This routine checks for proper alternation of points down the "rest"
 * leg of a backbone of the Zero Horizontal, Rest Vertical type.
 */

	static
	bool
alt_0HRV (

struct pset *		terms		/* IN - set of terminals. */
)
{
int			i;
int			n;
struct point *		ap;
struct point *		bp;
struct point *		p1;
enum side		last;
coord_t			prev_y;

	n = terms -> n;

	ap = &(terms -> a [0]);
	bp = &(terms -> a [1]);

	if (ap -> y <= bp -> y) {
		last = RIGHT;
		prev_y = bp -> y;
		p1 = &(terms -> a [2]);
		for (i = 2; i < n; p1++, i++) {
			/* Verify points properly descending... */
			if (p1 -> y >= prev_y) {
				return (FALSE);
			}
			if (BETWEEN_X (p1, ap, bp)) {
				if (last EQ LEFT) return (FALSE);
				last = LEFT;
			}
			else {
				if (last EQ RIGHT) return (FALSE);
				last = RIGHT;
			}
			prev_y = p1 -> y;
		}
	}
	else {
		last = RIGHT;
		prev_y = bp -> y;
		p1 = &(terms -> a [2]);
		for (i = 2; i < n; p1++, i++) {
			/* Verify points properly ascending... */
			if (p1 -> y <= prev_y) {
				return (FALSE);
			}
			if (BETWEEN_X (p1, ap, bp)) {
				if (last EQ LEFT) return (FALSE);
				last = LEFT;
			}
			else {
				if (last EQ RIGHT) return (FALSE);
				last = RIGHT;
			}
			prev_y = p1 -> y;
		}
	}

	return (TRUE);
}

/*
 * This routine checks for proper alternation of points down the "rest"
 * leg of a backbone of the One Horizontal, Rest Vertical type.
 */

	static
	bool
alt_1HRV (

struct pset *		terms		/* IN - set of terminals. */
)
{
int			i;
int			n;
struct point *		ap;
struct point *		bp;
struct point *		p1;
enum side		last;
coord_t			prev_y;

	n = terms -> n;

	ap = &(terms -> a [0]);
	bp = &(terms -> a [1]);

	p1 = &(terms -> a [2]);
	if (NOT BETWEEN_Y (p1, bp, ap)) {
		return (FALSE);
	}

	if (NOT STRICTLY_BETWEEN_X (ap, p1, bp)) {
		return (FALSE);
	}

	if (ap -> y <= bp -> y) {
		last = RIGHT;
		prev_y = bp -> y;
		p1 = &(terms -> a [3]);
		for (i = 3; i < n; p1++, i++) {
			/* Verify points properly descending... */
			if (p1 -> y >= prev_y) {
				return (FALSE);
			}
			if (BETWEEN_X (p1, ap, bp)) {
				if (last EQ LEFT) return (FALSE);
				last = LEFT;
			}
			else {
				if (last EQ RIGHT) return (FALSE);
				last = RIGHT;
			}
			prev_y = p1 -> y;
		}
	}
	else {
		last = RIGHT;
		prev_y = bp -> y;
		p1 = &(terms -> a [3]);
		for (i = 3; i < n; p1++, i++) {
			/* Verify points properly ascending... */
			if (p1 -> y <= prev_y) {
				return (FALSE);
			}
			if (BETWEEN_X (p1, ap, bp)) {
				if (last EQ LEFT) return (FALSE);
				last = LEFT;
			}
			else {
				if (last EQ RIGHT) return (FALSE);
				last = RIGHT;
			}
			prev_y = p1 -> y;
		}
	}

	return (TRUE);
}

/*
 * This routine checks for proper alternation of points down the "rest"
 * leg of a backbone of the Zero Vertical, Rest Horizontal type.
 */

	static
	bool
alt_0VRH (

struct pset *		terms		/* IN - set of terminals. */
)
{
int			i;
int			n;
struct point *		ap;
struct point *		bp;
struct point *		p1;
enum side		last;
coord_t			prev_x;

	n = terms -> n;

	ap = &(terms -> a [0]);
	bp = &(terms -> a [1]);

	if (ap -> x <= bp -> x) {
		last = RIGHT;
		prev_x = ap -> x;
		p1 = &(terms -> a [2]);
		for (i = 2; i < n; p1++, i++) {
			/* Verify points properly ascending... */
			if (p1 -> x <= prev_x) {
				return (FALSE);
			}
			if (BETWEEN_Y (p1, bp, ap)) {
				if (last EQ LEFT) return (FALSE);
				last = LEFT;
			}
			else {
				if (last EQ RIGHT) return (FALSE);
				last = RIGHT;
			}
			prev_x = p1 -> x;
		}
	}
	else {
		last = RIGHT;
		prev_x = ap -> x;
		p1 = &(terms -> a [2]);
		for (i = 2; i < n; p1++, i++) {
			/* Verify points properly descending... */
			if (p1 -> x >= prev_x) {
				return (FALSE);
			}
			if (BETWEEN_Y (p1, bp, ap)) {
				if (last EQ LEFT) return (FALSE);
				last = LEFT;
			}
			else {
				if (last EQ RIGHT) return (FALSE);
				last = RIGHT;
			}
			prev_x = p1 -> x;
		}
	}

	return (TRUE);
}

/*
 * This routine checks for proper alternation of points down the "rest"
 * leg of a backbone of the One Vertical, Rest Horizontal type.
 */

	static
	bool
alt_1VRH (

struct pset *		terms		/* IN - set of terminals. */
)
{
int			i;
int			n;
struct point *		ap;
struct point *		bp;
struct point *		p1;
enum side		last;
coord_t			prev_x;

	n = terms -> n;

	ap = &(terms -> a [0]);
	bp = &(terms -> a [1]);

	p1 = &(terms -> a [2]);
	if (NOT BETWEEN_X (p1, ap, bp)) {
		return (FALSE);
	}

	if (NOT STRICTLY_BETWEEN_Y (ap, p1, bp)) {
		return (FALSE);
	}

	if (ap -> x <= bp -> x) {
		last = RIGHT;
		prev_x = ap -> x;
		p1 = &(terms -> a [3]);
		for (i = 3; i < n; p1++, i++) {
			/* Verify points properly ascending... */
			if (p1 -> x <= prev_x) {
				return (FALSE);
			}
			if (BETWEEN_Y (p1, bp, ap)) {
				if (last EQ LEFT) return (FALSE);
				last = LEFT;
			}
			else {
				if (last EQ RIGHT) return (FALSE);
				last = RIGHT;
			}
			prev_x = p1 -> x;
		}
	}
	else {
		last = RIGHT;
		prev_x = ap -> x;
		p1 = &(terms -> a [3]);
		for (i = 3; i < n; p1++, i++) {
			/* Verify points properly descending... */
			if (p1 -> x >= prev_x) {
				return (FALSE);
			}
			if (BETWEEN_Y (p1, bp, ap)) {
				if (last EQ LEFT) return (FALSE);
				last = LEFT;
			}
			else {
				if (last EQ RIGHT) return (FALSE);
				last = RIGHT;
			}
			prev_x = p1 -> x;
		}
	}

	return (TRUE);
}

/*
 * This routine forms the Full Topology tree of the given style for the
 * given set of terminals and Steiner points.  It then determines (for
 * each pair of terminals X and Y) the longest edge along the path from
 * X to Y.  If this longest edge exceeds the same quantity for the
 * Minimum Spanning Tree of the original terminal set, then this cannot
 * be a Full Topolgy of any Steiner Minimal Tree.
 */

	static
	bool
longest_edge_too_long (

struct full_set *	fsp		/* IN - potential full-set to test. */
)
{
int			nedges;
int			n;
struct pset		comb;
struct edge		edge [MAX_POINTS - 1];

	/* Build edge-list of the full topology. */
	nedges = full_topology_graph (fsp, &comb, &edge [0]);

	n = fsp -> terminals -> n;

	if (full_topology_edge_too_long (n, &comb, edge, nedges)) {
		return (TRUE);
	}

	return (FALSE);
}

/*
 * This routine builds the graph for the given full-topology.  The
 * resulting graph is represented as an edge-list.  The output point
 * set has its "other_pnum" members set so as to refer to the original
 * point set (at least for the terminals, which occur first).
 *
 * As a convenient side-effect, it computes the length of the full-topology
 * and records it in the "tree_len" member.
 */

	static
	int
full_topology_graph (

struct full_set *	fsp,	/* IN - full-topology to build graph for. */
struct pset *		comb,	/* OUT - combined terminals/steiners. */
struct edge *		edges	/* OUT - graph edge list. */
)
{
int			i;
int			n;
struct point *		ap;
struct point *		bp;
struct point *		p1;
struct point *		p2;
struct edge *		ep;
pnum_t			prev;
struct pset *		terms;
struct pset *		steins;
dist_t			d;
dist_t			total;

	terms	= fsp -> terminals;
	steins	= fsp -> steiners;

	n = terms -> n;

	p1 = &(terms -> a [0]);
	p2 = &(comb -> a [0]);
	for (i = 0; i < n; p1++, p2++, i++) {
		*p2 = *p1;
		p2 -> other_pnum = p2 -> pnum;
		p2 -> pnum = ((pnum_t) i);
	}
	p1 = &(steins -> a [0]);
	for (i = 0; i < n - 2; p1++, p2++, i++) {
		*p2 = *p1;
		p2 -> pnum = n + i;
		p2 -> other_pnum = 0;
	}
	comb -> n = (n + n - 2);

	ap = &(comb -> a [0]);
	bp = &(comb -> a [1]);

	ep = edges;

	if (n <= 2) {
		ep -> len	= DISTANCE (ap, bp);
		ep -> p1	= 0;
		ep -> p2	= 1;
		fsp -> tree_len	= ep -> len;
		return (1);
	}

	switch (fsp -> style) {
	case ZERO_H_REST_V:
	case ONE_H_REST_V:
	case ONLY_V:
	case CROSS:
		ep -> p1	= 0;
		prev		= 1;
		break;

	case ZERO_V_REST_H:
	case ONE_V_REST_H:
	case ONLY_H:
		ep -> p1	= 1;
		prev		= 0;
		break;

	default:
		fatal ("full_topology_graph: Bug 1.");
		break;
	}
	ep -> p2 = n+n-3;
	++ep;
	for (i = 0; i < n-2; i++) {
		ep -> p1 = prev;
		ep -> p2 = n + i;
		++ep;
		ep -> p1 = n + i;
		ep -> p2 = i + 2;
		++ep;
		prev = n + i;
	}
	total = 0;
	while (edges < ep) {
		d = DISTANCE (&(comb -> a [edges -> p1]),
			      &(comb -> a [edges -> p2]));
		total += d;
		edges -> len = d;
		++edges;
	}
	fsp -> tree_len = total;

	return (n + n - 3);
}

/*
 * This routine performs the "corner flip" transform on the given full
 * topology, yielding a second full topology.
 */

	void
make_second_topology (

struct full_set *	ifsp,	/* IN - input full topology. */
struct full_set *	ofsp	/* OUT - output full topology. */
)
{
int			i;
int			n;
struct point *		p1;
struct point *		p2;
struct point *		endp;
coord_t			st_val;		/* New steiner point coord value. */
struct point *		iterms;		/* input terminals. */
struct point *		isteins;	/* input steiner points. */
struct point *		oterms;		/* output terminals. */
struct point *		osteins;	/* output steiner points. */

	iterms	= &(ifsp -> terminals -> a [0]);
	isteins	= &(ifsp -> steiners -> a [0]);
	oterms	= &(ofsp -> terminals -> a [0]);
	osteins	= &(ofsp -> steiners -> a [0]);

	n = ifsp -> terminals -> n;

	ofsp -> terminals -> n	= n;
	ofsp -> steiners -> n	= (n - 2);

	switch (ifsp -> style) {
	case ZERO_H_REST_V:
		if ((n % 2) EQ 0) {
			ofsp -> style = ZERO_H_REST_V;
			oterms [0]	= iterms [1];
			oterms [1]	= iterms [0];
			p2	= &oterms [2];
			endp	= &oterms [n];
			p1	= &iterms [n];
			while (p2 < endp) {
				*p2++ = *--p1;
			}
			p2	= &osteins [0];
			endp	= &osteins [n-2];
			p1	= &isteins [n-2];
			st_val	= oterms [0].x;
			i = 0;
			while (p2 < endp) {
				--p1;
				p2 -> x			= st_val;
				p2 -> y			= p1 -> y;
				p2 -> pnum		= ((pnum_t) i);
				p2 -> other_pnum	= p1 -> other_pnum;
				++p2;
				++i;
			}
		}
		else {
			ofsp -> style = ONE_H_REST_V;
			oterms [0]	= iterms [1];
			oterms [1]	= iterms [n-1];
			oterms [2]	= iterms [0];
			p2	= &oterms [3];
			endp	= &oterms [n];
			p1	= &iterms [n-1];
			while (p2 < endp) {
				*p2++ = *--p1;
			}
			osteins [0]	= isteins [n-3];
			osteins [0].pnum = 0;
			p2	= &osteins [1];
			endp	= &osteins [n-2];
			p1	= &isteins [n-3];
			st_val	= oterms [0].x;
			i = 1;
			while (p2 < endp) {
				--p1;
				p2 -> x			= st_val;
				p2 -> y			= p1 -> y;
				p2 -> pnum		= ((pnum_t) i);
				p2 -> other_pnum	= p1 -> other_pnum;
				++p2;
				++i;
			}
		}
		break;

	case ONE_H_REST_V:
		if ((n % 2) EQ 0) {
			ofsp -> style = ONE_H_REST_V;
			oterms [0]	= iterms [2];
			oterms [1]	= iterms [n-1];
			oterms [2]	= iterms [0];
			oterms [n-1]	= iterms [1];
			p2	= &oterms [3];
			endp	= &oterms [n-1];
			p1	= &iterms [n-1];
			while (p2 < endp) {
				*p2++ = *--p1;
			}
			osteins [0]	= isteins [n-3];
			osteins [0].pnum = 0;
			p2	= &osteins [1];
			endp	= &osteins [n-2];
			p1	= &isteins [n-3];
			st_val	= oterms [0].x;
			i = 1;
			while (p2 < endp) {
				--p1;
				p2 -> x			= st_val;
				p2 -> y			= p1 -> y;
				p2 -> pnum		= ((pnum_t) i);
				p2 -> other_pnum	= p1 -> other_pnum;
				++p2;
				++i;
			}
		}
		else {
			ofsp -> style = ZERO_H_REST_V;
			oterms [0]	= iterms [2];
			oterms [1]	= iterms [0];
			oterms [2]	= iterms [n-1];
			oterms [n-1]	= iterms [1];
			p2	= &oterms [3];
			endp	= &oterms [n-1];
			p1	= &iterms [n-1];
			while (p2 < endp) {
				*p2++ = *--p1;
			}
			p2	= &osteins [0];
			endp	= &osteins [n-2];
			p1	= &isteins [n-2];
			st_val	= oterms [0].x;
			i = 0;
			while (p2 < endp) {
				--p1;
				p2 -> x			= st_val;
				p2 -> y			= p1 -> y;
				p2 -> pnum		= ((pnum_t) i);
				p2 -> other_pnum	= p1 -> other_pnum;
				++p2;
				++i;
			}
		}
		break;

	case ZERO_V_REST_H:
		if ((n % 2) EQ 0) {
			ofsp -> style = ZERO_V_REST_H;
			oterms [0]	= iterms [1];
			oterms [1]	= iterms [0];
			p2	= &oterms [2];
			endp	= &oterms [n];
			p1	= &iterms [n];
			while (p2 < endp) {
				*p2++ = *--p1;
			}
			p2	= &osteins [0];
			endp	= &osteins [n-2];
			p1	= &isteins [n-2];
			st_val	= oterms [1].y;
			i = 0;
			while (p2 < endp) {
				--p1;
				p2 -> x			= p1 -> x;
				p2 -> y			= st_val;
				p2 -> pnum		= ((pnum_t) i);
				p2 -> other_pnum	= p1 -> other_pnum;
				++p2;
				++i;
			}
		}
		else {
			ofsp -> style = ONE_V_REST_H;
			oterms [0]	= iterms [n-1];
			oterms [1]	= iterms [0];
			oterms [2]	= iterms [1];
			p2	= &oterms [3];
			endp	= &oterms [n];
			p1	= &iterms [n-1];
			while (p2 < endp) {
				*p2++ = *--p1;
			}
			osteins [0]	= isteins [n-3];
			osteins [0].pnum = 0;
			p2	= &osteins [1];
			endp	= &osteins [n-2];
			p1	= &isteins [n-3];
			st_val	= oterms [1].y;
			i = 1;
			while (p2 < endp) {
				--p1;
				p2 -> x			= p1 -> x;
				p2 -> y			= st_val;
				p2 -> pnum		= ((pnum_t) i);
				p2 -> other_pnum	= p1 -> other_pnum;
				++p2;
				++i;
			}
		}
		break;

	case ONE_V_REST_H:
		if ((n % 2) EQ 0) {
			ofsp -> style = ONE_V_REST_H;
			oterms [0]	= iterms [n-1];
			oterms [1]	= iterms [2];
			oterms [2]	= iterms [1];
			oterms [n-1]	= iterms [0];
			p2	= &oterms [3];
			endp	= &oterms [n-1];
			p1	= &iterms [n-1];
			while (p2 < endp) {
				*p2++ = *--p1;
			}
			osteins [0]	= isteins [n-3];
			osteins [0].pnum = 0;
			p2	= &osteins [1];
			endp	= &osteins [n-2];
			p1	= &isteins [n-3];
			st_val	= oterms [1].y;
			i = 1;
			while (p2 < endp) {
				--p1;
				p2 -> x			= p1 -> x;
				p2 -> y			= st_val;
				p2 -> pnum		= ((pnum_t) i);
				p2 -> other_pnum	= p1 -> other_pnum;
				++p2;
				++i;
			}
		}
		else {
			ofsp -> style = ZERO_V_REST_H;
			oterms [0]	= iterms [1];
			oterms [1]	= iterms [2];
			oterms [n-1]	= iterms [0];
			p2	= &oterms [2];
			endp	= &oterms [n-1];
			p1	= &iterms [n];
			while (p2 < endp) {
				*p2++ = *--p1;
			}
			osteins [n-3]	= isteins [0];
			osteins [n-3].pnum = n-3;
			p2	= &osteins [0];
			endp	= &osteins [n-3];
			p1	= &isteins [n-2];
			st_val	= oterms [1].y;
			i = 0;
			while (p2 < endp) {
				--p1;
				p2 -> x			= p1 -> x;
				p2 -> y			= st_val;
				p2 -> pnum		= ((pnum_t) i);
				p2 -> other_pnum	= p1 -> other_pnum;
				++p2;
				++i;
			}
		}
		break;

	case ZERO_H_ZERO_V:
		if (n NE 2) {
			fatal ("make_second_topology: Bug 1.");
		}
		ofsp -> style = ZERO_H_ZERO_V;
		oterms [0]	= iterms [1];
		oterms [1]	= iterms [0];
		break;

	case ONLY_H:
	case ONLY_V:
	case CROSS:
		/* These types have no flips or slides, so they	*/
		/* don't change!				*/
		COPY_PSET (ofsp -> terminals, ifsp -> terminals);
		COPY_PSET (ofsp -> steiners,  ifsp -> steiners);
		ofsp -> style = ifsp -> style;
		break;

	default:
		fatal ("make_second_topology: Bug 2.");
		break;
	}

	/* Initialize other members of the full-set structure. */
	ofsp -> next		= NULL;
	ofsp -> tree_num	= 0;
	ofsp -> tree_len	= 0;
	(void) memcpy (ofsp -> tmap, ifsp -> tmap, sizeof (ofsp -> tmap));
}
