/***********************************************************************

	File:	gensets.c
	Rev:	a-1
	Date:	03/06/93

	Copyright (c) 1993, by The University of Virginia

************************************************************************

	Routines to recursively enumerate each potential full set.

************************************************************************

	Modification Log:

	a-1:	03/06/93	dmw
		: Created.

************************************************************************/

#include "steiner.h"


/*
 * Global Routines
 */

struct full_set *	generate_all_full_sets (struct pset *,
						bitmap_t *,
						int *);
struct full_set *	generate_duplicate_full_sets (struct pset *,
						      bitmap_t *);
struct pset *		save_pset (struct pset *);


/*
 * External References
 */

	/* none */

/*
 * Local Types
 */

	/* A structure to succinctly package data that must be	*/
	/* available at each level of the recursion.		*/

struct ginfo {
	struct pset *		pts;		/* original point set. */
	struct full_set *	list;		/* root of all full sets */
	struct full_set **	hookp;
	int32u			ccount;		/* # of coarse sets */
	int32u			fcount;		/* # of fine sets */
	struct point *		ap;		/* backbone point A */
	struct point *		bp;		/* backbone point B */
	enum backbone		style;		/* style of backbone. */
	struct pset		extras;		/* pts for oneXrestY. */
	int			lbound;		/* min "rest" points. */
	struct pset		gen;		/* pts of set being gen'd */
	struct pset		stn;		/* steiner points. */
	bitmap_t *		tmap;		/* terminals to consider */
};

struct pair {
	struct point	a;
	struct point	b;
};

/*
 * Local Routines
 */

static void		check_potential_full_set (struct ginfo *);
static void		horiz_inc (struct ginfo *, int, struct pset *);
static void		gen_E (struct ginfo *, int, struct pset *);
static void		gen_N (struct ginfo *, int, struct pset *);
static void		gen_NE (struct ginfo *, int, struct pset *);
static void		gen_SE (struct ginfo *, int, struct pset *);
static void		generate_all_cross_full_sets (struct ginfo *,
						      int,
						      int,
						      struct pair *,
						      struct pair *,
						      int *);
static void		generate_all_horiz_full_sets (struct ginfo *,
						      int,
						      struct pair *,
						      int *);
static void		generate_all_vert_full_sets (struct ginfo *,
						     int,
						     struct pair *,
						     int *);
static void		generate_horiz_vert_and_cross (struct ginfo *, int *);
static int		list_all_horiz_pairs (struct ginfo *, struct pair *);
static int		list_all_vert_pairs (struct ginfo *, struct pair *);
static void		recurse_horz_inc (struct ginfo *,
					  struct point *,
					  struct point *,
					  int,
					  struct point **,
					  int);
static void		recurse_vert_dec (struct ginfo *,
					  struct point *,
					  struct point *,
					  int,
					  struct point **,
					  int);
static void		recurse_vert_inc (struct ginfo *,
					  struct point *,
					  struct point *,
					  int,
					  struct point **,
					  int);
static void		vert_dec (struct ginfo *, int, struct pset *);
static void		vert_inc (struct ginfo *, int, struct pset *);


/*
 * Local Macros
 */

#define	BOUND(lb, n, pnt1, pnt2)			\
		((lb) [(pnt1).pnum * (n) + (pnt2).pnum])

/*
 * This routine recursively generates all Full Sets.
 */

	struct full_set *
generate_all_full_sets (

struct pset *		pts,		/* IN - the point set. */
bitmap_t *		tmap,		/* IN - subset of terminals to use */
int *			lb		/* IN - backbone lower bounds */
)
{
int			i;
int			j;
int			n;
struct point *		p1;
struct point *		p2;
struct ginfo		g;
struct pset		valid_pts;

	g.pts		= pts;
	g.list		= NULL;
	g.hookp		= &g.list;
	g.ccount	= 0;
	g.fcount	= 0;
	g.tmap		= tmap;

	n = pts -> n;

	p1 = &g.stn.a [0];
	for (i = 0; i < n; p1++, i++) {
		p1 -> x		= 0;
		p1 -> y		= 0;
		p1 -> pnum	= ((pnum_t) i);
		p1 -> other_pnum = 0;
	}

	generate_horiz_vert_and_cross (&g, lb);

	p1 = &(pts -> a [0]);
	for (i = 0; i < n; p1++, i++) {
		if (NOT BITON (tmap, i)) continue;
		g.ap = p1;
		p2 = &(pts -> a [0]);
		for (j = 0; j < n; p2++, lb++, j++) {
			if (*lb < 0) continue;
			if (NOT BITON (tmap, j)) continue;

			if ((p1 -> x <= p2 -> x) AND (p1 -> y <= p2 -> y)) {
				g.bp		= p2;
				list_valid_points (pts,
						   p1,
						   p2,
						   tmap,
						   &valid_pts);
				gen_NE (&g, *lb - 2, &valid_pts);
			}
			else if (p1 -> x < p2 -> x) {
				g.bp		= p2;
				list_valid_points (pts,
						   p1,
						   p2,
						   tmap,
						   &valid_pts);
				gen_SE (&g, *lb - 2, &valid_pts);
			}
			else {
				/* Other orientations are skipped... */
			}
		}
	}

	(void) printf ("  %%  ff = %lu\n", g.ccount);
	(void) printf ("  %%  rf = %lu\n", g.fcount);

	return (g.list);
}

/*
 * This routine finds all pairs of points whose coordinates are exactly
 * identical.  This is a degeneracy that causes successive algorithms
 * much heartburn, so we take care of them immediately by keeping only
 * the earliest (lowest pnum) point, and marking the others as duplicates.
 * For each ommitted (duplicate) point, we generate a 2-terminal full-set
 * (of zero-length) containing the first point and the duplicate.
 */

	struct full_set *
generate_duplicate_full_sets (

struct pset *		inpts,	/* IN - original terminal set. */
bitmap_t *		tmap	/* IN/OUT - set of terminals to consider. */
)
{
int			i;
int			n;
struct point *		p1;
struct point *		p2;
struct point *		endp;
struct point *		dup_end;
struct point *		lowest;
struct full_set *	root;
struct full_set **	hookp;
struct full_set *	fsp;
struct pset		pts;
struct pset		gen;
struct pset		stn;

	root	= NULL;
	hookp	= &root;

	/* Sort a COPY of the points on both X and Y keys */
	copy_terminal_subset (&pts, inpts, tmap);
	sort_xy_inc (&pts);

	n = pts.n;

	p1 = &pts.a [0];
	endp = p1 + (n - 1);

	while (p1 < endp) {
		if ((p1 [0].x NE p1 [1].x) OR (p1 [0].y NE p1 [1].y)) {
			/* Not duplicates -- skip to next */
			++p1;
			continue;
		}

		for (dup_end = p1 + 1; dup_end < endp; dup_end++) {
			if (p1 [0].x NE dup_end [1].x) break;
			if (p1 [0].y NE dup_end [1].y) break;
		}
		++dup_end;
		/* dup_end now points to the first non-equal terminal */

		lowest = p1;
		for (p2 = p1 + 1; p2 < dup_end; p2++) {
			if (p2 -> pnum < lowest -> pnum) {
				lowest = p2;
			}
		}

		/* Make one full-set for each extra occurrence... */
		gen.n		= 2;
		gen.a [0]	= *lowest;
		stn.n		= 0;
		for (p2 = p1; p2 < dup_end; p2++) {
			if (p2 EQ lowest) continue;
			gen.a [1]	= *p2;

			/* Never need to consider this terminal again! */
			CLRBIT (tmap, p2 -> pnum);

			fsp = new (sizeof (*fsp));
			fsp -> next		= NULL;
			fsp -> style		= ZERO_H_ZERO_V;
			fsp -> tree_num		= 0;
			fsp -> tree_len		= 0;
			fsp -> terminals	= save_pset (&gen);
			fsp -> steiners		= save_pset (&stn);

			*hookp = fsp;
			hookp = &(fsp -> next);
		}

		/* Skip whole group of coincident points and continue */
		p1 = dup_end;
	}

	return (root);
}

/*
 * This routine generates all degenerate full-sets that have backbones
 * that are either precisely horizontal, precisely vertical.  It also
 * generates backbones for the ultimate degeneracies -- crosses.
 */

	static
	void
generate_horiz_vert_and_cross (

struct ginfo *		gp,	/* IN/OUT - recursion info for backbone. */
int *			lb	/* IN - backbone lower bounds */
)
{
int			nhoriz;
int			nvert;
struct pair		hpairs [MAX_TERMINALS - 1];
struct pair		vpairs [MAX_TERMINALS - 1];

	nhoriz	= list_all_horiz_pairs (gp, &hpairs [0]);
	nvert	= list_all_vert_pairs  (gp, &vpairs [0]);

	if (nhoriz > 0) {
		if (nvert > 0) {
			generate_all_cross_full_sets (gp,
						      nhoriz,
						      nvert,
						      &hpairs [0],
						      &vpairs [0],
						      lb);
		}
		generate_all_horiz_full_sets (gp, nhoriz, &hpairs [0], lb);
	}

	if (nvert > 0) {
		generate_all_vert_full_sets (gp, nvert, &vpairs [0], lb);
	}
}

/*
 * This routine computes a list of all pairs of terminals that represent
 * backbones that are precisely horizontal.  The generated backbones will
 * have NO TERMINALS anywhere along them except the two (unique) terminals
 * at the endpoints.  Because of this constraint, there can be at most
 * N-1 such backbones given N original terminals.
 */

	static
	int
list_all_horiz_pairs (

struct ginfo *		gp,	/* IN - recursion info for backbone */
struct pair *		pairs	/* OUT - pairs of horizontal points */
)
{
int			i;
int			n;
int			npairs;
struct point *		p1;
struct point *		endp;
struct pset		pts;

	/* Get a copy of all relevent terminals. */
	copy_terminal_subset (&pts, gp -> pts, gp -> tmap);

	/* Sort them by increasing Y, and then increasing X. */
	sort_yx_inc (&pts);

	n = pts.n;
	p1 = &(pts.a [1]);
	endp = p1 + n;
	npairs = 0;

	for (; p1 < endp; p1++) {
		if (p1 [-1].y NE p1 [0].y) continue;
		/* These two points are horizontal.  Because of the	*/
		/* dual-key sort, and no duplicates, we know that	*/
		/* p1 [-1].x < p1 [0].x.  This is a horizontal pair!	*/
		pairs -> a	= p1 [-1];
		pairs -> b	= p1 [0];
		++npairs;
	}

	return (npairs);
}

/*
 * This routine computes a list of all pairs of terminals that represent
 * backbones that are precisely vertical.  The generated backbones will
 * have NO TERMINALS anywhere along them except the two (unique) terminals
 * at the endpoints.  Because of this constraint, there can be at most
 * N-1 such backbones given N original terminals.
 */

	static
	int
list_all_vert_pairs (

struct ginfo *		gp,	/* IN - recursion info for backbone */
struct pair *		pairs	/* OUT - pairs of vertical points */
)
{
int			i;
int			n;
int			npairs;
struct point *		p1;
struct point *		endp;
struct pset		pts;

	/* Get a copy of all relevent terminals. */
	copy_terminal_subset (&pts, gp -> pts, gp -> tmap);

	/* Sort them by increasing X, and then increasing Y. */
	sort_xy_inc (&pts);

	n = pts.n;
	p1 = &(pts.a [1]);
	endp = p1 + n;
	npairs = 0;

	for (; p1 < endp; p1++) {
		if (p1 [-1].x NE p1 [0].x) continue;
		/* These two points are vertical.  Because of the	*/
		/* dual-key sort, and no duplicates, we know that	*/
		/* p1 [-1].y < p1 [0].y.  This is a vertical pair!	*/
		pairs -> a	= p1 [-1];
		pairs -> b	= p1 [0];
		++npairs;
	}

	return (npairs);
}

/*
 * This routine generates all full-sets that are crosses.  We use as input
 * the set of all horizontal and vertical backbones.
 */

	static
	void
generate_all_cross_full_sets (

struct ginfo *		gp,	/* IN/OUT - generation info */
int			nhor,	/* IN - number of horizontal backbones */
int			nver,	/* IN - number of vertical backbones */
struct pair *		hpair,	/* IN - horizontal backbones */
struct pair *		vpair,	/* IN - vertical backbones */
int *			lb	/* IN - backbone lower bounds */
)
{
int			i;
int			j;
int			n;
int			bnd;
struct pair *		hp;
struct pair *		vp;
struct full_set *	fsp;
struct point		st;
struct full_set		fset;
struct pset		terms;
struct pset		steins;

#define	SKIP_TEST(lb, n, pnt1, pnt2)			\
	{ int bnd;					\
		bnd = BOUND (lb, n, pnt1, pnt2);	\
		if ((bnd < 0) OR (bnd > 4)) continue;	\
	}
#define	SKIP_IF_BAD_BOUNDS(lb, n, pnt1, pnt2)		\
	{						\
		SKIP_TEST (lb, n, pnt1, pnt2);		\
		SKIP_TEST (lb, n, pnt2, pnt1);		\
	}

	n = gp -> pts -> n;

	hp = hpair;
	for (i = 0; i < nhor; i++, hp++) {
		/* Backbone must permit full-sets of 4 or fewer points... */
		SKIP_IF_BAD_BOUNDS (lb, n, hp -> a, hp -> b);

		vp = vpair;
		for (j = 0; j < nver; j++, vp++) {
			if (NOT ((hp -> a.x < vp -> a.x) AND
				 (vp -> a.x < hp -> b.x) AND
				 (vp -> a.y < hp -> a.y) AND
				 (hp -> a.y < vp -> b.y))) {
					continue;
			}
			/* We have a cross! */

			/* Backbone must permit 4 or more points... */
			SKIP_IF_BAD_BOUNDS (lb, n, vp -> a, vp -> b);

			terms.n		= 4;
			terms.a [0]	= vp -> a;
			terms.a [1]	= vp -> b;
			terms.a [2]	= hp -> a;
			terms.a [3]	= hp -> b;

			st.x		= vp -> a.x;
			st.y		= hp -> a.y;
			st.pnum		= 0;
			st.other_pnum	= 0;

			steins.n	= 2;
			steins.a [0]	= st;
			steins.a [1]	= st;

			fset.next	= NULL;
			fset.style	= CROSS;
			fset.tree_num	= 0;
			fset.tree_len	= 0;
			fset.terminals	= &terms;
			fset.steiners	= &steins;

			if (passes_fine_tests (gp -> pts, &fset)) {
				++(gp -> fcount);	/* tally fine sets. */

				fsp = new (sizeof (*fsp));
				fsp -> next		= NULL;
				fsp -> style		= CROSS;
				fsp -> tree_num		= 0;
				fsp -> tree_len		= fset.tree_len;
				fsp -> terminals	= save_pset (&terms);
				fsp -> steiners		= save_pset (&steins);

				*(gp -> hookp) = fsp;
				gp -> hookp = &(fsp -> next);
			}
		}
	}

#undef	SKIP_TEST
#undef	SKIP_IF_BAD_BOUNDS
}

/*
 * This routine generates all full-sets whose backbones are precisely
 * horizontal.
 */

	static
	void
generate_all_horiz_full_sets (

struct ginfo *		gp,	/* IN/OUT - generation info */
int			nhor,	/* IN - number of horizontal backbones */
struct pair *		hpair,	/* IN - horizontal backbones */
int *			lb	/* IN - backbone lower bounds */
)
{
int			i;
int			j;
int			n;
int			bnd1;
int			bnd2;
int			lbound;
int			nvalid;
struct point *		p1;
struct point *		ap;
struct point *		bp;
struct point *		dstp;
struct pair *		hp;
struct point		s;
struct pset		valid_pts;

	n = gp -> pts -> n;

	hp = hpair;
	for (i = 0; i < nhor; i++, hp++) {
		bnd1 = BOUND (lb, n, hp -> a, hp -> b);
		if (bnd1 < 2) continue;
		bnd2 = BOUND (lb, n, hp -> b, hp -> a);
		if (bnd2 < 2) continue;

		lbound = min (bnd1, bnd2);

		ap = &(hp -> a);
		bp = &(hp -> b);

		dstp = &valid_pts.a [0];
		nvalid = 0;

		p1 = &(gp -> pts -> a [0]);
		for (j = 0; j < n; j++, p1++) {
			if (NOT BITON (gp -> tmap, j)) continue;
			if (NOT STRICTLY_BETWEEN_X (ap, p1, bp)) continue;
			s.x	= p1 -> x;
			s.y	= ap -> y;
			if (NOT is_point_close_to (gp -> pts, p1, &s)) {
				*dstp++ = *p1;
				++nvalid;
			}
		}
		valid_pts.n = nvalid;
		gp -> ap	= ap;
		gp -> bp	= bp;
		gen_E (gp, lbound, &valid_pts);
	}
}

/*
 * This routine generates all full-sets whose backbones are precisely
 * vertical.
 */

	static
	void
generate_all_vert_full_sets (

struct ginfo *		gp,	/* IN/OUT - generation info */
int			nver,	/* IN - number of vertical backbones */
struct pair *		vpair,	/* IN - vertical backbones */
int *			lb	/* IN - backbone lower bounds */
)
{
int			i;
int			j;
int			n;
int			bnd1;
int			bnd2;
int			lbound;
int			nvalid;
struct point *		p1;
struct point *		ap;
struct point *		bp;
struct point *		dstp;
struct pair *		vp;
struct point		s;
struct pset		valid_pts;

	n = gp -> pts -> n;

	vp = vpair;
	for (i = 0; i < nver; i++, vp++) {
		bnd1 = BOUND (lb, n, vp -> a, vp -> b);
		if (bnd1 < 2) continue;
		bnd2 = BOUND (lb, n, vp -> b, vp -> a);
		if (bnd2 < 2) continue;

		lbound = min (bnd1, bnd2);

		ap = &(vp -> a);
		bp = &(vp -> b);

		dstp = &valid_pts.a [0];
		nvalid = 0;

		p1 = &(gp -> pts -> a [0]);
		for (j = 0; j < n; j++, p1++) {
			if (NOT BITON (gp -> tmap, j)) continue;
			if (NOT STRICTLY_BETWEEN_Y (ap, p1, bp)) continue;
			s.x	= ap -> x;
			s.y	= p1 -> y;
			if (NOT is_point_close_to (gp -> pts, p1, &s)) {
				*dstp++ = *p1;
				++nvalid;
			}
		}
		valid_pts.n = nvalid;
		gp -> ap	= ap;
		gp -> bp	= bp;
		gen_N (gp, lbound, &valid_pts);
	}
}

/*
 * This routine generates all potential full sets for the Horizontal
 * backbone described by the given "ginfo".
 */

	static
	void
gen_E (

struct ginfo *		gp,	/* IN/OUT - recursion info for backbone. */
int			lb,	/* IN - lower bound (less backbone endpts). */
struct pset *		vpts	/* IN - valid points for this backbone. */
)
{
int			i;
int			n;
struct point *		p1;
struct point *		ap;	/* backbone point A. */
struct point *		bp;	/* backbone point B. */
enum side		last;
int			n_groups;
struct point *		groups [MAX_TERMINALS + 1];

	ap = gp -> ap;
	bp = gp -> bp;

	/* Sort valid points in increasing order by X coord. */
	sort_x_inc (vpts);

	last = UNKNOWN;
	n_groups = 0;
	n = vpts -> n;
	p1 = &(vpts -> a [0]);
	for (i = 0; i < n; p1++, i++) {
		if (p1 -> y > ap -> y) {
			if (last NE LEFT) {
				groups [n_groups++] = p1;
				last = LEFT;
			}
		}
		else {
			if (last NE RIGHT) {
				groups [n_groups++] = p1;
				last = RIGHT;
			}
		}
	}
	groups [n_groups] = p1;		/* terminate last group. */

	/* The first two points in each full set are the backbone endpoints. */
	gp -> gen.a [0] 	= *ap;
	gp -> gen.a [1] 	= *bp;
	gp -> gen.n		= 2;
	gp -> style		= ONLY_H;
	gp -> extras.n		= 0;
	gp -> lbound		= lb - 2;

	/* Gen all full-sets that alternate starting with the 1st group. */
	recurse_horz_inc (gp,
			  &(gp -> gen.a [2]),
			  &(gp -> stn.a [0]),
			  0,
			  &groups [0],
			  n_groups);

	/* Gen all full-sets that alternate starting with the 2nd group. */
	if (gp -> lbound < 1) {
		/* Require at least 1 point this time... */
		gp -> lbound = 1;
	}
	recurse_horz_inc (gp,
			  &(gp -> gen.a [2]),
			  &(gp -> stn.a [0]),
			  0,
			  &groups [1],
			  n_groups - 1);
}

/*
 * This routine generates all potential full sets for the Vertical
 * backbone described by the given "ginfo".
 */

	static
	void
gen_N (

struct ginfo *		gp,	/* IN/OUT - recursion info for backbone. */
int			lb,	/* IN - lower bound (less backbone endpts). */
struct pset *		vpts	/* IN - valid points for this backbone. */
)
{
int			i;
int			n;
struct point *		p1;
struct point *		ap;	/* backbone point A. */
struct point *		bp;	/* backbone point B. */
enum side		last;
int			n_groups;
struct point *		groups [MAX_TERMINALS + 1];

	ap = gp -> ap;
	bp = gp -> bp;

	/* Sort valid points in increasing order by Y coord. */
	sort_y_inc (vpts);

	last = UNKNOWN;
	n_groups = 0;
	n = vpts -> n;
	p1 = &(vpts -> a [0]);
	for (i = 0; i < n; p1++, i++) {
		if (p1 -> x < ap -> x) {
			if (last NE LEFT) {
				groups [n_groups++] = p1;
				last = LEFT;
			}
		}
		else {
			if (last NE RIGHT) {
				groups [n_groups++] = p1;
				last = RIGHT;
			}
		}
	}
	groups [n_groups] = p1;		/* terminate last group. */

	/* The first two points in each full set are the backbone endpoints. */
	gp -> gen.a [0] 	= *ap;
	gp -> gen.a [1] 	= *bp;
	gp -> gen.n		= 2;
	gp -> style		= ONLY_V;
	gp -> extras.n		= 0;
	gp -> lbound		= lb - 2;

	/* Gen all full-sets that alternate starting with the 1st group. */
	recurse_vert_inc (gp,
			  &(gp -> gen.a [2]),
			  &(gp -> stn.a [0]),
			  0,
			  &groups [0],
			  n_groups);

	/* Gen all full-sets that alternate starting with the 2nd group. */
	if (gp -> lbound < 1) {
		/* Require at least 1 point this time... */
		gp -> lbound = 1;
	}
	recurse_vert_inc (gp,
			  &(gp -> gen.a [2]),
			  &(gp -> stn.a [0]),
			  0,
			  &groups [1],
			  n_groups - 1);
}

/*
 * This routine generates all potential full sets for the backbone
 * described by the given "ginfo".  The backbone has angle 0 < theta < 90.
 * The backbone therefore always has a corner.
 */

	static
	void
gen_NE (

struct ginfo *		gp,	/* IN/OUT - recursion info for backbone. */
int			lb,	/* IN - lower bound (less backbone endpts). */
struct pset *		vpts	/* IN - valid points for this backbone. */
)
{
int			i;
int			n;
struct point *		p1;
struct point *		p2;
struct point *		ap;	/* backbone point A. */
struct point *		bp;	/* backbone point B. */
struct point		c;	/* corner point */
struct point		s;	/* Steiner point */

	ap = gp -> ap;
	bp = gp -> bp;

	/* The first two points in each full set are the backbone endpoints. */
	gp -> gen.a [0] = *ap;
	gp -> gen.a [1] = *bp;

	c.x	= ap -> x;
	c.y	= bp -> y;

	if (NOT is_point_close_to (gp -> pts, &c, bp)) {
		/* Can do ZERO_H_REST_V backbones! */
		gp -> style	= ZERO_H_REST_V;
		gp -> extras.n	= 0;
		gp -> gen.n	= 2;	/* first 2 points pre-determined */
		vert_dec (gp, lb, vpts);
	}

	n = vpts -> n;
	p1 = &(vpts -> a [0]);
	gp -> extras.n = 0;
	p2 = &(gp -> extras.a [0]);
	for (i = 0; i < n; p1++, i++) {
		if (NOT STRICTLY_BETWEEN_X (ap, p1, bp)) continue;
		if (p1 -> y <= bp -> y) continue;
		s.x = p1 -> x;
		s.y = bp -> y;
		if (is_point_close_to (gp -> pts, &c, &s)) continue;
		if (is_point_close_to (gp -> pts, &s, bp)) continue;

		*p2++ = *p1;
		++(gp -> extras.n);
	}

	if (gp -> extras.n > 0) {
		gp -> style = ONE_H_REST_V;
		gp -> gen.n = 3;	/* first 3 points pre-determined */
		vert_dec (gp, lb - 1, vpts);
	}

	if (NOT is_point_close_to (gp -> pts, ap, &c)) {
		gp -> style	= ZERO_V_REST_H;
		gp -> extras.n	= 0;
		gp -> gen.n	= 2;	/* first 2 points pre-determined */
		horiz_inc (gp, max (lb, 2), vpts);
	}

	p1 = &(vpts -> a [0]);
	gp -> extras.n = 0;
	p2 = &(gp -> extras.a [0]);
	for (i = 0; i < n; p1++, i++) {
		if (NOT STRICTLY_BETWEEN_Y (ap, p1, bp)) continue;
		if (p1 -> x >= ap -> x) continue;
		s.x = ap -> x;
		s.y = p1 -> y;
		if (is_point_close_to (gp -> pts, ap, &s)) continue;
		if (is_point_close_to (gp -> pts, &s, &c)) continue;

		*p2++ = *p1;
		++(gp -> extras.n);
	}

	if (gp -> extras.n > 0) {
		gp -> style = ONE_V_REST_H;
		gp -> gen.n = 3;	/* first 3 points pre-determined */
		horiz_inc (gp, max (lb - 1, 2), vpts);
	}
}

/*
 * This routine generates all potential full sets for the backbone
 * described by the given "ginfo".  The backbone has angle -90 < theta < 0.
 */

	static
	void
gen_SE (

struct ginfo *		gp,	/* IN/OUT - recursion info for backbone. */
int			lb,	/* IN - lower bound (less backbone endpts). */
struct pset *		vpts	/* IN - valid points for this backbone. */
)
{
int			i;
int			n;
struct point *		p1;
struct point *		p2;
struct point *		ap;	/* backbone point A. */
struct point *		bp;	/* backbone point B. */
struct point		c;	/* corner point */
struct point		s;	/* Steiner point */

	ap = gp -> ap;
	bp = gp -> bp;

	/* The first two points in each full set are the backbone endpoints. */
	gp -> gen.a [0] = *ap;
	gp -> gen.a [1] = *bp;

	c.x	= ap -> x;
	c.y	= bp -> y;

	if (NOT is_point_close_to (gp -> pts, &c, bp)) {
		/* Can do ZERO_H_REST_V backbones! */
		gp -> style	= ZERO_H_REST_V;
		gp -> extras.n	= 0;
		gp -> gen.n	= 2;	/* first 2 points pre-determined */
		vert_inc (gp, lb, vpts);
	}

	n = vpts -> n;
	p1 = &(vpts -> a [0]);
	gp -> extras.n = 0;
	p2 = &(gp -> extras.a [0]);
	for (i = 0; i < n; p1++, i++) {
		if (NOT STRICTLY_BETWEEN_X (ap, p1, bp)) continue;
		if (p1 -> y >= bp -> y) continue;
		s.x = p1 -> x;
		s.y = bp -> y;
		if (is_point_close_to (gp -> pts, &c, &s)) continue;
		if (is_point_close_to (gp -> pts, &s, bp)) continue;

		*p2++ = *p1;
		++(gp -> extras.n);
	}

	if (gp -> extras.n > 0) {
		gp -> style = ONE_H_REST_V;
		gp -> gen.n = 3;	/* first 3 points pre-determined */
		vert_inc (gp, lb - 1, vpts);
	}

	if (NOT is_point_close_to (gp -> pts, ap, &c)) {
		/* Can to ZERO_V_REST_H backbones! */
		gp -> style	= ZERO_V_REST_H;
		gp -> extras.n	= 0;
		gp -> gen.n	= 2;	/* first 2 points pre-determined */
		horiz_inc (gp, max (lb, 2), vpts);
	}

	p1 = &(vpts -> a [0]);
	gp -> extras.n = 0;
	p2 = &(gp -> extras.a [0]);
	for (i = 0; i < n; p1++, i++) {
		if (NOT STRICTLY_BETWEEN_Y (ap, p1, bp)) continue;
		if (p1 -> x >= ap -> x) continue;
		s.x = ap -> x;
		s.y = p1 -> y;
		if (is_point_close_to (gp -> pts, ap, &s)) continue;
		if (is_point_close_to (gp -> pts, &s, &c)) continue;

		*p2++ = *p1;
		++(gp -> extras.n);
	}

	if (gp -> extras.n > 0) {
		gp -> style = ONE_V_REST_H;
		gp -> gen.n = 3;	/* first 3 points pre-determined */
		horiz_inc (gp, max (lb - 1, 2), vpts);
	}
}

/*
 * This routine generates all full sets of the given style off of the
 * given backbone.  The "rest" points lie along the vertical leg in
 * DECREASING Y coordinates.  The given lower bound is for points added
 * along this leg.
 */

	static
	void
vert_dec (

struct ginfo *		gp,	/* IN/OUT - recursion info for backbone. */
int			lb,	/* IN - lower bound of pts on "rest" leg. */
struct pset *		vpts	/* IN - valid points for backbone. */
)
{
int			i;
int			n;
struct point *		ap;
struct point *		bp;
struct point *		p1;
struct point *		p2;
enum side		last;
int			n_groups;
struct pset		y_sorted;
struct point *		groups [MAX_TERMINALS + 1];

	gp -> lbound	= lb;

	ap = gp -> ap;
	bp = gp -> bp;

	n = 0;
	p2 = &y_sorted.a [0];

	p1 = &(vpts -> a [0]);
	for (i = 0; i < vpts -> n; p1++, i++) {
		if (STRICTLY_BETWEEN_Y (ap, p1, bp)) {
			*p2++ = *p1;
			++n;
		}
	}
	y_sorted.n = n;
	sort_y_dec (&y_sorted);

	last = UNKNOWN;
	n_groups = 0;
	p1 = &y_sorted.a [0];
	for (i = 0; i < n; p1++, i++) {
		if (STRICTLY_BETWEEN_X (p1, ap, bp)) {
			if (last NE LEFT) {
				groups [n_groups++] = p1;
				last = LEFT;
			}
		}
		else {
			if (last EQ LEFT) {
				groups [n_groups++] = p1;
				last = RIGHT;
			}
		}
	}
	groups [n_groups] = p1;		/* terminate last group. */
	/* The points X in group I are groups [I] <= X < groups [I+1]. */
	n = gp -> gen.n;
	recurse_vert_dec (gp,
			  &(gp -> gen.a [n]),
			  &(gp -> stn.a [n-2]),
			  0,
			  &groups [0],
			  n_groups);
}

/*
 * This routine generates all full sets of the given style off of the
 * given backbone.  The "rest" points lie along the vertical leg in
 * INCREASING Y coordinates.  The given lower bound is for points added
 * along this leg.
 */

	static
	void
vert_inc (

struct ginfo *		gp,	/* IN/OUT - recursion info for backbone. */
int			lb,	/* IN - lower bound of pts on "rest" leg. */
struct pset *		vpts	/* IN - valid points for backbone. */
)
{
int			i;
int			n;
struct point *		ap;
struct point *		bp;
struct point *		p1;
struct point *		p2;
enum side		last;
int			n_groups;
struct pset		y_sorted;
struct point *		groups [MAX_TERMINALS + 1];

	gp -> lbound	= lb;

	ap = gp -> ap;
	bp = gp -> bp;

	n = 0;
	p2 = &y_sorted.a [0];

	p1 = &(vpts -> a [0]);
	for (i = 0; i < vpts -> n; p1++, i++) {
		if (STRICTLY_BETWEEN_Y (ap, p1, bp)) {
			*p2++ = *p1;
			++n;
		}
	}
	y_sorted.n = n;
	sort_y_inc (&y_sorted);

	last = UNKNOWN;
	n_groups = 0;
	p1 = &y_sorted.a [0];
	for (i = 0; i < n; p1++, i++) {
		if (BETWEEN_X (p1, ap, bp)) {
			if (last NE LEFT) {
				groups [n_groups++] = p1;
				last = LEFT;
			}
		}
		else {
			if (last EQ LEFT) {
				groups [n_groups++] = p1;
				last = RIGHT;
			}
		}
	}
	groups [n_groups] = p1;		/* terminate last group. */
	/* The points X in group I are groups [I] <= X < groups [I+1]. */
	n = gp -> gen.n;
	recurse_vert_inc (gp,
			  &(gp -> gen.a [n]),
			  &(gp -> stn.a [n-2]),
			  0,
			  &groups [0],
			  n_groups);
}

/*
 * This routine generates all full sets of the given style off of the
 * given backbone.  The "rest" points lie along the horizontal leg in
 * INCREASING X coordinates.  The given lower bound is for points added
 * along this leg.
 */

	static
	void
horiz_inc (

struct ginfo *		gp,	/* IN/OUT - recursion info for backbone. */
int			lb,	/* IN - lower bound of pts on "rest" leg. */
struct pset *		vpts	/* IN - valid points for backbone. */
)
{
int			i;
int			n;
struct point *		ap;
struct point *		bp;
struct point *		p1;
struct point *		p2;
enum side		last;
int			n_groups;
struct pset		x_sorted;
struct point *		groups [MAX_TERMINALS + 1];

	gp -> lbound	= lb;

	ap = gp -> ap;
	bp = gp -> bp;

	n = 0;
	p2 = &x_sorted.a [0];

	p1 = &(vpts -> a [0]);
	for (i = 0; i < vpts -> n; p1++, i++) {
		if (STRICTLY_BETWEEN_X (ap, p1, bp)) {
			*p2++ = *p1;
			++n;
		}
	}
	x_sorted.n = n;
	sort_x_inc (&x_sorted);

	last = UNKNOWN;
	n_groups = 0;
	p1 = &x_sorted.a [0];
	for (i = 0; i < n; p1++, i++) {
		if (BETWEEN_Y (p1, bp, ap)) {
			if (last NE LEFT) {
				groups [n_groups++] = p1;
				last = LEFT;
			}
		}
		else {
			if (last EQ LEFT) {
				groups [n_groups++] = p1;
				last = RIGHT;
			}
		}
	}
	groups [n_groups] = p1;		/* terminate last group. */
	/* The points X in group I are groups [I] <= X < groups [I+1]. */
	n = gp -> gen.n;
	recurse_horz_inc (gp,
			  &(gp -> gen.a [n]),
			  &(gp -> stn.a [n-2]),
			  0,
			  &groups [0],
			  n_groups);
}

/*
 * This routine recursively enumerates all possible combinations of
 * properly alternating "side barbs" down the "rest" leg of the current
 * backbone, which is the vertical leg.  The lower bound places a
 * minimum constraint on the number of these alternating points we
 * include in our enumerated sets.
 */

	static
	void
recurse_vert_inc (

struct ginfo *		gp,	/* IN/OUT - recursion info for backbone. */
struct point *		genp,	/* IN/OUT - where to put next gen'd point. */
struct point *		stnp,	/* IN/OUT - where to put next steiner pt. */
int			ngen,	/* IN - number of points gen'd so far. */
struct point **		grp,	/* IN - next left/right group to take from. */
int			nleft	/* IN - number of left/right groups left. */
)
{
int			i;
int			n;
struct point *		p1;
struct point *		p2;

	if (ngen >= gp -> lbound) {
		/* We have a solution! */
		if (gp -> extras.n <= 0) {
			gp -> gen.n = 2 + ngen;
			check_potential_full_set (gp);
		}
		else {
			gp -> gen.n = 3 + ngen;
			n = gp -> extras.n;
			p1 = &(gp -> extras.a [0]);
			for (i = 0; i < n; p1++, i++) {
				gp -> gen.a [2] = *p1;
				gp -> stn.a [0].x = p1 -> x;
				gp -> stn.a [0].y = gp -> bp -> y;
				check_potential_full_set (gp);
			}
		}
	}

	/* The following loops choose each possible single point to add	*/
	/* to the set next.  The iteration stops either if we run out	*/
	/* of groups to choose points from, or if too few groups are	*/
	/* left to achieve a set big enough to use.			*/
	while ((nleft > 0) AND ((ngen + nleft) >= gp -> lbound)) {
		/* Try using each point from the current group in turn. */
		p1 = grp [0];
		p2 = grp [1];
		for (; p1 < p2; p1++) {
			if ((ngen > 0) AND (p1 -> y <= genp [-1].y)) {
				/* The alternating points must ADVANCE	*/
				/* down the backbone, not alternate in	*/
				/* place!				*/
				continue;
			}
			*genp = *p1;
			stnp -> x = gp -> ap -> x;
			stnp -> y = p1 -> y;
			recurse_vert_inc (gp,
					  genp+1,
					  stnp+1,
					  ngen+1,
					  grp+1,
					  nleft-1);
		}
		/* Now throw away the current left/right or right/left	*/
		/* pair of groups and repeat.  This corrsponds to	*/
		/* taking no point from either of these groups...	*/
		grp += 2;
		nleft -= 2;
	}
}

/*
 * This routine recursively enumerates all possible combinations of
 * properly alternating "side barbs" down the "rest" leg of the current
 * backbone, which is the vertical leg.  The lower bound places a
 * minimum constraint on the number of these alternating points we
 * include in our enumerated sets.
 */

	static
	void
recurse_vert_dec (

struct ginfo *		gp,	/* IN/OUT - recursion info for backbone. */
struct point *		genp,	/* IN/OUT - where to put next gen'd point. */
struct point *		stnp,	/* IN/OUT - where to put next steiner pt. */
int			ngen,	/* IN - number of points gen'd so far. */
struct point **		grp,	/* IN - next left/right group to take from. */
int			nleft	/* IN - number of left/right groups left. */
)
{
int			i;
int			n;
struct point *		p1;
struct point *		p2;

	if (ngen >= gp -> lbound) {
		/* We have a solution! */
		if (gp -> extras.n <= 0) {
			gp -> gen.n = 2 + ngen;
			check_potential_full_set (gp);
		}
		else {
			gp -> gen.n = 3 + ngen;
			n = gp -> extras.n;
			p1 = &(gp -> extras.a [0]);
			for (i = 0; i < n; p1++, i++) {
				gp -> gen.a [2] = *p1;
				gp -> stn.a [0].x = p1 -> x;
				gp -> stn.a [0].y = gp -> bp -> y;
				check_potential_full_set (gp);
			}
		}
	}

	/* The following loops choose each possible single point to add	*/
	/* to the set next.  The iteration stops either if we run out	*/
	/* of groups to choose points from, or if too few groups are	*/
	/* left to achieve a set big enough to use.			*/
	while ((nleft > 0) AND ((ngen + nleft) >= gp -> lbound)) {
		/* Try using each point from the current group in turn. */
		p1 = grp [0];
		p2 = grp [1];
		for (; p1 < p2; p1++) {
			if ((ngen > 0) AND (p1 -> y >= genp [-1].y)) {
				/* The alternating points must ADVANCE	*/
				/* down the backbone, not alternate in	*/
				/* place!				*/
				continue;
			}
			*genp = *p1;
			stnp -> x = gp -> ap -> x;
			stnp -> y = p1 -> y;
			recurse_vert_dec (gp,
					  genp+1,
					  stnp+1,
					  ngen+1,
					  grp+1,
					  nleft-1);
		}
		/* Now throw away the current left/right or right/left	*/
		/* pair of groups and repeat.  This corrsponds to	*/
		/* taking no point from either of these groups...	*/
		grp += 2;
		nleft -= 2;
	}
}

/*
 * This routine recursively enumerates all possible combinations of
 * properly alternating "side barbs" down the "rest" leg of the current
 * backbone, which is the horizontal leg.  The lower bound places a
 * minimum constraint on the number of these alternating points we
 * include in our enumerated sets.
 */

	static
	void
recurse_horz_inc (

struct ginfo *		gp,	/* IN/OUT - recursion info for backbone. */
struct point *		genp,	/* IN/OUT - where to put next gen'd point. */
struct point *		stnp,	/* IN/OUT - where to put next steiner pt. */
int			ngen,	/* IN - number of points gen'd so far. */
struct point **		grp,	/* IN - next left/right group to take from. */
int			nleft	/* IN - number of left/right groups left. */
)
{
int			i;
int			n;
struct point *		p1;
struct point *		p2;

	if (ngen >= gp -> lbound) {
		/* We have a solution! */
		if (gp -> extras.n <= 0) {
			gp -> gen.n = 2 + ngen;
			check_potential_full_set (gp);
		}
		else {
			gp -> gen.n = 3 + ngen;
			n = gp -> extras.n;
			p1 = &(gp -> extras.a [0]);
			for (i = 0; i < n; p1++, i++) {
				gp -> gen.a [2] = *p1;
				gp -> stn.a [0].x = gp -> ap -> x;
				gp -> stn.a [0].y = p1 -> y;
				check_potential_full_set (gp);
			}
		}
	}

	/* The following loops choose each possible single point to add	*/
	/* to the set next.  The iteration stops either if we run out	*/
	/* of groups to choose points from, or if too few groups are	*/
	/* left to achieve a set big enough to use.			*/
	while ((nleft > 0) AND ((ngen + nleft) >= gp -> lbound)) {
		/* Try using each point from the current group in turn. */
		p1 = grp [0];
		p2 = grp [1];
		for (; p1 < p2; p1++) {
			if ((ngen > 0) AND (p1 -> x <= genp [-1].x)) {
				/* The alternating points must ADVANCE	*/
				/* down the backbone, not alternate in	*/
				/* place!				*/
				continue;
			}
			*genp = *p1;
			stnp -> x = p1 -> x;
			stnp -> y = gp -> bp -> y;
			recurse_horz_inc (gp,
					  genp+1,
					  stnp+1,
					  ngen+1,
					  grp+1,
					  nleft-1);
		}
		/* Now throw away the current left/right or right/left	*/
		/* pair of groups and repeat.  This corrsponds to	*/
		/* taking no point from either of these groups...	*/
		grp += 2;
		nleft -= 2;
	}
}

/*
 * This routine takes a set that has passed all of the coarse-grained
 * tests, and applies the fine-grained (i.e. expensive) tests on it.
 * If it passes these tests also, then we commit it to memory.
 */

	static
	void
check_potential_full_set (

struct ginfo *		gp	/* IN/OUT - recursion info for backbone. */
)
{
struct pset *		terms;
struct pset *		steiners;
struct full_set *	fsp;
enum backbone		style;
struct full_set		fset;

	++(gp -> ccount);	/* chalk up one more coarse set... */
	terms = &(gp -> gen);
	steiners = &(gp -> stn);
	steiners -> n = (terms -> n - 2);
	style = gp -> style;
	if (terms -> n <= 2) {
		style = ZERO_H_ZERO_V;
	}

	/* We now have a potential full-set! */
	fset.next	= NULL;
	fset.style	= style;
	fset.tree_num	= 0;
	fset.tree_len	= 0;
	fset.terminals	= terms;
	fset.steiners	= steiners;

	if (passes_fine_tests (gp -> pts, &fset)) {
		++(gp -> fcount);	/* tally fine sets. */

		fsp = new (sizeof (*fsp));
		fsp -> next		= NULL;
		fsp -> style		= style;
		fsp -> tree_num		= 0;
		fsp -> tree_len		= fset.tree_len;
		fsp -> terminals	= save_pset (terms);
		fsp -> steiners		= save_pset (steiners);

		*(gp -> hookp) = fsp;
		gp -> hookp = &(fsp -> next);
	}
}

/*
 * This routine creates a "this bed is just right" copy of the given
 * point set in heap storage.
 */

	struct pset *
save_pset (

struct pset *		p	/* IN - point set to copy. */
)
{
struct pset *		newp;
size_t			size;

	size = PSET_SIZE (p -> n);
	newp = new (size);
	(void) memcpy (newp, p, size);

	return (newp);
}
