/***********************************************************************

	File:	main.c
	Rev:	a-1
	Date:	02/20/93

	Copyright (c) 1993, by The University of Virginia

************************************************************************

	The main routine for the Rectilinear Steiner Tree program.

************************************************************************

	Modification Log:

	a-1:	02/20/93	dmw
		: Created.

************************************************************************/

#include "steiner.h"
#include <signal.h>


/*
 * Global Routines
 */

int			main (int, char **);


/*
 * External References
 */

	/* none */


/*
 * Local Routines
 */

static void		check_limit (int);
static void		decode_params (int, char **);
static dist_t		do_kahng_robins (struct pset *);
static dist_t		do_minimum_spanning_tree (struct pset *);
static int32u		get_cpu_time_limit (char *);
static bool		get_coord (coord_t *);
static void		get_points (struct pset *);
static void		usage (void);


/*
 * Local Variables
 */

static int32u		cpu_time_limit;
static char *		me;

/*
 * This is the main routine for the Rectilinear Steiner Tree program.
 */

	int
main (

int		argc,
char **		argv
)
{
int			i;
int			ntrees;
int			num_fset_masks;
bitmap_t *		fset_mask;
bitmap_t *		smt_mask;
struct full_set *	fsp;
struct full_set *	trivial_fsp;
struct full_set *	smt;
struct full_set *	tmp;
dist_t			smt_len;
cpu_time_t		T0;
cpu_time_t		T1;
bool			failed;
int			count;
char			tbuf [20];
char			title [128];
struct cinfo		cinfo;
struct pset		pts;
bitmap_t		tmap [N_TERM_BMAP];
bitmap_t		arts [N_TERM_BMAP];
int			lb [MAX_TERMINALS * MAX_TERMINALS];

	setbuf (stdout, NULL);

	decode_params (argc, argv);

	get_points (&pts);

	if (cpu_time_limit > 0) {
		/* Enable CPU time limitation. */
		check_limit (0);
	}

	T0 = get_cpu_time ();

	/* Create a bit-map of all terminals to consider */
	for (i = 0; i < N_TERM_BMAP; i++) {
		tmap [i] = 0;
	}
	for (i = 0; i < pts.n; i++) {
		SETBIT (tmap, i);
	}

	/* Prepare for plotting all terminals. */
	define_Plot_Terminals (&pts);

	init_tables ();

#if 1	/* SKIP_STEINER_TREE */

	/* Remove all duplicate terminals from the input -- create one	*/
	/* trivial full-set for each terminal that duplicates another.	*/
	trivial_fsp = generate_duplicate_full_sets (&pts, &tmap [0]);

	compute_mst_longest_edges (&pts);

	coarse_grained_tests (&pts, &tmap [0], &lb [0]);

	fsp = generate_all_full_sets (&pts, &tmap [0], &lb [0]);

#if 1
	fsp = sort_full_trees (fsp);
#endif

	/* Number all of the full-sets. */
	ntrees = 0;
	for (tmp = fsp; tmp NE NULL; tmp = tmp -> next) {
		tmp -> tree_num = ntrees++;
	}

	/* Create a bit-mask of all full-sets to consider. */
	num_fset_masks = BMAP_ELTS (ntrees);
	fset_mask = NEWA (num_fset_masks, bitmap_t);
	smt_mask  = NEWA (num_fset_masks, bitmap_t);
	(void) memset (fset_mask, 0, num_fset_masks * sizeof (bitmap_t));
	(void) memset (smt_mask,  0, num_fset_masks * sizeof (bitmap_t));
	for (i = 0; i < ntrees; i++) {
		SETBIT (fset_mask, i);
	}

	build_compatibility_data (&pts, &tmap [0], fsp, fset_mask, &cinfo);

	if (Print_Full_Sets) {
		plot_full_sets (fset_mask, &cinfo);
	}

	/* Thin out impossible full-sets.  (Will turn off the bit in	*/
	/* fset_mask corresponding to each full-set removed.)  It also	*/
	/* modifies the compatibility arrays.				*/
	initial_pruning (&tmap [0], fset_mask, &cinfo);

	/* Count number of full-sets remaining after Initial Pruning. */
	count = 0;
	for (i = 0; i < num_fset_masks; i++) {
		count += NBITSON (fset_mask [i]);
	}

	T1 = get_cpu_time ();
	convert_cpu_time (T1 - T0, tbuf);
	(void) printf ("  %%  stage one time:  %s seconds\n", tbuf);

	print_mask (" % FSets after pruning",
		    fset_mask,
		    cinfo.full_tree_count);

#if 1
	find_articulation_points (tmap, fset_mask, &cinfo, arts);
	print_mask (" % Articulation Points", arts, pts.n);
#endif

	if (Print_Relief_Map) {
		(void) sprintf (title,
				"Relief Plot, %ld Points, %ld Full Sets, %s seconds",
				(long) pts.n,
				(long) count,
				tbuf);
		relief_plot_full_sets (title, fset_mask, &cinfo);
	}

	if (Print_Overlaid_Full_Sets) {
		sprintf (title,
			 "All Full Sets:  %d points,  %s seconds",
			 pts.n, tbuf);
		overlay_plot_subset (title, fset_mask, &cinfo);
	}

	if (Print_Split_Full_Sets) {
		plot_split_full_sets (fset_mask, &cinfo);
	}

	if (Stop_After_Stage_1) {
		exit (0);
	}

	failed = split_and_solve (&tmap [0], fset_mask, &cinfo, smt_mask);

	if (failed) {
		/* For some reason, no SMT could be found.  This is	*/
		/* definitely a bug, since one is certain to exist!	*/
		fatal ("main: Bug 1.");
	}

	smt = NULL;
	for (i = 0; i < cinfo.full_tree_count; i++) {
		if (NOT BITON (smt_mask, i)) continue;
		tmp = copy_full_set (cinfo.full_trees [i]);
		tmp -> next = smt;
		smt = tmp;
	}

	/* Add the trivial full-sets to the final solution. */
	while (trivial_fsp NE NULL) {
		tmp = trivial_fsp -> next;
		trivial_fsp -> next = smt;
		smt = trivial_fsp;
		trivial_fsp = tmp;
	}

	T1 = get_cpu_time ();

	smt_len = 0;
	for (tmp = smt; tmp NE NULL; tmp = tmp -> next) {
		smt_len += tmp -> tree_len;
	}
	convert_cpu_time (T1 - T0, tbuf);

	sprintf (title,
		 "Steiner Minimal Tree:  %d points,  length = %lu,  %s seconds",
		 pts.n, smt_len, tbuf);

	overlay_plot_subset (title, smt_mask, &cinfo);

#else
	smt_len = 0;	/* Avoid fatal errors! */
#endif	/* SKIP_STEINER_TREE */

	if (NOT Skip_Minimum_Spanning_Tree) {
		dist_t	mst_len;

		mst_len = do_minimum_spanning_tree (&pts);
		if (mst_len < smt_len) {
			fatal ("MST shorter than SMT!");
		}
	}
	if (NOT Skip_Kahng_Robins) {
		dist_t	kr_len;
		kr_len = do_kahng_robins (&pts);
		if (kr_len < smt_len) {
			fatal ("Kahng-Robins shorter than SMT!");
		}
	}

	return (0);
}

/*
 * This routine decodes the various command-line arguments.
 */

	static
	void
decode_params (

int		argc,
char **		argv
)
{
char *		ap;
char		c;

	--argc;
	me = *argv++;
	while (argc > 0) {
		ap = *argv++;
		if (*ap NE '-') {
			usage ();
		}
		++ap;
		while ((c = *ap++) NE '\0') {
			switch (c) {
			case '1':
				Stop_After_Stage_1 = TRUE;
				break;

			case 'b':
				Print_Bounds = TRUE;
				break;

			case 'c':
				Print_Compatibility_Matrix = TRUE;
				break;

			case 'f':
				Print_Full_Sets = TRUE;
				break;

			case 'i':
				Use_Thorough_Pruning = TRUE;
				break;

			case 'k':
				Skip_Kahng_Robins = TRUE;
				break;

			case 'l':
				if (*ap EQ '\0') {
					if (argc <= 0) {
						usage ();
					}
					ap = *argv++;
					--argc;
				}
				cpu_time_limit = get_cpu_time_limit (ap);
				ap = "";
				break;

			case 'm':
				Skip_Minimum_Spanning_Tree = TRUE;
				break;

			case 'o':
				Print_Overlaid_Full_Sets = TRUE;
				break;

			case 'r':
				Print_Relief_Map = TRUE;
				break;

			case 'R':
				Disable_Recursive_Decomposition = TRUE;
				break;

			case 's':
				Print_Split_Full_Sets = TRUE;
				break;

			case 't':
				if (*ap EQ '\0') {
					if (argc <= 0) {
						usage ();
					}
					ap = *argv++;
					--argc;
				}
				Piece_Split_Threshold = atoi (ap);
				ap = "";
				break;

			default:
				usage ();
				break;
			}
		}
		--argc;
	}
}

/*
 * This routine decodes an argument string that represents an amount of
 * CPU time.  The time is the sum of a sequence of time groups.  Each
 * time group is a sequence of decimal digits followed by an optional
 * letter defining units of minutes, hours, days, etc.
 * The regular expression accepted is:
 *
 * ([0-9]+([dhms][a-z]*)?)*
 */

	static
	int32u
get_cpu_time_limit (ap)

char *		ap;
{
int32u		total;
int32u		group;
char		c;

	total = 0;
	c = *ap++;
	while (c NE '\0') {
		if ((c < '0') OR (c > '9')) {
			usage ();
		}
		group = c - '0';
		while (((c = *ap++) NE '\0') AND ('0' <= c) AND (c <= '9')) {
			group = (10 * group) + (c - '0');
		}
		switch (c) {
		case 'd':	group *= (24 * 60 * 60);	break;
		case 'h':	group *= (60 * 60);		break;
		case 'm':	group *= 60;			break;
		case 's':					break;
		case '\0':					break;
		default:	usage ();			break;
		}
		/* strip rest of unit name... */
		while ((c > 0) AND ((c < '0') OR (c > '9'))) {
			c = *ap++;
		}
		total += group;
	}

	return (total);
}

/*
 * This routine handles the "ALARM" signal, which goes off after a set
 * amount of *REAL* time.  We check the CPU time usage -- if it has
 * exceeded the limit, we print a message and exit.  Otherwise, we
 * reschedule another alarm signal.
 */

	static
	void
check_limit (

int		sig
)
{
int32u		T;

	T = (get_cpu_time () / 100);

	if (T >= cpu_time_limit) {
		(void) printf ("\n\n%% ***** CPU time limit of %lu"
				" seconds exceeded *****\n",
			       cpu_time_limit);
		exit (1);
	}

	T = cpu_time_limit - T;

	signal (SIGALRM, check_limit);
	alarm (T);
}

/*
 * This routine prints out the proper usage and exits.
 */

static char *	arg_doc [] = {
	"",
	"\t-1\tStops after stage 1 of the algorithm.",
	"\t-b\tPrints lower bounds during stage 1.",
	"\t-c\tPrints compatibility matrix.",
	"\t-f\tPrints all full-sets in \"fly specks\" fashion.",
	"\t-i\tEnables thorough (but very slow) initial pruning.",
	"\t-k\tDisables computation of Kahng-Robins Tree.",
	"\t-m\tDisables computation of Minimum Spanning Tree.",
	"\t-o\tPrints all full-sets in overlaid fashion.",
	"\t-r\tPrint all full-sets overlaid in \"relief map\" format.",
	"\t-R\tDisables recursive decomposition of tree sub-problems.",
	"\t-s\tPrints all full-sets as splitting pieces.",
	"\t-t N\tSets the piece-split threshold to N.",
	"",
	"\tExample CPU times are:",
	"\t\t-l 3days2hours30minutes15seconds",
	"\t\t-l1000seconds -l1000 -l 2h30m",
	"",
	NULL
};

	static
	void
usage (void)

{
char **		pp;
char *		p;

	(void) fprintf (stderr,
			"\nUsage: %s [-1bcfkmor] [-l cpu-time-limit] [-t threshold]\n",
			me);
	pp = &arg_doc [0];
	while ((p = *pp++) NE NULL) {
		(void) fprintf (stderr, "%s\n", p);
	}
	exit (1);
}

/*
 * This routine computes and plots the Minimum Spanning Tree for the
 * given point set.
 */

	static
	dist_t
do_minimum_spanning_tree (

struct pset *		pts		/* IN - the set of terminals. */
)
{
int			i;
int			n;
cpu_time_t		T0;
cpu_time_t		T1;
struct edge *		ep;
struct point *		p1;
struct point *		p2;
dist_t			mst_len;
char			tbuf [20];
struct edge		m [MAX_TERMINALS - 1];

	T0 = get_cpu_time ();
	n = mst (pts, &m [0]);

	T1 = get_cpu_time ();

	(void) printf ("\n  %%  Minimum Spanning Tree\n");

	(void) printf ("BeginPlot\n");
	(void) printf ("\tPlot_Terminals\n");
	mst_len = 0;
	for (i = 0; i < n; i++) {
		ep = &m [i];
		p1 = &(pts -> a [ep -> p1]);
		p2 = &(pts -> a [ep -> p2]);
		plot_segment (p1, p2);
		mst_len += ep -> len;
	}
	convert_cpu_time (T1 - T0, tbuf);
	(void) printf ("  (Minimum Spanning Tree:  %d points,  length = %lu,  %s seconds)\n",
		       pts -> n, mst_len, tbuf);
	(void) printf ("EndPlot\n\n");

	return (mst_len);
}

/*
 * This routine computes an approximate Steiner Minimal Tree using
 * the Kahng-Robins heuristic, and plots the result.
 */

	static
	dist_t
do_kahng_robins (

struct pset *		pts		/* IN - the set of terminals. */
)
{
int			i;
int			n;
int			save_n;
cpu_time_t		T0;
cpu_time_t		T1;
struct edge *		ep;
struct point *		p1;
struct point *		p2;
dist_t			kr_len;
char			tbuf [20];
struct edge		m [MAX_POINTS];

	save_n = pts -> n;

	T0 = get_cpu_time ();

	n = kahng_robins (pts, 0, &m [0]);

	T1 = get_cpu_time ();

	(void) printf ("\n  %%  Kahng Robins\n");

	(void) printf ("BeginPlot\n");
	pts -> n = save_n;
	(void) printf ("\tPlot_Terminals\n");
	kr_len = 0;
	for (i = 0; i < n; i++) {
		ep = &m [i];
		p1 = &(pts -> a [ep -> p1]);
		p2 = &(pts -> a [ep -> p2]);
		plot_segment (p1, p2);
		kr_len += ep -> len;
	}
	convert_cpu_time (T1 - T0, tbuf);
	(void) printf ("  (1-Steiner Heuristic:  %d points,  length = %lu,  %s seconds)\n",
		       pts -> n, kr_len, tbuf);
	(void) printf ("EndPlot\n\n");

	return (kr_len);
}

/*
 * This routine reads in a set of points from the standard input.
 */

	static
	void
get_points (

struct pset *		pts		/* OUT - point set. */
)

{
int			n;
struct point *		p;
coord_t			x;
coord_t			y;

	/* Zero out the entire point set. */
	(void) memset (pts, 0, sizeof (*pts));

	n = 0;
	p = &(pts -> a [0]);

	while (TRUE) {
		if (NOT get_coord (&x)) break;

		if (NOT get_coord (&y)) {
			(void) fprintf (stderr, "X coord without Y coord.\n");
			exit (1);
		}

		if (n >= MAX_TERMINALS) {
			(void) fprintf (stderr, "Too many terminals.\n");
			exit (1);
		}

		p -> x		= x;
		p -> y		= y;
		p -> pnum	= ((pnum_t) n);
		p -> other_pnum	= ((pnum_t) n);

		++n;
		++p;
	}
	pts -> n = (pnum_t) n;
}

/*
 * This routine reads in a single coordinate from the standard input.
 * If EOF is encountered before any digits, then FALSE is returned,
 * otherwise TRUE is returned.
 */

	static
	bool
get_coord (

coord_t *		coord_ptr	/* OUT - coordinate value. */
)
{
int		c;
int32u		val;
static bool	eof_flag = FALSE;

	*coord_ptr = 0;

	/* Check for EOF on previous call. */
	if (eof_flag) return (FALSE);

	do {
		c = getchar ();
		if (c < 0) {
			eof_flag = TRUE;
			return (FALSE);
		}
		if (ispunct (c)) {
			/* Strip comment line... */
			do {
				c = getchar ();
				if (c < 0) {
					eof_flag = TRUE;
					return (FALSE);
				}
			} while (c NE '\n');
			continue;
		}
	} while (NOT isdigit (c));

	val = 0;
	do {
		val = (val * 10) + (c - '0');
		c = getchar ();
		if (c < 0) {
			eof_flag = TRUE;
			break;
		}
	} while (isdigit (c));

	if (val > MAX_COORD) {
		(void) fprintf (stderr, "Coordinate out of range.\n");
		exit (1);
	}

	*coord_ptr = (coord_t) val;

	return (TRUE);
}
