/***********************************************************************

	File:	mst.c
	Rev:	a-1
	Date:	02/20/93

	Copyright (c) 1993, by The University of Virginia

************************************************************************

	Routines to compute Minimum Spanning Trees.

************************************************************************

	Modification Log:

	a-1:	02/20/93	dmw
		: Created.

************************************************************************/

#include "steiner.h"


/*
 * Global Routines
 */

int			kahng_robins (struct pset *, dist_t, struct edge *);
dist_t			kahng_robins_length (struct pset *, dist_t);
int			mst (struct pset *, struct edge *);


/*
 * External References
 */

	/* none */

/*
 * Local Equates
 */

#define	MAX_EDGES		((MAX_POINTS * (MAX_POINTS - 1)) / 2)
#define	MAX_UFIND		MAX_POINTS


/*
 * Local Routines
 */

static int		build_edges (struct pset *, struct edge *);
static int		find (int);
static void		initialize_union_find (int);
static dist_t		kr_main (struct pset *, dist_t);
static void		makeset (int);
static dist_t		mst_length (struct pset *);
static void		sort_edge_list (struct edge *, int);
static void		unite (int, int);


/*
 * Local Variables
 */


static int		parent [MAX_UFIND];
static int		rank   [MAX_UFIND];
static int		set_size;

/*
 * This routine computes a Minimum Spanning Tree for the given point set.
 * The output is a list of edges.
 */

	int
mst (

struct pset *		pts,		/* IN - point set. */
struct edge *		edges		/* OUT - edge list. */
)
{
int		i;
int		nedges;
int		n;
int		mst_edge_count;
int		components;
struct edge *	ep;
struct edge *	ep_endp;
int		root1;
int		root2;
struct edge	edge_array [MAX_EDGES];

	nedges = build_edges (pts, &edge_array [0]);

	sort_edge_list (&edge_array [0], nedges);

	n = pts -> n;

	initialize_union_find (n);
	for (i = 0; i < n; i++) {
		makeset (i);
	}

	components = n;
	mst_edge_count = 0;
	ep = &edge_array [0];
	ep_endp = (ep + nedges);

	while (components > 1) {
		if (ep >= ep_endp) {
			/* Ran out of edges before MST complete! */
			fatal ("mst: Bug 1.");
		}
		root1 = find (ep -> p1);
		root2 = find (ep -> p2);
		if (root1 NE root2) {
			unite (root1, root2);
			*edges = *ep;
			++edges;
			++mst_edge_count;
			--components;
		}
		++ep;
	}

	return (mst_edge_count);
}

/*
 * This routine builds an edge-list containing all pairs of points.
 */

	static
	int
build_edges (

struct pset *		pts,		/* IN - set of points. */
struct edge *		edges		/* OUT - edge list. */
)
{
int		i;
int		j;
int		n;
int		nedges;
struct point *	p1;
struct point *	p2;

	n = pts -> n;

	nedges = 0;
	p1 = &(pts -> a [0]);
	for (i = 0; i < n; i++) {
		p2 = (p1 + 1);
		for (j = i + 1; j < n; j++) {
			edges -> len	= DISTANCE (p1, p2);
			edges -> p1	= ((pnum_t) i);
			edges -> p2	= ((pnum_t) j);
			++edges;
			++nedges;
			++p2;
		}
		++p1;
	}

	return (nedges);
}

/*
 * This routine sorts the given edge list in INCREASING order by edge length.
 */

	void
sort_edge_list (

struct edge *		a,	/* IN/OUT - array of edges to be sorted. */
int			n	/* IN - number of elements in array. */
)
{
int		h;
struct edge	tmp;
dist_t		key;
struct edge *	p1;
struct edge *	p2;
struct edge *	p3;
struct edge *	p4;
struct edge *	endp;

	endp = &a [n];

	for (h = 1; h <= n; h = 3*h+1) {
	}

	do {
		h = h / 3;
		p4 = &a [h];
		p1 = p4;
		while (p1 < endp) {
			tmp = *p1;
			key = tmp.len;
			p2 = p1;
			while (TRUE) {
				p3 = (p2 - h);
				if (p3 -> len <= key) break;
				*p2 = *p3;
				p2 = p3;
				if (p2 < p4) break;
			}
			*p2 = tmp;
			++p1;
		}
	} while (h > 1);
}

/*
 * This routine initializes a collection of N disjoint sets.
 */

	static
	void
initialize_union_find (

int		n		/* IN - number of disjoint sets. */
)
{
int		i;

	set_size = n;

	if (n > MAX_UFIND) {
		/* Asking for too many disjoint sets. */
		fatal ("initialize_union_find: Bug 1.");
	}

	for (i = 0; i < MAX_UFIND; i++) {
		parent [i]	= 0;
		rank [i]	= 0;
	}
}

/*
 * This routine makes a single disjoint set for item "i".
 */

	static
	void
makeset (

int		i		/* IN - item to make into a disjoint set. */
)
{
	if ((i < 0) OR (i >= set_size)) {
		/* Item out of bounds. */
		fatal ("makeset: Bug 1.");
	}
	parent [i]	= i;
	rank [i]	= 0;
}

/*
 * This routine "unites" two sets that were previously disjoint.  I and J
 * must be the "canonical" member of each disjoint set (i.e. they must
 * each be the output of a "find" operation), and must be distinct.
 *
 * We perform the "union by rank" heuristic here.
 */

	static
	void
unite (

int		i,		/* IN - first set to unite. */
int		j		/* IN - second set to unite. */
)
{
int		ri;
int		rj;

	if ((i < 0) OR (i >= set_size)) {
		/* Item I is out of range. */
		fatal ("unite: Bug 1.");
	}
	if ((j < 0) OR (j >= set_size)) {
		/* Item J is out of range. */
		fatal ("unite: Bug 2.");
	}
	if (i EQ j) {
		/* Attempt to unite I with I. */
		fatal ("unite: Bug 3.");
	}

	ri = rank [i];
	rj = rank [j];

	if (ri EQ rj) {
		/* Both subtrees have the same maximum depth.  We	*/
		/* arbitrarily choose I to be underneath J.  The rank	*/
		/* of J must then increase.				*/
		parent [i] = j;
		rank [j]   = rj + 1;
	}
	else if (ri > rj) {
		/* Tree I is (probably) deeper.  Putting J underneath	*/
		/* will not increase I's rank.				*/
		parent [j] = i;
	}
	else {
		/* Tree J is (probably) deeper... */
		parent [i] = j;
	}
}

/*
 * This routine, given a member I of one of the disjoint sets A, will
 * choose a cannonical member J of set A and return it.  Until set A gets
 * united with some other set, find (I) will always return the same J.
 *
 * This routine performs the "path compression" heuristic.
 */

	static
	int
find (

int		i		/* IN - item to find cannonical item for. */
)
{
int		j;
int		k;

	/* Yes, I know this routine is very elegent when coded	*/
	/* recursively...  Here's the iterative version.	*/

	j = parent [i];
	if (i EQ j) {
		/* A cannonical element has itself as parent. */
		return (i);
	}

	/* We must search up the tree -- and compress when done... */
	while (TRUE) {
		k = parent [j];
		if (j EQ k) break;
		j = k;
	}

	/* Now compress the path (make all items in chain point directly */
	/* at the root K) -- we never have to do this search again!	 */
	while (i NE k) {
		j = parent [i];
		parent [i] = k;
		i = j;
	}

	return (k);
}

/*
 * This routine computes an approximate Steiner Minimal Tree using the
 * heuristic method of Kahng and Robins.
 *
 * Inputs:	pts	- the input point set -- MODIFIED to contain
 *			  additional Steiner points.
 *
 *		limit	- we quit if we ever get a tree shorter than this.
 *
 * Outputs:	edges	- The list of edges in the Steiner tree.
 *
 * Returns:		- Number of edges in Steiner tree.
 */

	int
kahng_robins (

struct pset *	pts,		/* IN/OUT - terminals and steiner points. */
dist_t		limit,		/* IN - quit if tree below this in length. */
struct edge *	edges		/* OUT - edge list for Steiner tree. */
)
{
int		nedges;

	/* Do the Kahng-Robins stuff -- ignore the resulting tree length... */
	(void) kr_main (pts, limit);

	nedges = mst (pts, edges);

	return (nedges);
}

/*
 * This routine computes the LENGTH of an approximate Steiner Minimal Tree
 * using the heuristic method of Kahng and Robins.  We use this routine
 * when we don't care what the tree looks like and we only want to know
 * its length.
 *
 * Inputs:	pts	- the input point set.
 *		limit	- we quit if we ever get a tree shorter than this.
 *
 * Returns:		- length of Steiner tree obtained.
 */

	dist_t
kahng_robins_length (

struct pset *	in_pts,		/* IN - terminals. */
dist_t		limit		/* IN - quit if tree below this in length. */
)
{
dist_t		len;
struct pset	pts_copy;

	/* Make a local copy of the point set that we can modify... */
	COPY_PSET (&pts_copy, in_pts);

	len = kr_main (&pts_copy, limit);

	return (len);
}

/*
 * This routine is the guts of the Kahng-Robins heuristic for computing
 * an approximate Steiner Minimal Tree.
 *
 * Inputs:	pts	- the input point set -- MODIFIED to contain
 *			  additional Steiner points.
 *
 *		limit	- we quit if we ever get a tree shorter than this.
 *
 * Returns:		- length of approximate Steiner Minimal Tree.
 */

	dist_t
kr_main (

struct pset *	pts,		/* IN/OUT - terminals and steiner points. */
dist_t		limit		/* IN - quit if tree below this in length. */
)
{
int		i;
int		j;
int		k;
int		m;
int		n;
int		nterm;
int		components;
dist_t		last;
dist_t		best;
dist_t		len;
int		best_j;
int		best_k;
int		nedges;
struct point *	p1;
struct point *	p2;
struct point *	p3;
struct point *	newpt;
struct edge *	ep;
struct edge *	endp;
struct edge *	ep1;
struct edge *	ep2;
struct edge *	endp1;
struct edge *	endp2;
int		root1;
int		root2;
int		nmask;
bitmap_t *	mark_rowp;
struct point	bestpt;
struct edge	new_edges [MAX_POINTS];
struct edge	edge_array [MAX_EDGES];
bitmap_t	mark [MAX_TERMINALS * N_TERM_BMAP];

	nterm = pts -> n;

	nedges = build_edges (pts, &edge_array [0]);

	sort_edge_list (&edge_array [0], nedges);

	initialize_union_find (nterm);
	for (i = 0; i < nterm; i++) {
		makeset (i);
	}

	components = nterm;
	ep = &edge_array [0];
	endp = ep + nedges;
	len = 0;

	while (components > 1) {
		if (ep >= endp) {
			/* Ran out of edges before MST built! */
			fatal ("kahng_robins: Bug 1.");
		}
		root1 = find (ep -> p1);
		root2 = find (ep -> p2);
		if (root1 NE root2) {
			unite (root1, root2);
			len += ep -> len;
			--components;
		}
		++ep;
	}
	best = len;

	nmask = BMAP_ELTS (nterm);
	memset ((char *) mark, 0, nterm * nmask * sizeof (bitmap_t));

	/* Mark the diagonal to dis-allow terminals as steiners. */
	mark_rowp = &mark [0];
	for (i = 0; i < nterm; i++) {
		SETBIT (mark_rowp, i);
		mark_rowp += nmask;
	}

	n = nterm;
	newpt = &(pts -> a [n]);
	for (i = 0; (i < nterm) AND (n < MAX_POINTS); i++) {
		last = best;
		newpt -> pnum = ((pnum_t) n);
		mark_rowp = &mark [0];
		p1 = &(pts -> a [0]);
		for (j = 0; j < nterm; mark_rowp += nmask, p1++, j++) {
			newpt -> x = p1 -> x;
			p2 = &(pts -> a [0]);
			for (k = 0; k < nterm; p2++, k++) {
				if (BITON (mark_rowp, k)) continue;
				newpt -> y = p2 -> y;
				ep2 = &new_edges [0];
				p3 = &(pts -> a [0]);
				for (m = 0; m < n; ep2++, p3++, m++) {
					ep2 -> p1	= n;
					ep2 -> p2	= m;
					ep2 -> len	= DISTANCE (newpt, p3);
				}
				sort_edge_list (&new_edges [0], n);

				initialize_union_find (n + 1);
				for (m = 0; m <= n; m++) {
					makeset (m);
				}

				components = n + 1;
				ep1	= &edge_array [0];
				endp1	= ep1 + nedges;
				ep2	= &new_edges [0];
				endp2	= ep2 + n;
				len	= 0;

				while (components > 1) {
					if (ep1 < endp1) {
						if (ep2 < endp2) {
							if (ep1 -> len < ep2 -> len) {
								ep = ep1++;
							}
							else {
								ep = ep2++;
							}
						}
						else {
							ep = ep1++;
						}
					}
					else if (ep2 < endp2) {
						ep = ep2++;
					}
					else {
						fatal ("kahng_robins: Bug 2.");
					}
					root1 = find (ep -> p1);
					root2 = find (ep -> p2);
					if (root1 NE root2) {
						unite (root1, root2);
						len += ep -> len;
						--components;
					}
				}
				if (len < best) {
					bestpt = *newpt;
					best = len;
					best_j = j;
					best_k = k;
					if (best < limit) break;
				}
			}
			if (best < limit) break;
		}
		if (best < limit) {
			/* Include this point in the output... */
			++n;
			break;
		}
		if (best >= last) break;

		/* Avoid trying this steiner point again... */
		SETBIT (&mark [j * nmask], k);

		/* Re-create edges for best point, and merge into */
		/* list of all edges, maintaining sorted order... */
		*newpt = bestpt;
		p3 = &(pts -> a [0]);
		ep = &new_edges [0];
		for (m = 0; m < n; p3++, ep++, m++) {
			ep -> p1	= n;
			ep -> p2	= m;
			ep -> len	= DISTANCE (newpt, p3);
		}
		sort_edge_list (&new_edges [0], n);
		ep1	= &edge_array [0];
		endp1	= ep1 + nedges;
		ep2	= &new_edges [0];
		endp2	= ep2 + n;
		ep	= endp1 + n;
		nedges += n;

		while (ep2 < endp2) {
			if (ep1 < endp1) {
				if (endp1 [-1].len >= endp2 [-1].len) {
					*--ep = *--endp1;
				}
				else {
					*--ep = *--endp2;
				}
			}
			else {
				*--ep = *--endp2;
			}
		}

		/* Edge list updated.  Now add best point to	*/
		/* the set and try for another.			*/
		++n;
		++newpt;
	}

	pts -> n = n;

	return (best);
}

/*
 * This routine computes the total length of the Minimum Spanning Tree
 * of the given set of points.
 */

	static
	dist_t
mst_length (

struct pset *		pts	/* IN - point set to find MST length of. */
)
{
int		i;
int		nedges;
dist_t		total;
struct edge *	ep;
struct edge	edges [MAX_POINTS - 1];

	nedges = mst (pts, &edges [0]);

	total = 0;
	ep = &edges [0];
	for (i = 0; i < nedges; i++) {
		total += ep -> len;
		++ep;
	}
	return (total);
}
