/***********************************************************************

	File:	output.c
	Rev:	a-1
	Date:	04/18/93

	Copyright (c) 1993, by The University of Virginia

************************************************************************

	Routines for outputting stuff.

************************************************************************

	Modification Log:

	a-1:	04/18/93	dmw
		: Created.  Moved most output stuff into this file
		:  from other places.

************************************************************************/

#include "steiner.h"


/*
 * Global Routines
 */

void		define_Plot_Terminals (struct pset *);
void		overlay_plot_innards (bitmap_t *, struct cinfo *);
void		overlay_plot_subset (char *, bitmap_t *, struct cinfo *);
void		plot_full_set_split_pair (int, int, bitmap_t *, struct cinfo *);
void		plot_full_sets (bitmap_t *, struct cinfo *);
void		plot_segment (struct point *, struct point *);
void		plot_split_full_sets (bitmap_t *, struct cinfo *);


/*
 * External References
 */

	/* none */


/*
 * Local Routines
 */

static void		plot_backbone (pnum_t, pnum_t, bool);
static void		plot_single_full_set (struct full_set *, bool);
static void		plot_term_steiner_segment (struct point *,
						   struct point *,
						   bool);

/*
 * This routine emits a series of Postscript definitions, one per
 * terminal.  The word is called Term_N where N is the terminal number.
 * The word pushes the terminal's X and Y coordinate -- in that order.
 *
 * It then emits another Postscript definition for the word Plot_Terminals
 * which causes all N terminals to be plotted and labeled.
 */

	void
define_Plot_Terminals (

struct pset *		pts		/* IN - terminals to plot */
)
{
int			i;
int			n;
struct point *		p1;

	n = pts -> n;
	p1 = &(pts -> a [0]);
	for (i = 0; i < n; i++, p1++) {
		(void) printf ("/Term_%lu {\t%lu\t%lu\t} def\n",
			       (int32u) i,
			       (int32u) (p1 -> x),
			       (int32u) (p1 -> y));
	}

	(void) printf ("\n/Plot_Terminals {\n");
	for (i = 0; i < n; i++) {
		(void) printf ("\t(%lu)\tTerm_%lu\tTerm\n",
			       (int32u) i,
			       (int32u) i);
	}
	(void) printf ("} def\n\n");
}

/*
 * This routine generates some Postscript commands to plot each full-set
 * in the given mask.
 */

	void
plot_full_sets (

bitmap_t *		fset_mask,	/* IN - subset of full-sets to plot */
struct cinfo *		cip		/* IN - compatibility info */
)
{
int			i;
int			n;
struct full_set *	fsp;

	n = cip -> full_tree_count;

	for (i = 0; i < n; i++) {
		if (NOT BITON (fset_mask, i)) continue;
		fsp = cip -> full_trees [i];
		(void) printf ("BeginSmallPlot\n");
		(void) printf ("\tPlot_Terminals\n");
		plot_single_full_set (fsp, FALSE);
		(void) printf ("  (Full Set %d,  Length = %lu)\n",
			       i, fsp -> tree_len);
		(void) printf ("EndSmallPlot2\n\n");
	}
	(void) printf ("FlushSmallPlot\n\n");
}

/*
 * This routine generates some Postscript commands to plot a SUBSET of
 * the given full-sets in overlaid fashion.
 */

	void
overlay_plot_subset (

char *			title,		/* IN - title to display with. */
bitmap_t *		fset_mask,	/* IN - subset of full-sets to plot */
struct cinfo *		cip		/* IN - compatibility info */
)
{
	(void) printf ("BeginPlot\n");
	(void) printf ("\tPlot_Terminals\n");

	overlay_plot_innards (fset_mask, cip);

	(void) printf ("  (%s)\n", title);
	(void) printf ("EndPlot\n\n");
}

/*
 * This routine is the "guts" of the overlay_plot stuff.
 */

	void
overlay_plot_innards (

bitmap_t *		fset_mask,	/* IN - subset of full-sets to plot */
struct cinfo *		cip		/* IN - compatibility info */
)
{
int			i;
int			n;
struct full_set *	fsp;

	n = cip -> full_tree_count;

	for (i = 0; i < n; i++) {
		if (NOT BITON (fset_mask, i)) continue;
		fsp = cip -> full_trees [i];
		plot_single_full_set (fsp, FALSE);
	}
}

/*
 * This routine plots each of the specified subset of the full-sets in
 * a fashion that reveals its efficacy as a "splitting piece."  That is,
 * it plots a piece X in bold, and plots every Y that is NOT incompatible
 * with X with light lines.  This clearly shows how X would split the
 * problem up...
 */

	void
plot_split_full_sets (

bitmap_t *		fset_mask,	/* IN - subset of full-sets to plot */
struct cinfo *		cip		/* IN - compatibility info */
)
{
int			i;
int			j;
int			n;
struct full_set *	fsp1;
struct full_set *	fsp2;
int8u *			carray;

	n = cip -> full_tree_count;

	for (i = 0; i < n; i++) {
		if (NOT BITON (fset_mask, i)) continue;
		fsp1 = cip -> full_trees [i];
		carray = &(cip -> carray [i * n]);

		(void) printf ("BeginSmallPlot\n");

		for (j = 0; j < n; j++) {
			if (NOT BITON (fset_mask, j)) continue;
			if (carray [j] EQ INCOMPATIBLE) continue;
			fsp2 = cip -> full_trees [j];
			plot_single_full_set (fsp2, TRUE);
		}
		plot_single_full_set (fsp1, FALSE);
		(void) printf ("\tPlot_Terminals\n");
		(void) printf ("  (Full Set %d,  Length = %lu)\n",
			       i, fsp1 -> tree_len);
		(void) printf ("EndSmallPlot2\n\n");
	}
	(void) printf ("FlushSmallPlot\n\n");
}

/*
 * This routine plots the segments of a single full-set.
 */

	static
	void
plot_single_full_set (

struct full_set *	fsp,		/* IN - full set to plot */
bool			plot_lightly	/* IN - TRUE ==> don't plot bold */
)
{
int			i;
int			n;
struct point *		p1;
struct point *		p2;
struct pset *		terms;
struct pset *		steins;
struct point		c;

	terms = fsp -> terminals;
	steins = fsp -> steiners;

	/* Put in a comment describing the full-set... */
	(void) printf ("  %%");
	n = terms -> n;
	p1 = &(terms -> a [0]);
	for (i = 0; i < n; p1++, i++) {
		(void) printf (" %d", p1 -> pnum);
	}
	(void) printf ("\n");

	p1 = &(terms -> a [0]);
	p2 = &(terms -> a [1]);

	plot_backbone (p1 -> pnum, p2 -> pnum, plot_lightly);

	n = terms -> n;
	p1 = &(terms -> a [2]);
	p2 = &(steins -> a [0]);
	for (i = 2; i < n; p1++, p2++, i++) {
		plot_term_steiner_segment (p1, p2, plot_lightly);
	}
}

/*
 * This routine plots a line segment between two points.
 */

	void
plot_segment (

struct point *		p1,	/* IN - first point. */
struct point *		p2	/* IN - second point. */
)
{
	(void) printf ("	%d	%d	%d	%d	Seg\n",
		       p1 -> x, p1 -> y, p2 -> x, p2 -> y);
}


/*
 * This routine plots a line segment between the given terminal,
 * and the given Steiner point.
 */

	static
	void
plot_term_steiner_segment (

struct point *		p1,		/* IN - first point. */
struct point *		p2,		/* IN - second point. */
bool			plot_lightly	/* IN - TRUE ==> don't plot bold */
)
{
	(void) printf ("\tTerm_%lu\t%lu\t%lu\t%s\n",
		       (int32u) (p1 -> pnum),
		       (int32u) (p2 -> x),
		       (int32u) (p2 -> y),
		       plot_lightly ? "LightSeg" : "Seg");
}


/*
 * This routine plots a backbone between two terminals.
 */

	static
	void
plot_backbone (

pnum_t		p1,		/* IN - first terminal. */
pnum_t		p2,		/* IN - second terminal. */
bool		plot_lightly	/* IN - TRUE ==> don't plot bold */
)
{
	(void) printf ("\tTerm_%lu\tTerm_%lu\t%s\n",
		       (int32u) p1,
		       (int32u) p2,
		       plot_lightly ? "LightBackBone2" : "BackBone2");
}

/*
 * This routine plots the two given full sets, together with all other
 * full-sets except those that are incompatible with either of them.
 */

	void
plot_full_set_split_pair (

int			I,		/* IN - first full set number */
int			J,		/* IN - second full set number */
bitmap_t *		fset_mask,	/* IN - subset of full sets */
struct cinfo *		cip		/* IN - compatibility info */
)
{
int			k;
int			nft;
int8u *			cp1;
int8u *			cp2;

	nft = cip -> full_tree_count;

	cp1 = &(cip -> carray [I*nft]);
	cp2 = &(cip -> carray [J*nft]);

	(void) printf ("BeginSmallPlot\n");

	for (k = 0; k < nft; k++) {
		if (NOT BITON (fset_mask, k)) continue;
		if (cp1 [k] EQ INCOMPATIBLE) continue;
		if (cp2 [k] EQ INCOMPATIBLE) continue;
		plot_single_full_set (cip -> full_trees [k], TRUE);
	}
	plot_single_full_set (cip -> full_trees [I], FALSE);
	plot_single_full_set (cip -> full_trees [J], FALSE);
	(void) printf ("\tPlot_Terminals\n");
	(void) printf ("  (Full Sets %d and %d together)\n", I, J);
	(void) printf ("EndSmallPlot2\n\n");
}
