/***********************************************************************

	File:	relief.c
	Rev:	a-1
	Date:	07/28/94

	Copyright (c) 1994, by The University of Virginia

************************************************************************

	Routines to print out a "Topographical Relief Map" of a
	given subset of the full sets.

************************************************************************

	Modification Log:

	a-1:	07/28/94	dmw
		: Created.

************************************************************************/

#include "steiner.h"


/*
 * Global Routines
 */

void		relief_plot_full_sets (char *, bitmap_t *, struct cinfo *);


/*
 * External References
 */

	/* none */


/*
 * Local Types
 */

enum ev_type {
	RECT_START,
	RECT_END
};

struct event {
	coord_t		x;
	enum ev_type	type;
	coord_t		x0;	/* Lower left X */
	coord_t		y0;	/* Lower left Y */
	coord_t		x1;	/* Upper right X */
	coord_t		y1;	/* Upper right Y */
	struct event *	next;
};

struct status {
	coord_t		y0;
	coord_t		y1;
	coord_t		x0;
	int		thickness;
	struct status *	next;
};

struct rect {
	coord_t		x0;
	coord_t		x1;
	coord_t		y0;
	coord_t		y1;
	int		thickness;
	struct rect *	next;
};


/*
 * Local Routines
 */

static struct status *		cleanup_status (struct status *);
static struct status *		delete_status (coord_t,
					       coord_t,
					       coord_t,
					       struct status *);
static struct event *		fset_rect (struct full_set *);
static struct event *		insert_event (struct event *, struct event *);
static struct status *		merge_status (struct status *,
					      struct status *, coord_t);
static void			prect (coord_t,
				       coord_t,
				       coord_t,
				       coord_t,
				       int);


/*
 * Local Variables
 */

static struct rect *	rlist;
static int		max_thick;

/*
 * This routine plots the given subset of the full-sets in a "Topographical
 * Relief Map" format.  Full sets whose backbone has a bend in it (i.e.
 * almost every full set) actually have an infinite number of different
 * full topologies, the so-called "second topology" being one of them.  If
 * we overlay the 1st and 2nd topology, we see a rectangle in the middle
 * such that ANY point within the rectangle is a member of SOME full-topology
 * for the set.
 *
 * We plot the rectangle for each specified full-set in an overlaid
 * fashion.  Regions where K of these rectangles overlap are plotted
 * darker than regions where K-1 or fewer overlap.  This results in
 * a plot wherein "busy" sections of the problem appear dark, and
 * places with fewer alternatives show up lighter.  Such plots resemble
 * classic topographical relief maps.
 */

	void
relief_plot_full_sets (

char *			title,		/* IN - title for plot */
bitmap_t *		fset_mask,	/* IN - subset of full-sets to plot */
struct cinfo *		cip		/* IN - compatibility info */
)
{
int			i;
struct full_set *	fsp;
struct event *		ev;
struct event *		root;
struct status *		stroot;
struct status *		sp1;
coord_t			cur_x;
struct rect *		rp;

	root = NULL;
	for (i = 0; i < cip -> full_tree_count; i++) {
		if (NOT BITON (fset_mask, i)) continue;
		fsp = cip -> full_trees [i];
		ev = fset_rect (fsp);
		if (ev NE NULL) {
			root = insert_event (ev, root);
		}
	}

	/* Begin the sweep-line algorithm to overlay all rectangles... */
	stroot = NULL;
	ev = root;
	rlist = NULL;
	max_thick = 0;
	while (ev NE NULL) {
		/* Examine next event in direction of sweep */
		cur_x = ev -> x;
#if 0
printf ("---------------------------------------------------------------\n");
print_status ("Initial Status", stroot, cur_x);
print_event ("Current Event", ev);
#endif
		if (ev -> type EQ RECT_START) {
			/* Construct a new rectangle status... */
			sp1 = NEW (struct status);
			sp1 -> y0		= ev -> y0;
			sp1 -> y1		= ev -> y1;
			sp1 -> x0		= cur_x;
			sp1 -> thickness	= 1;
			sp1 -> next		= NULL;
			stroot = merge_status (sp1, stroot, cur_x);
			ev -> x		= ev -> x1;
			ev -> type	= RECT_END;
			ev = insert_event (ev, ev -> next);
#if 0
print_status ("After Adding Rectangle", stroot, cur_x);
#endif
		}
		else {
			stroot = delete_status (ev -> y0,
						ev -> y1,
						cur_x,
						stroot);
			root = ev;
			ev = ev -> next;
			free ((char *) root);
#if 0
print_status ("After Deleting Rectangle", stroot, cur_x);
#endif
		}
		stroot = cleanup_status (stroot);
#if 0
print_status ("After Cleanup", stroot, cur_x);
#endif
	}

	/* All rectangles have been generated.  Now plot them! */
	(void) printf ("BeginPlot\n");
	while (rlist NE NULL) {
		rp = rlist;
		rlist = rp -> next;

		(void) printf ("\t%lu\t%lu\t%lu\t%lu\t%lu\t%lu\tShadedRect\n",
			       (int32u) (rp -> thickness),
			       (int32u) max_thick,
			       (int32u) rp -> x0,
			       (int32u) rp -> y0,
			       (int32u) rp -> x1,
			       (int32u) rp -> y1);
		free ((char *) rp);
	}

	overlay_plot_innards (fset_mask, cip);

	(void) printf ("\tPlot_Terminals\n");
	(void) printf ("  (%s)\n", title);
	(void) printf ("EndPlot\n");
}

/*
 * This routine finds the rectangle (if any) for the given full-set.
 */

	static
	struct event *
fset_rect (

struct full_set *	fsp	/* IN - input full-set */
)
{
int			n;
struct point *		terms;		/* input terminals. */
coord_t			x0;
coord_t			y0;
coord_t			x1;
coord_t			y1;
coord_t			tmp;
struct event *		rp;

	n = fsp -> terminals -> n;
	terms	= &(fsp -> terminals -> a [0]);

	/* First rectangle corner is always the backbone corner... */
	x0 = terms [0].x;
	y0 = terms [1].y;

	switch (fsp -> style) {
	case ZERO_H_REST_V:
		if ((n % 2) EQ 0) {
			x1 = terms [1].x;
			y1 = terms [0].y;
		}
		else {
			x1 = terms [1].x;
			y1 = terms [n-1].y;
		}
		break;

	case ONE_H_REST_V:
		if ((n % 2) EQ 0) {
			x1 = terms [2].x;
			y1 = terms [n-1].y;
		}
		else {
			x1 = terms [2].x;
			y1 = terms [0].y;
		}
		break;

	case ZERO_V_REST_H:
		if ((n % 2) EQ 0) {
			x1 = terms [1].x;
			y1 = terms [0].y;
		}
		else {
			x1 = terms [n-1].x;
			y1 = terms [0].y;
		}
		break;

	case ONE_V_REST_H:
		if ((n % 2) EQ 0) {
			x1 = terms [n-1].x;
			y1 = terms [2].y;
		}
		else {
			x1 = terms [1].x;
			y1 = terms [2].y;
		}
		break;

	case ZERO_H_ZERO_V:
		if (n NE 2) {
			fatal ("fset_rect: Bug 1.");
		}
		x1 = terms [1].x;
		y1 = terms [0].y;
		/* Backbones that are stricly horizontal or vertical	*/
		/* are valid here, but have no rectangle!		*/
		if ((x0 EQ x1) OR (y0 EQ y1)) return (NULL);
		break;

	case ONLY_H:
	case ONLY_V:
	case CROSS:
		/* These degenerate cases do not have rectangles... */
		return (NULL);

	default:
		fatal ("fset_rect: Bug 2.");
		return (NULL);
	}

	if ((x0 EQ x1) OR (y0 EQ y1)) {
		/* Should only happen on ONLY_H or ONLY_V backbones... */
		fatal ("fset_rect: Bug 3.");
	}

	/* Swap coordinates (if necessary) so that (x0,y0) is the	*/
	/* lower-left corner, and (x1,y1) is upper-right.		*/
	if (x0 > x1) {
		tmp = x0;
		x0 = x1;
		x1 = tmp;
	}
	if (y0 > y1) {
		tmp = y0;
		y0 = y1;
		y1 = tmp;
	}

	rp = NEW (struct event);

	rp -> x		= x0;
	rp -> type	= RECT_START;
	rp -> x0	= x0;
	rp -> y0	= y0;
	rp -> x1	= x1;
	rp -> y1	= y1;
	rp -> next	= NULL;

	return (rp);
}

/*
 * This routine inserts the given event into the given list of events.
 */

	static
	struct event *
insert_event (

struct event *	evp,		/* IN - event to insert */
struct event *	list		/* IN - list to insert it into */
)
{
struct event *	root;
struct event **	hookp;
struct event *	p;

	root = list;
	hookp = &root;
	while ((p = *hookp) NE NULL) {
		if (p -> x >= evp -> x) break;
		hookp = &(p -> next);
	}
	evp -> next = p;
	*hookp = evp;

	return (root);
}

/*
 * This routine merges the given single status entry into the list of
 * all current status entries.  This current list shows the status along
 * the sweep-line going from minimum to maximum Y value.
 */

	static
	struct status *
merge_status (

struct status *		sp,	/* IN - status entry to merge */
struct status *		list,	/* IN - list to merge into */
coord_t			cur_x	/* IN - current sweep line X coord */
)
{
struct status *		root;
struct status **	hookp;
struct status *		p;
struct status *		np;

	root = list;
	hookp = &root;
	while ((p = *hookp) NE NULL) {
		if (p -> y1 <= sp -> y0) {
			/* entry "p" is disjoint from entry "sp" */
			hookp = &(p -> next);
			continue;
		}
		if (sp -> y1 <= p -> y0) {
			/* Entry "sp" comes entirely before entry "p". */
			break;
		}
		if (sp -> y0 < p -> y0) {
			/* There is a slice of "sp" that must come	*/
			/* before "p".  Insert it right here.		*/
			np = NEW (struct status);
			np -> y0	= sp -> y0;
			np -> y1	= p -> y0;
			np -> x0	= sp -> x0;
			np -> thickness	= sp -> thickness;
			np -> next	= p;
			*hookp = np;
			hookp = &(np -> next);
			sp -> y0 = p -> y0;
		}
		else if (sp -> y0 > p -> y0) {
			/* There is a slice of "p" that must come	*/
			/* before "sp".  Insert it right here.		*/
			np = NEW (struct status);
			np -> y0	= p -> y0;
			np -> y1	= sp -> y0;
			np -> x0	= p -> x0;
			np -> thickness	= p -> thickness;
			np -> next	= p;
			*hookp = np;
			hookp = &(np -> next);
			p -> y0 = sp -> y0;
		}
		/* At this point, entries "p" and "sp" begin at the	*/
		/* same Y coordinate (p->y0 EQ sp->y0), and overlap by	*/
		/* a non-zero amount.					*/
		if (sp -> y1 < p -> y1) {
			prect (p -> x0,
			       cur_x,
			       sp -> y0,
			       sp -> y1,
			       p -> thickness);
			sp -> next = p;
			*hookp = sp;
			p -> y0 = sp -> y1;
			sp -> thickness += p -> thickness;
			return (root);
		}
		else if (sp -> y1 > p -> y1) {
			prect (p -> x0,
			       cur_x,
			       p -> y0,
			       p -> y1,
			       p -> thickness);
			p -> x0		= cur_x;
			p -> thickness	+= sp -> thickness;
			sp -> y0 = p -> y1;
			hookp = &(p -> next);
		}
		else {
			prect (p -> x0,
			       cur_x,
			       p -> y0,
			       p -> y1,
			       p -> thickness);
			p -> thickness += sp -> thickness;
			free ((char *) sp);
			return (root);
		}
	}
	if (sp -> y0 >= sp -> y1) {
		/* infinitely thin -- nothing to insert! */
		free ((char *) sp);
	}
	else {
		sp -> next = p;
		*hookp = sp;
	}

	return (root);
}

/*
 * This routine outputs a single rectangular region of the topographical
 * relief map.
 */

	static
	void
prect (

coord_t		x0,		/* IN - min X coord. */
coord_t		x1,		/* IN - max X coord. */
coord_t		y0,		/* IN - min Y coord. */
coord_t		y1,		/* IN - max X coord. */
int		thickness	/* IN - topographical "altitude" */
)
{
struct rect *	rp;

	if ((thickness > 0) AND (x0 < x1) AND (y0 < y1)) {
#if 0
printf ("prect: X=[%4lu,%4lu] Y=[%4lu,%4lu] Thick=%d\n",
	(int32u) x0, (int32u) x1, (int32u) y0, (int32u) y1, thickness);
#endif
		rp = NEW (struct rect);
		rp -> x0	= x0;
		rp -> x1	= x1;
		rp -> y0	= y0;
		rp -> y1	= y1;
		rp -> thickness	= thickness;
		rp -> next	= rlist;
		rlist = rp;
		if (max_thick < thickness) {
			max_thick = thickness;
		}
	}
}

/*
 * This routine cleans up the status list, deleting empty elements,
 * and merging adjacent equivalent ones.
 */

	static
	struct status *
cleanup_status (

struct status *	list		/* IN - status list to clean up */
)
{
struct status *		root;
struct status **	hookp;
struct status *		p1;
struct status *		p2;

	root = list;
	hookp = &root;

	while ((p1 = *hookp) NE NULL) {
		if ((p1 -> thickness <= 0) OR (p1 -> y0 >= p1 -> y1)) {
			/* entry describes no rectangles -- delete! */
			*hookp = p1 -> next;
			free ((char *) p1);
			continue;
		}
		p2 = p1 -> next;
		if (p2 EQ NULL) break;
		if ((p1 -> y1 EQ p2 -> y0) AND
		    (p1 -> x0 EQ p2 -> x0) AND
		    (p1 -> thickness EQ p2 -> thickness)) {
			p2 -> y0 = p1 -> y0;
			*hookp = p2;
			free ((char *) p1);
			continue;
		}
		/* Entry p1 is okay -- advance to next. */
		hookp = &(p1 -> next);
	}

	return (root);
}

/*
 * This routine removes the effects of an "outgoing" rectangle from the
 * current status.  This automatically forces us to emit at least one
 * rectangle to the output.  The status is also updated appropriately.
 */

	static
	struct status *
delete_status (

coord_t		y0,		/* IN - min Y coord of outgoing rectangle. */
coord_t		y1,		/* IN - max Y coord of outgoing rectangle. */
coord_t		cur_x,		/* IN - current X coordinate.		   */
struct status *	list		/* IN - current status list.		   */
)
{
struct status *		root;
struct status **	hookp;
struct status *		p;
struct status *		np;

	root = list;
	hookp = &root;
	while (y0 < y1) {
		p = *hookp;
		if (p EQ NULL) break;
		if (p -> y1 <= y0) {
			/* entry "p" is disjoint from outgoing rect. */
			hookp = &(p -> next);
			continue;
		}
		if (y1 <= p -> y0) {
			/* Outgoing rectangle comes entirely before	*/
			/* entry "p".  This should never happen!	*/
			fatal ("delete_status: Bug 1.");
		}
		/* Outgoing rectangle overlaps entry "p". */
		if (y0 < p -> y0) {
			/* There is a slice of the outgoing rectangle	*/
			/* that comes before entry "p".  This should	*/
			/* never happen...				*/
			fatal ("delete_status: Bug 2.");
		}
		else if (y0 > p -> y0) {
			/* There is a slice of entry "p" that must come	*/
			/* before this outgoing rectangle.  Insert a	*/
			/* new entry for it here.			*/
			np = NEW (struct status);
			np -> y0	= p -> y0;
			np -> y1	= y0;
			np -> x0	= p -> x0;
			np -> thickness	= p -> thickness;
			np -> next	= p;
			*hookp = np;
			hookp = &(np -> next);
			p -> y0 = y0;
		}
		/* At this point, entry "p" and the outgoing rectangle	*/
		/* begin at the same Y coordinate (p->y0 EQ y0), and	*/
		/* overlap by a non-zero amount.			*/
		if (y1 < p -> y1) {
			prect (p -> x0,
			       cur_x,
			       y0,
			       y1,
			       p -> thickness);
			np = NEW (struct status);
			np -> y0	= y0;
			np -> y1	= y1;
			np -> x0	= cur_x;
			np -> thickness	= p -> thickness - 1;
			np -> next	= p;
			*hookp = np;
			hookp = &(np -> next);
			p -> y0 = y1;
			break;
		}
		if (y1 > p -> y1) {
			prect (p -> x0,
			       cur_x,
			       p -> y0,
			       p -> y1,
			       p -> thickness);
			p -> x0		= cur_x;
			--(p -> thickness);
			y0 = p -> y1;
			hookp = &(p -> next);
		}
		else {
			prect (p -> x0,
			       cur_x,
			       p -> y0,
			       p -> y1,
			       p -> thickness);
			p -> x0		= cur_x;
			--(p -> thickness);
			break;
		}
	}

	return (root);
}

print_event (title, ev)

char *		title;
struct event *	ev;
{
	printf ("%s: %5s  cur_x=%4lu  X=[%4lu,%4lu]  Y=[%4lu,%4lu]\n",
		title,
		(ev -> type EQ RECT_START) ? "START" : "END  ",
		(int32u) (ev -> x),
		(int32u) (ev -> x0),
		(int32u) (ev -> x1),
		(int32u) (ev -> y0),
		(int32u) (ev -> y1));
}

print_status (title, sp, cur_x)

char *		title;
struct status *	sp;
coord_t		cur_x;
{
	if (sp EQ NULL) {
		printf ("%s: EMPTY\n", title);
		return;
	}
	printf ("%s:\n", title);
	while (sp NE NULL) {
		printf ("\tX=[%4lu,%4lu], Y=[%4lu,%4lu], Thick=%d\n",
			(int32u) sp -> x0,
			(int32u) cur_x,
			(int32u) sp -> y0,
			(int32u) sp -> y1,
			(int32u) sp -> thickness);
		sp = sp -> next;
	}
}
