/***********************************************************************

	File:	search.c
	Rev:	a-1
	Date:	03/25/93

	Copyright (c) 1993, by The University of Virginia

************************************************************************

	Routines to perform the backtrack search for a solution.

************************************************************************

	Modification Log:

	a-1:	03/25/93	dmw
		: Created.

************************************************************************/

#include "steiner.h"


/*
 * Global Routines
 */

bool			backtrack_search (bitmap_t *,
					  bitmap_t *,
					  struct cinfo *,
					  bitmap_t *);
struct full_set *	sort_full_trees (struct full_set *);


/*
 * External References
 */

	/* none */

/*
 * Local Routines
 */

static void			allocate_search_stacks (struct cinfo *);
static bool			better_split (int, int *, int *,
					      int, int *, int *);
static void			check_piece_splits (bitmap_t *,
						    bitmap_t *,
						    struct cinfo *);
static void			do_piece_splits (bitmap_t *,
						 bitmap_t *,
						 struct cinfo *);
static int			find_components (bitmap_t *, bitmap_t *,
						 int *, int *,
						 int *, bitmap_t *,
						 struct cinfo *);
static bool			find_inaccessible_terminal (struct cinfo *,
							    int);
static pnum_t			find_starting_point (bitmap_t *,
						     struct cinfo *);
static void			free_search_stacks (struct sinfo *);
static void			handle_subproblems (int, int,
						    int *, int *, int *, int *,
						    bitmap_t *, bitmap_t *,
						    struct cinfo *);
static void			perform_search (bitmap_t *,
						bitmap_t *,
						struct cinfo *);
static void			search_recurse (struct cinfo *,
						int, int, dist_t);
static void			sort_components (int, int *, int *);
static void			walk (int, bitmap_t *, int *, int *,
				      bitmap_t *, bitmap_t *,
				      int *, struct full_set **,
				      int **);

/*
 * This routine performs a backtrack search for a Steiner Minimal Tree of
 * the given point set.  It composes a minimum-length solution from some
 * combination of the given full-sets.
 */

	bool
backtrack_search (

bitmap_t *		tmap,		/* IN - subset of terminals to use */
bitmap_t *		fset_mask,	/* IN - subset of full-sets to use */
struct cinfo *		cip,		/* IN - compatibility info */
bitmap_t *		smt_mask	/* OUT - SMT as a set of full-sets */
)
{
int			i;
int			nmasks;
int			kmasks;
int			term_count;
int			tree_count;
bitmap_t *		bp1;
bitmap_t *		bp2;
bitmap_t		word;
struct sinfo *		sip;
bool			failed;

	nmasks	= cip -> num_fset_masks;
	kmasks	= cip -> num_term_masks;

	/* First, zero out the result bit-mask... */
	bp1 = smt_mask;
	for (i = 0; i < nmasks; i++) {
		*bp1++ = 0;
	}

	/* Count the number of terminals we must connect... */
	term_count = 0;
	bp1 = tmap;
	for (i = 0; i < kmasks; i++) {
		word = *bp1++;
		term_count += NBITSON (word);
	}
	if (term_count <= 1) return (FALSE);

	/* Count the number of full-sets we have... */
	tree_count = 0;
	bp1 = fset_mask;
	for (i = 0; i < nmasks; i++) {
		word = *bp1++;
		tree_count += NBITSON (word);
	}

	(void) printf (" %% Entering search module with %d terms and %d trees\n",
		       term_count, tree_count);

#if 0
	print_mask ("  % terminals: ", tmap, cip -> num_terms);
	print_mask ("  % full-sets: ", fset_mask, cip -> full_tree_count);
#endif

	/* Create a worst-case number of search-context structures... */
	sip = NEWA ((tree_count + 1), struct sinfo);
	cip -> sinfo_ptr = sip;

	/* Initialize so that no memory is allocated to any of	*/
	/* the context structures yet...			*/
	(void) memset (sip, 0, ((tree_count + 1) * sizeof (*sip)));

	/* Initialize first search context... */
	sip -> best_length		= INF_DISTANCE;
	sip -> best_count		= 0;
	sip -> best_solution		= NULL;
	sip -> found			= FALSE;

	if (tree_count <= Piece_Split_Threshold) {
		/* Problem is too small to bother with piece-splits... */
		perform_search (tmap, fset_mask, cip);
	}
	else {
		/* Big search.  Let's try breaking it up... */
		do_piece_splits (tmap, fset_mask, cip);
	}

	if (sip -> found) {
		bp1 = smt_mask;
		bp2 = sip -> best_solution;
		for (i = 0; i < nmasks; i++) {
			*bp1++ = *bp2++;
		}

#if 0
		print_mask ("  % search solution: ", smt_mask, cip -> full_tree_count);
		(void) printf ("  %% length = %ld\n",
			       total_tree_length (smt_mask, cip));
#endif
		failed = FALSE;
	}

	else {
		(void) printf ("  %% NO SOLUTION FOUND!\n");
		failed = TRUE;
	}

	(void) printf (" %% Exiting search module\n");

	/* Free up the search stacks in reverse order we allocated them. */
	while (sip -> solution NE NULL) {
		++sip;
	}
	while (--sip >= cip -> sinfo_ptr) {
		free_search_stacks (sip);
	}

	free ((char *) cip -> sinfo_ptr);
	cip -> sinfo_ptr = NULL;

	return (failed);
}

/*
 * This routine pre-sorts the list of full-sets and re-numbers them so that
 * at each node in the search tree, we select candidate sub-trees in an
 * order that heuristically puts the most likely solutions first.  This has
 * a tendency to reduce the search space by producing earlier length cutoffs
 * on the later sub-trees.
 */

	struct full_set *
sort_full_trees (

struct full_set *	fsp		/* IN - full-trees to sort. */
)
{
int			i;
int			n;
int			h;
struct full_set *	tmp;
struct full_set **	trees;
struct full_set **	endp;
struct full_set **	p1;
struct full_set **	p2;
struct full_set **	p3;
struct full_set **	p4;
dist_t			dist_1;
dist_t			dist_2;
int			n_1;
int			n_2;
struct full_set *	root;
struct full_set **	hookp;

	trees = put_trees_in_array (fsp, &n);

	endp = trees + n;

	for (h = 1; h <= n; h = 3*h+1) {
	}

	do {
		h = h / 3;
		p4 = trees + h;
		p1 = p4;
		while (p1 < endp) {
			tmp = *p1;
			dist_1 = tmp -> tree_len;
			n_1	= tmp -> terminals -> n - 1;
			p2 = p1;
			while (TRUE) {
				p3 = (p2 - h);
				dist_2	= (*p3) -> tree_len;
				n_2	= (*p3) -> terminals -> n - 1;
				if ((dist_2 * n_1) <= (dist_1 * n_2)) break;
				*p2 = *p3;
				p2 = p3;
				if (p2 < p4) break;
			}
			*p2 = tmp;
			++p1;
		}
	} while (h > 1);

	i = 0;
	root = NULL;
	hookp = &root;
	p1 = trees;
	while (p1 < endp) {
		tmp = *p1++;
		tmp -> tree_num = i++;
		tmp -> next	= NULL;
		*hookp = tmp;
		hookp = &(tmp -> next);
	}

	return (root);
}

/*
 * This routine checks to see if the problem is big enough so that we should
 * try using the piece-splits trick.  If not, we skip it.
 */

	static
	void
check_piece_splits (

bitmap_t *		tmap,		/* IN - subset of terminals to use */
bitmap_t *		fset_mask,	/* IN - subset of full-sets to use */
struct cinfo *		cip		/* IN - compatibility info */
)
{
int			i;
int			nmasks;
int			count;
bitmap_t *		bp1;
bitmap_t		word;

	nmasks	= cip -> num_fset_masks;

	count = 0;
	bp1 = fset_mask;
	for (i = 0; i < nmasks; i++) {
		word = *bp1++;
		count += NBITSON (word);
	}

	if (count <= Piece_Split_Threshold) {
		/* Problem is too small to bother with this trick! */
		perform_search (tmap, fset_mask, cip);
	}
	else {
		/* Big search.  Let's try breaking it up... */
		do_piece_splits (tmap, fset_mask, cip);
	}
}

/*
 * This routine handles the top-most level of the search.
 */

	static
	void
perform_search (

bitmap_t *		tmap,		/* IN - subset of terminals to use */
bitmap_t *		fset_mask,	/* IN - subset of full-sets to use */
struct cinfo *		cip		/* IN - compatibility/search info */
)
{
int			i;
int			j;
int			nft;
int			nmasks;
int			kmasks;
int			count;
int			nterms;
bitmap_t *		bp1;
bitmap_t *		bp2;
bitmap_t *		bp3;
bitmap_t *		bp4;
pnum_t			first_point;
struct sinfo *		sip;
int			t;
int *			tp1;
int *			endp;
struct point *		p1;
struct full_set *	fsp;
struct full_set **	full_trees;
size_t			size;

	nft		= cip -> full_tree_count;
	nmasks		= cip -> num_fset_masks;
	kmasks		= cip -> num_term_masks;

#if 0
	count = 0;
	bp1 = fset_mask;
	bp2 = bp1 + nmasks;
	while (bp1 < bp2) {
		count += NBITSON (*bp1);
		++bp1;
	}
	(void) printf ("  %%  Doing exhaustive search on %d full-trees\n",
		       count);
#endif

	/* Get current search context. */
	sip = cip -> sinfo_ptr;

	if (sip -> solution EQ NULL) {
		/* Haven't been this deep before.  Allocate new	*/
		/* search stacks, as large as we could EVER	*/
		/* need again at this same level...		*/
		allocate_search_stacks (cip);
	}

	bp1 = sip -> terms_left;
	bp2 = tmap;
	for (j = 0; j < kmasks; j++) {
		*bp1++ = *bp2++;
	}

	nterms = 0;
	bp1 = sip -> terms_left;
	bp2 = bp1 + kmasks;
	while (bp1 < bp2) {
		nterms += NBITSON (*bp1);
		++bp1;
	}

	size = (nmasks * sizeof (bitmap_t));
	(void) memset (sip -> compat, 0, size);
	(void) memset (sip -> incompat, 0, size);
	(void) memset (sip -> try, 0, size);
	(void) memset (sip -> omit, 0, size);
	(void) memset (sip -> solution, 0, size);

	full_trees = cip -> full_trees;

	/* Do the search once for each full-tree that contains the */
	/* starting point...					   */
	first_point = find_starting_point (fset_mask, cip);

	tp1	= cip -> term_trees [first_point];
	endp	= cip -> term_trees [first_point + 1];
	while (tp1 < endp) {
		t = *tp1++;

		/* Skip trees that are not part of this component... */
		if (NOT BITON (fset_mask, t)) continue;

		fsp = full_trees [t];
		SETBIT (sip -> solution, t);

		bp1 = sip -> terms_left;
		bp2 = &(fsp -> tmap [0]);
		bp3 = bp1 + kmasks;
		for (i = 0; i < kmasks; i++) {
			*bp3++ = *bp1++ & ~*bp2++;
		}

		bp1 = &(cip -> cmasks [t * nmasks]);
		bp2 = sip -> compat + nmasks;
		for (i = 0; i < nmasks; i++) {
			*bp2++ = *bp1++;
		}

		bp1 = &(cip -> incmasks [t * nmasks]);
		bp2 = sip -> omit;
		bp3 = fset_mask;
		bp4 = sip -> incompat + nmasks;
		for (i = 0; i < nmasks; i++) {
			*bp4++ = *bp1++ | *bp2++ | ~*bp3++;
		}

		search_recurse (cip,
				1,
				nterms - fsp -> terminals -> n,
				fsp -> tree_len);

		CLRBIT (sip -> solution, t);
		SETBIT (sip -> omit, t);
	}
}

/*
 * This routine allocates all of the search stacks needed.
 */

	static
	void
allocate_search_stacks (

struct cinfo *		cip		/* IN - compatibility/search info. */
)
{
int		nft;
int		nmasks;
int		kmasks;
int		n;
struct sinfo *	sip;

	nft	= cip -> full_tree_count;
	nmasks	= cip -> num_fset_masks;
	kmasks	= cip -> num_term_masks;
	sip	= cip -> sinfo_ptr;

	sip -> best_solution	= NEWA (nmasks, bitmap_t);
	sip -> solution		= NEWA (nmasks, bitmap_t);

	n = (nft + 1) * kmasks;

	sip -> terms_left	= NEWA (n, bitmap_t);

	n = (nft + 1) * nmasks;

	sip -> compat		= NEWA (n, bitmap_t);
	sip -> incompat		= NEWA (n, bitmap_t);
	sip -> try		= NEWA (n, bitmap_t);
	sip -> omit		= NEWA (n, bitmap_t);
}

/*
 * This routine frees up the search stacks.
 */

	static
	void
free_search_stacks (

struct sinfo *		sip		/* IN - search info. */
)
{
	free ((char *) sip -> omit);
	free ((char *) sip -> try);
	free ((char *) sip -> incompat);
	free ((char *) sip -> compat);
	free ((char *) sip -> terms_left);
	free ((char *) sip -> solution);
	free ((char *) sip -> best_solution);

	sip -> best_solution	= NULL;
	sip -> solution		= NULL;
	sip -> terms_left	= NULL;
	sip -> compat		= NULL;
	sip -> incompat		= NULL;
	sip -> try		= NULL;
	sip -> omit		= NULL;
}

/*
 * This routine selects a single terminal to use as a starting point in
 * the search.  We choose a point that is a member of the least number
 * of full-sets.
 */

	static
	pnum_t
find_starting_point (

bitmap_t *		fset_mask,	/* IN - full-sets to search over. */
struct cinfo *		cip		/* IN - compatibility/search info. */
)
{
int			i;
int			j;
int			n;
int			nft;
struct point *		p1;
struct full_set *	fsp;
struct pset *		terms;
int			cnt;
int			fewest;
pnum_t			starting;
int32u			counts [MAX_TERMINALS];

	(void) memset (counts, 0, sizeof (counts));

	nft = cip -> full_tree_count;

	for (i = 0; i < nft; i++) {
		if (BITON (fset_mask, i)) {
			fsp = cip -> full_trees [i];
			terms = fsp -> terminals;
			n = terms -> n;
			p1 = &(terms -> a [0]);
			for (j = 0; j < n; p1++, j++) {
				++(counts [p1 -> pnum]);
			}
		}
	}

	starting = MAX_TERMINALS;
	fewest	 = INT_MAX;
	for (i = 0; i < MAX_TERMINALS; i++) {
		cnt = counts [i];
		if (cnt <= 0) {
			/* Ignore terminals that are not part	*/
			/* of this sub-problem...		*/
		}
		else if (cnt < fewest) {
			starting = i;
			fewest = cnt;
		}
	}

	if (starting >= MAX_TERMINALS) {
		fatal ("find_starting_point: Bug 2.");
	}

	return (starting);
}

/*
 * This routine performs the recursive portion of the backtrack search.
 */

	static
	void
search_recurse (

struct cinfo *	cip,		/* IN/OUT - compatibility/search info. */
int		level,		/* IN - recursion level. */
int		nleft,		/* IN - number of terminals left to connect. */
dist_t		length		/* IN - length of current partial soln. */
)
{
int			i;
int			j;
int			k;
int			kmasks;
int			nmasks;
int			fset_offset;
int			real_num_left;
bitmap_t *		bp1;
bitmap_t *		bp2;
bitmap_t *		bp3;
bitmap_t *		bp4;
bitmap_t		mask;
bitmap_t *		terms_left;
bitmap_t *		compat;
bitmap_t *		incompat;
bitmap_t *		try;
bitmap_t *		omit;
struct full_set *	fsp;
int			i1;
bitmap_t *		wordp;
bitmap_t		word;
int			basebit;
int			bitno;
int8u *			p1;
int8u *			p2;
int			byte;
dist_t			budget;
struct sinfo *		sip;

	sip = cip -> sinfo_ptr;

	if (length >= sip -> best_length) {
		/* Current partial solution is already too long	*/
		/* to ever become the best seen so far...	*/
		return;
	}

	nmasks	= cip -> num_fset_masks;

	if (nleft <= 0) {
		if (nleft < 0) {
			fatal ("search_recurse: Bug 1.");
		}
		/* All terminals have been connected!  We have	*/
		/* a new BEST solution!  Save it off.		*/
		bp1 = sip -> best_solution;
		bp2 = sip -> solution;
		for (i = 0; i < nmasks; i++) {
			*bp1++ = *bp2++;
		}
		sip -> best_length	= length;
		sip -> found		= TRUE;
		return;
	}

	if ((nleft >= 10) AND
	    find_inaccessible_terminal (cip, level)) {
		/* There is a terminal still be be hooked up such that	*/
		/* none of its full-trees is compatible with the	*/
		/* current partial solution!  Early cutoff!		*/
		return;
	}

	kmasks		= cip -> num_term_masks;
	terms_left	= sip -> terms_left + level * kmasks;

	fset_offset	= level * nmasks;

	compat		= sip -> compat + fset_offset;
	incompat	= sip -> incompat + fset_offset;
	try		= sip -> try + fset_offset;
	omit		= sip -> omit + fset_offset;

	/* Do the length/n ratio cutoff test -- if none of the feasible	*/
	/* pieces has a length/n ratio that beats budget/nleft, then	*/
	/* there is now way to get a better solution!			*/
	if ((nleft >= 4) AND (sip -> best_length < INF_DISTANCE)) {

		/* Compute the "length budget" with which we must	*/
		/* connect all remaining points...			*/
		budget = sip -> best_length - length;

		/* Compute the *real* number of terminals left,		*/
		/* excluding those that can be hooked up for free,	*/
		/* (because they belong to a "fake" full-set).		*/
		real_num_left = nleft;
		for (i = cip -> first_fake_term; i < cip -> num_terms; i++) {
			if (BITON (terms_left, i)) {
				/* Each fake full-set has 2 other terms	*/
				/* that get hooked up for free...	*/
				real_num_left -= 2;
			}
		}
		if (real_num_left <= 0) goto no_ratio_cutoff;

		/* Find piece with the lowest length/terms ratio that	*/
		/* is not incompatible with current partial solution...	*/
		bp1 = incompat;
		bp2 = bp1 + nmasks;
		bitno = 0;
		word = 0;
		while (bp1 < bp2) {
			word = ~(*bp1++);
			if (word NE 0) break;
			bitno += BPW;
		}
		byte = 0;
		if (word NE 0) {
			do {
				byte = (word & 0xFF);
				if (byte NE 0) break;
				bitno += 8;
				word >>= 8;
			} while (word NE 0);
		}
		if (byte NE 0) {
			k = bitno + one_bits_in_byte [byte] [0];
			if (k < cip -> full_tree_count) {
				fsp = cip -> full_trees [k];
				if ((fsp -> tree_len * real_num_left) >=
				    (budget * (fsp -> terminals -> n - 1))) {
					/* Even the piece with the best	*/
					/* length/n ratio does not beat	*/
					/* whats required for a better	*/
					/* solution...  Cutoff!		*/
#if 0
					(void) printf ("%% RATIO CUTOFF!\n");
#endif
					return;
				}
			}
		}
		no_ratio_cutoff:	;
	}

	/* Determine the set of trees we will try to add to the	*/
	/* partial solution.					*/

	bp1 = compat;
	bp2 = incompat;
	bp3 = try;
	for (i = 0; i < nmasks; i++) {
		*bp3++ = *bp1++ & ~*bp2++;
	}

	bp1 = omit;
	for (i = 0; i < nmasks; i++) {
		*bp1++ = 0;
	}

	/* Loop over each feasible full-tree. */
	wordp = try;
	for (j = 0, basebit = 0; j < nmasks; basebit += 32, j++) {
		word = *wordp++;
		if (word EQ 0) continue;
		for (bitno = basebit; word NE 0; bitno += 8, word >>= 8) {
			byte = (word & 0xFF);
			if (byte EQ 0) continue;
			p1 = one_bits_in_byte [byte];
			p2 = one_bits_in_byte [byte + 1];
			while (p1 < p2) {
				k = bitno + *p1++;
				/* Full-set K is feasible.  Test to see	*/
				/* if it is truly compatible...		*/
				fsp = cip -> full_trees [k];
				bp1 = &(fsp -> tmap [0]);
				bp2 = terms_left;
				i1 = 0;
				for (i = 0; i < kmasks; i++) {
					mask = *bp1++ & ~*bp2++;
					i1 += NBITSON (mask);
				}
				if (i1 NE 1) continue;
				/* Tree K intersects the current partial */
				/* solution at exactly one point!	 */

				bp1 = compat;
				bp2 = &(cip -> cmasks [k * nmasks]);
				bp3 = bp1 + nmasks;
				for (i = 0; i < nmasks; i++) {
					*bp3++ = *bp1++ | *bp2++;
				}

				bp1 = incompat;
				bp2 = &(cip -> incmasks [k * nmasks]);
				bp3 = omit;
				bp4 = bp1 + nmasks;
				for (i = 0; i < nmasks; i++) {
					*bp4++ = (*bp1++ | *bp2++ | *bp3++);
				}

				bp1 = terms_left;
				bp2 = &(fsp -> tmap [0]);
				bp3 = bp1 + kmasks;
				for (i = 0; i < kmasks; i++) {
					*bp3++ = *bp1++ & ~*bp2++;
				}

				SETBIT (sip -> solution, k);

				search_recurse (cip,
						level + 1,
						nleft - fsp -> terminals -> n + 1,
						length + fsp -> tree_len);

				CLRBIT (sip -> solution, k);
				SETBIT (omit, k);
			}
		}
	}
}

/*
 * This routine checks the given node level to see if there are any
 * inaccessible terminals.
 */

	static
	bool
find_inaccessible_terminal (

struct cinfo *	cip,		/* IN - compatibility/search info. */
int		level		/* IN - recursion level. */
)
{
int		t;
int		basebit;
int		bitno;
int		byte;
int		kmasks;
int		nmasks;
bitmap_t *	cur_incompat;
bitmap_t *	wordp;
bitmap_t *	endp;
bitmap_t	word;
int8u *		p1;
int8u *		p2;
bitmap_t *	bp1;
bitmap_t *	bp2;
bitmap_t *	bp3;
struct sinfo *	sip;

	sip = cip -> sinfo_ptr;

	kmasks	= cip -> num_term_masks;
	nmasks	= cip -> num_fset_masks;

	cur_incompat = sip -> incompat + (level * nmasks);
	bp3 = cur_incompat + nmasks;

	wordp	= sip -> terms_left + (level * kmasks);
	endp	= wordp + kmasks;

	for (basebit = 0; wordp < endp; basebit += 32) {
		word = *wordp++;
		if (word EQ 0) continue;
		for (bitno = basebit; word NE 0; bitno += 8, word >>= 8) {
			byte = (word & 0xFF);
			if (byte EQ 0) continue;
			p1 = one_bits_in_byte [byte];
			p2 = one_bits_in_byte [byte + 1];
			while (p1 < p2) {
				t = bitno + *p1++;

				bp1 = cip -> term_tree_masks + (t * nmasks);
				bp2 = cur_incompat;
				while (TRUE) {
					if (bp2 >= bp3) {
						/* EVERY tree involving	*/
						/* T has been found to	*/
						/* be incompatible with	*/
						/* >= 1 trees in the	*/
						/* current partial	*/
						/* solution!		*/
						return (TRUE);
					}
					if (*bp1 NE (*bp1 & *bp2)) break;
					++bp1;
					++bp2;
				}
			}
		}
	}

	return (FALSE);
}

/*
 * This routine applies the trick of using a single full-set to split the
 * search problem into smaller sub-problems.  This rapidly dispenses with
 * all possibilities involving that full-set.  Therefore, that full-set
 * can be discarded from all further searches.
 */

	static
	void
do_piece_splits (

bitmap_t *		tmap,		/* IN - subset of terminals to use */
bitmap_t *		fset_mask,	/* IN - subset of full-sets to use */
struct cinfo *		cip		/* IN/OUT - compatibility/search info. */
)
{
int			i;
int			j;
int			k;
int			nft;
int			nmasks;
int			ncomps;
int			numleft;
int			t;
bitmap_t *		fs_mask;
bitmap_t *		temp;
bitmap_t *		bp1;
bitmap_t *		bp2;
bitmap_t *		bp3;
int			best_tree;
int			best_ncomps;
int			size;
struct full_set *	fsp;
bool			bad_comp_found;
bool			changed;
bitmap_t *		comptrees;
bitmap_t *		best_trees;
int			basebit;
bitmap_t *		wordp;
bitmap_t *		endp;
bitmap_t		word;
int			byte;
int8u *			p1;
int8u *			p2;
int			newpnum [MAX_TERMINALS];
int			comp [MAX_TERMINALS + 1];
int			compsize [MAX_TERMINALS];
int			ordercomp [MAX_TERMINALS];
int			best_pnum [MAX_TERMINALS];
int			best_comp [MAX_TERMINALS + 1];
int			best_size [MAX_TERMINALS];
int			best_order [MAX_TERMINALS];

	if (cip -> sinfo_ptr -> solution EQ NULL) {
		/* Haven't been this deep before.  Allocate new	*/
		/* search stacks, as large as we could EVER	*/
		/* need again at this same level...		*/
		allocate_search_stacks (cip);
	}

	nft	= cip -> full_tree_count;
	nmasks	= cip -> num_fset_masks;

	fs_mask = NEWA (nmasks, bitmap_t);
	temp	= NEWA (nmasks, bitmap_t);

	best_trees	= NEWA (cip -> num_terms * nmasks, bitmap_t);
	comptrees	= NEWA (cip -> num_terms * nmasks, bitmap_t);

	/* Create a local MODIFIABLE copy of the full-sets mask... */
	bp1 = fset_mask;
	bp2 = fs_mask;
	numleft = 0;
	for (i = 0; i < nmasks; i++) {
		word = *bp1++;
		numleft += NBITSON (word);
		*bp2++ = word;
	}

#if 1
	(void) printf (" %% Entering piece split routine with %d pieces\n",
		       numleft);
#endif

#if 0
 {
 int			i;
 struct full_set *	fsp;
 char			buf [32];

	/* Print out all the pieces for debugging purposes... */
	for (i = 0; i < nft; i++) {
		if (NOT BITON (fs_mask, i)) continue;
		(void) sprintf (buf, "   %% piece %d:\t", i);
		fsp = cip -> full_trees [i];
		print_mask (buf, fsp -> tmap, cip -> num_terms);
	}
 }
#endif

	/* Each iteration of this loop identifies a single full-set	*/
	/* that splits the search problem.  We do the searches, discard	*/
	/* the full-set, and repeat.  If the remaining full-sets are	*/
	/* already split, we are DONE!  If no such splitting full-set	*/
	/* exists, then we brute-force search the rest.			*/

	while (TRUE) {
		bp1 = fs_mask;
		bp2 = temp;
		for (i = 0; i < nmasks; i++) {
			*bp2++ = *bp1++;
		}
		ncomps = find_components (temp,
					  tmap,
					  newpnum,
					  comp,
					  compsize,
					  comptrees,
					  cip);
		if (ncomps > 1) {
			/* We have thinned out so many full-sets that	*/
			/* it is no longer possible to connect every	*/
			/* terminal together with what's left.  DONE!	*/
#if 0
			(void) printf (" %% done because of split!\n");
#endif
			break;
		}
		do {
			changed = FALSE;
			best_tree = -1;
			best_ncomps = 1;
			best_order [0] = 0;
			best_size [0] = nft + 1;
			for (t = 0; t < nft; t++) {
				if (NOT BITON (fs_mask, t)) continue;
				bp1 = fs_mask;
				bp2 = &(cip -> incmasks [t * nmasks]);
				bp3 = temp;
				for (i = 0; i < nmasks; i++) {
					*bp3++ = *bp1++ & ~*bp2++;
				}
				ncomps = find_components (temp,
							  tmap,
							  newpnum,
							  comp,
							  compsize,
							  comptrees,
							  cip);
				if (ncomps <= 1) {
					/* Using this full-set does not split	*/
					/* the problem...  skip it.		*/
					continue;
				}
				/* Every component must intersect full-	*/
				/* set T in at least 1 place.  If not,	*/
				/* we can discard T immediately!	*/
				fsp = cip -> full_trees [t];
				bad_comp_found = FALSE;
				for (i = 0; i < ncomps; i++) {
					bad_comp_found = TRUE;
					for (j = comp [i]; j < comp [i+1]; j++) {
						if (BITON (fsp -> tmap, newpnum [j])) {
							bad_comp_found = FALSE;
							break;
						}
					}
					if (bad_comp_found) break;
				}
				if (bad_comp_found) {
#if 0
					(void) printf (" %% piece %d IMPOSSIBLE!\n", t);
#endif
					changed = TRUE;
					CLRBIT (fs_mask, t);
					--numleft;
					continue;
				}

				/* Sort components from biggest to smallest */
				sort_components (ncomps, compsize, ordercomp);

#if 0
				/* Skip if 2nd-largest component empty... */
				if (compsize [ordercomp [1]] <= 0) continue;
#endif

				/* Save if best split so far... */
				if (better_split (ncomps, compsize, ordercomp,
						  best_ncomps, best_size, best_order)) {
					best_ncomps = ncomps;
					for (i = 0; i < ncomps; i++) {
						best_size [i] = compsize [i];
					}
					for (i = 0; i < ncomps; i++) {
						best_order [i] = ordercomp [i];
					}
					for (i = 0; i <= ncomps; i++) {
						best_comp [i] = comp [i];
					}
					for (i = 0; i < comp [ncomps]; i++) {
						best_pnum [i] = newpnum [i];
					}
					size = ncomps * nmasks;
					for (i = 0; i < size; i++) {
						best_trees [i] = comptrees [i];
					}
					best_tree = t;
				}
			}
		} while (changed);
		if (best_tree < 0) {
			/* Ran out of split opportunities -- search	*/
			/* whats left in one swell foop...		*/
#if 0
			(void) printf ("  %% no split found with %d left.\n",
				       numleft);
#endif
			if (numleft > 0) {
				perform_search (tmap, fs_mask, cip);
			}
			break;
		}

#if 0
		if (numleft <= Piece_Split_Threshold) {
#if 0
			(void) printf ("  %% enough pieces gone -- use BFI\n");
#endif
			if (numleft > 0) {
				perform_search (tmap, fs_mask, cip);
			}
			break;
		}
#endif

		bp1 = fs_mask;
		bp2 = &(cip -> incmasks [best_tree * nmasks]);
		bp3 = temp;
		for (i = 0; i < nmasks; i++) {
			*bp3++ = *bp1++ & ~*bp2++;
		}

		handle_subproblems (best_tree,
				    best_ncomps,
				    best_comp,
				    best_order,
				    best_size,
				    best_pnum,
				    best_trees,
				    temp,
				    cip);

		CLRBIT (fs_mask, best_tree);
		--numleft;
	}

#if 1
	(void) printf (" %% Exiting piece split routine\n");
#endif

	free ((char *) comptrees);
	free ((char *) best_trees);
	free ((char *) temp);
	free ((char *) fs_mask);
}

/*
 * This routine determines the connected components of the given subset of
 * full-trees.
 */

	static
	int
find_components (

bitmap_t *		fs_mask,	/* IN - set of full-trees */
bitmap_t *		term_set,	/* IN - set of terminals to span */
int *			newpnum,	/* OUT - points of components */
int *			comp,		/* OUT - partition of points */
int *			compsize,	/* OUT - sizes of each component */
bitmap_t *		comptrees,	/* OUT - trees for each component */
struct cinfo *		cip		/* IN - compatibility/search info */
)
{
int			i;
int			j;
int			ncomps;
int			nmasks;
int			idx;
int			component_size;
bitmap_t *		bp1;
bitmap_t *		bp2;
bitmap_t		terms_left [N_TERM_BMAP];

	nmasks	= cip -> num_fset_masks;

	/* Make a MODIFIABLE copy of the terminals to be spanned... */
	bp1 = &terms_left [0];
	bp2 = &term_set [0];
	for (i = 0; i < N_TERM_BMAP; i++) {
		*bp1++ = *bp2++;
	}

	ncomps = 0;
	idx = 0;
	for (i = 0; i < cip -> num_terms; i++) {
		if (BITON (terms_left, i)) {
			comp [ncomps] = idx;
			component_size = 0;
			for (j = 0; j < nmasks; j++) {
				comptrees [j] = 0;
			}
			walk (i,
			      fs_mask,
			      &idx,
			      newpnum,
			      terms_left,
			      comptrees,
			      &component_size,
			      cip -> full_trees,
			      cip -> term_trees);
			compsize [ncomps] = component_size;
			++ncomps;
			comptrees += nmasks;
		}
	}
	/* Terminate last component. */
	comp [ncomps] = idx;

	return (ncomps);
}

	static
	void
walk (

int		pnum,		/* IN - point to walk from */
bitmap_t *	fs_mask,	/* IN/OUT - full-trees not yet visited */
int *		idx,		/* IN/OUT - component index */
int *		newpnum,	/* OUT - terminals in component */
bitmap_t *	terms_left,	/* IN/OUT - terminals not yet visted */
bitmap_t *	comptrees,	/* OUT - full-sets in component. */
int *		comp_size,	/* IN/OUT - count of full-trees in component */
struct full_set ** full_trees,	/* IN - array of full-trees */
int **		term_trees	/* IN - trees involving each terminal */
)
{
int			i;
int			j;
int			n;
struct point *		p1;
struct pset *		terms;
struct full_set *	fsp;
int *			tp;
int *			endp;
int			t;

	/* Unmark this terminal so that we don't visit it again. */
	CLRBIT (terms_left, pnum);
	/* Record this terminal as being in this component. */
	newpnum [(*idx)++] = pnum;

	/* Walk every full-tree involving this terminal... */
	tp	= term_trees [pnum];
	endp	= term_trees [pnum + 1];
	while (tp < endp) {
		t = *tp++;
		if (BITON (fs_mask, t)) {
			/* This full-tree has now been visited... */
			CLRBIT (fs_mask, t);
			SETBIT (comptrees, t);
			/* Component contains one more full-tree! */
			++(*comp_size);
			terms = full_trees [t] -> terminals;
			n = terms -> n;
			p1 = &(terms -> a [0]);
			for (i = 0; i < n; p1++, i++) {
				if (NOT BITON (terms_left, p1 -> pnum)) continue;
				walk (p1 -> pnum,
				      fs_mask,
				      idx,
				      newpnum,
				      terms_left,
				      comptrees,
				      comp_size,
				      full_trees,
				      term_trees);
			}
		}
	}
}

/*
 * This routine sorts the given set of components in DESCENDING order by
 * the number of full-trees for each component.  Rather than shuffling
 * the data (which is variable size), we simply generate an array
 * (ordercomp) containing the index of each component in proper order.
 */

	static
	void
sort_components (

int		ncomps,		/* IN - number of components */
int *		compsize,	/* IN - number of full-trees in each comp */
int *		ordercomp	/* OUT - sorted order of components */
)
{
int		i;
int		i1;
int		i2;
int		n;
bool		changed;

	/* Set up indexes according to current order. */
	for (i = 0; i < ncomps; i++) {
		ordercomp [i] = i;
	}

	n = ncomps - 1;

	do {
		changed = FALSE;
		for (i = 0; i < n; i++) {
			i1 = ordercomp [i];
			i2 = ordercomp [i + 1];
			if (compsize [i1] < compsize [i2]) {
				ordercomp [i]	  = i2;
				ordercomp [i + 1] = i1;
				changed = TRUE;
			}
		}
	} while (changed);
}

/*
 * This routine determines if the new split is better than the previous
 * best.
 */

	static
	bool
better_split (

int		ncomps,		/* IN - number of new components */
int *		size,		/* IN - sizes of new components */
int *		order,		/* IN - order of new components */
int		best_ncomps,	/* IN - number of components in best so far */
int *		best_size,	/* IN - sizes of components in best split */
int *		best_order	/* IN - order of components in best split */
)
{
int		i1;
int		i2;

	while (TRUE) {
		if (ncomps <= 0) return (best_ncomps > 0);
		if (best_ncomps <= 0) return (FALSE);
		i1 = *order++;
		i2 = *best_order++;
		if (size [i1] < best_size [i2]) return (TRUE);
		if (size [i1] > best_size [i2]) return (FALSE);
		--ncomps;
		--best_ncomps;
	}
}

/*
 * This routine launches combinatorial searches on each of the sub-problems
 * obtained by using the full-set "split_fs" in the solution.  We have
 * already determined that by putting this piece down first, the problem
 * splits into "ncomp" number of components, each completely independent,
 * and all bridged together by this single full-set.
 */

	static
	void
handle_subproblems (

int		split_fs,		/* IN - full-set that splits problem */
int		ncomps,			/* IN - number of sub-problems */
int *		comp,			/* IN - splits pnum into components */
int *		ordercomp,		/* IN - sorts components by size */
int *		compsize,		/* IN - full-sets in each component */
int *		pnum,			/* IN - point numbers in components */
bitmap_t *	comptrees,		/* IN - mask of trees for each comp */
bitmap_t *	fset_mask,		/* IN - full-sets of sub-problems */
struct cinfo *	cip			/* IN - compatibility/search info */
)
{
int			i;
int			j;
int			k;
int			nmasks;
int			t;
bitmap_t *		bp1;
bitmap_t *		bp2;
bitmap_t *		sub_fs;
bitmap_t *		combined_solution;
dist_t			total_len;
dist_t			split_len;
bool			failure;
struct sinfo *		old_sip;
struct sinfo *		sip;
bitmap_t		new_tmap [N_TERM_BMAP];

	/* The components are listed from largest to smallest in the	*/
	/* "ordercomp" array.  Strip off any sub-problems that contain	*/
	/* only zero full-sets.  These require no action!		*/
	while (TRUE) {
		if (ncomps <= 0) {
			/* WHOA!  There must be SOME work here!!! */
			fatal ("handle_subproblems: Bug 1.");
		}
		k = ordercomp [ncomps - 1];	/* K = smallest subproblem. */
		if (compsize [k] > 0) break;
		--ncomps;
	}

#if 0
 {
 struct full_set *	fsp;

	(void) sprintf (buf, " %% split using piece %d with pts:", split_fs);
	fsp = cip -> full_trees [split_fs];
	print_mask (buf, fsp -> tmap, cip -> num_terms);
	for (i = 0; i < ncomps; i++) {
		k = ordercomp [i];
		(void) printf ("  %% component %d with %d trees:",
			       i, compsize [k]);
		for (j = comp [k]; j < comp [k+1]; j++) {
			(void) printf (" %d", pnum [j]);
		}
		(void) printf ("\n");
	}
 }
#endif

	nmasks = cip -> num_fset_masks;

	/* Create a bit-mask into which to assemble a solution. */
	combined_solution	= NEWA (nmasks, bitmap_t);

	bp1 = combined_solution;
	for (i = 0; i < nmasks; i++) {
		*bp1++ = 0;
	}

	/* Combined solution always contains the split-piece... */
	SETBIT (combined_solution, split_fs);
	split_len	= cip -> full_trees [split_fs] -> tree_len;
	total_len	= split_len;

	/* Begin a new search context... */
	old_sip = cip -> sinfo_ptr;
	sip	= old_sip + 1;

	cip -> sinfo_ptr = sip;

	if (sip -> solution EQ NULL) {
		/* Haven't been this deep before.  Allocate new	*/
		/* search stacks, as large as we could EVER	*/
		/* need again at this same level...		*/
		allocate_search_stacks (cip);
	}

	sub_fs = NEWA (nmasks, bitmap_t);

	/* Loop through the components from smallest to largest, doing	*/
	/* a search sub-problem for each...				*/

	failure = FALSE;

	for (j = ncomps; j > 0; j--) {
		/* Process component number K... */
		k = ordercomp [j - 1];

		/* Form a mask of all full-sets in this component... */
		bp1 = comptrees + (k * nmasks);
		bp2 = sub_fs;
		for (i = 0; i < nmasks; i++) {
			*bp2++ = *bp1++;
		}

		/* Be certain that the splitting piece is in there too... */
		SETBIT (sub_fs, split_fs);

		/* Create a mask of all terminals in this component... */
		/* (include the splitting piece first...)		*/
		bp1 = &new_tmap [0];
		bp2 = &(cip -> full_trees [split_fs] -> tmap [0]);
		for (i = 0; i < N_TERM_BMAP; i++) {
			*bp1++ = *bp2++;
		}

		for (i = comp [k]; i < comp [k+1]; i++) {
			SETBIT (new_tmap, pnum [i]);
		}

		/* Set up to receive best solution of this sub-problem... */
		sip -> best_length	= INF_DISTANCE;
		sip -> found		= FALSE;

		/* Any combined solution will still have to beat the	*/
		/* best solution found by our caller so far...		*/
		if (old_sip -> found) {
			sip -> best_length = old_sip -> best_length - total_len + split_len;
		}

#if 1
		/* Handle sub-problem.  If it is big enougn,	*/
		/* try doing the piece splits recursively!	*/
		check_piece_splits (new_tmap, sub_fs, cip);
#else
		/* Do exhaustive search on sub-problem. */
		perform_search (new_tmap, sub_fs, cip);
#endif

		if (NOT sip -> found) {
			/* One of the sub-problems could not be solved!	*/
			/* This is not fatal, but it does mean that we	*/
			/* can toss out the split-piece...		*/
#if 0
			(void) printf ("  %% Component %d could NOT be solved!\n",
				       j - 1);
#endif
			failure = TRUE;
			break;
		}

		/* Merge this sub-problem's solution into combined soln	*/
		bp1 = combined_solution;
		bp2 = sip -> best_solution;
		for (i = 0; i < nmasks; i++) {
			*bp1++ |= *bp2++;
		}
		total_len += (sip -> best_length - split_len);
#if 0
		CLRBIT (sip -> best_solution, split_fs);
		print_tree ("solution is", sip -> best_solution, cip);
		(void) printf ("  %%    length = %d\n", len);
		SETBIT (sip -> best_solution, split_fs);
#endif
	}

	if (NOT failure) {
		/* We have solved all sub-problems!  Copy combined	*/
		/* solution back to caller if it is better than his!	*/
		if ((NOT old_sip -> found) OR (total_len < old_sip -> best_length)) {
			old_sip -> found = TRUE;
			bp1 = old_sip -> best_solution;
			bp2 = combined_solution;
			for (i = 0; i < nmasks; i++) {
				*bp1++ = *bp2++;
			}
			old_sip -> best_length = total_len;
#if 0
			(void) printf (" %% NEW BEST FOUND!\n");
			(void) printf (" %% ->");
			print_mask (combined_solution, cip -> full_tree_count);
			(void) printf (" %%   length = %d\n", total_len);
#endif
		}
	}

	free ((char *) sub_fs);
	free ((char *) combined_solution);

	cip -> sinfo_ptr = old_sip;
}
