/***********************************************************************

	File:	steiner.h
	Rev:	a-1
	Date:	02/20/93

	Copyright (c) 1993, by The University of Virginia

************************************************************************

	General declarations for the Rectilinear Steiner Tree program.

************************************************************************

	Modification Log:

	a-1:	02/20/93	dmw
		: Created.

************************************************************************/

#ifndef	STEINER_H
#define	STEINER_H

#include <ctype.h>
#include <limits.h>
#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>


/*
 * Typedef's to insulate us from the ugliness of C types.
 */

typedef unsigned char		int8u;
#if	defined(__STDC__) && ((__STDC__+0)==1)
typedef	signed char		int8s;
#else
typedef char			int8s;
#endif
typedef unsigned short		int16u;
typedef short			int16s;
typedef unsigned long		int32u;
typedef long			int32s;
typedef char			bool;


/*
 * Macros to protect us from some of C's sharpest edges.
 */

#define NOT	!
#define AND	&&
#define OR	||
#define FALSE	0
#define	TRUE	1
#define EQ	==
#define NE	!=

/*
 * Fundamental constants.
 */

#define	MAX_TERMINALS		128
#define	MAX_STEINERS		(MAX_TERMINALS - 2)
#define	MAX_POINTS		(MAX_TERMINALS + MAX_STEINERS)


/*
 * Primitive types associated with points.
 */

typedef int16u		coord_t;	/* type of X and Y coordinates. */
typedef	int16s		pnum_t;		/* type of point number indices. */
typedef int32u		dist_t;		/* distance. */

#define	INF_DISTANCE	((dist_t) ULONG_MAX)	/* "infinite" distance. */
#define	MAX_COORD	10000


/*
 * Bit Maps
 */

typedef	int32u		bitmap_t;	/* Element of Bit-map vector. */
#define	BPW		32		/* Bits per word of a bit-map. */
#define	BMAP_ELTS(n)	(((n) + (BPW-1)) / BPW)
#define	N_TERM_BMAP	BMAP_ELTS (MAX_TERMINALS)


/*
 * The following structure represents a single point.
 */

struct point {
	coord_t		x;
	coord_t		y;
	pnum_t		pnum;
	pnum_t		other_pnum;
};


/*
 * The following represents an input set of points, with the count.
 */

struct pset {
	int		n;
	struct point	a [MAX_POINTS];
};

/*
 * The folling represents a single edge in a tree or graph of points.
 */

struct edge {
	dist_t		len;		/* Length of edge. */
	pnum_t		p1;		/* First endpoint of edge. */
	pnum_t		p2;		/* Second endpoint of edge. */
};


/*
 * The following represents a single adjacency relationship when we represent
 * a tree or graph using adjacency lists.
 */

struct adj {
	dist_t		len;		/* Length of edge. */
	pnum_t		pnum;		/* Index of other point. */
};

/*
 * A type to represent which side of the backbone a particular point is on.
 */
enum side {
	UNKNOWN,
	LEFT,
	RIGHT
};


/*
 * The various types of backbone styles.
 */

enum backbone {
	ZERO_H_REST_V,		/* Zero on Horizontal leg, rest on Vertical */
	ONE_H_REST_V,		/* One on Horizontal leg, rest on Vertical */
	ZERO_V_REST_H,		/* Zero on Vertical let, rest on Horizontal */
	ONE_V_REST_H,		/* One on Vertical leg, rest on Horizontal */
	ZERO_H_ZERO_V,		/* Simple two-point backbone */
	ONLY_H,			/* Backbone is exactly horizontal */
	ONLY_V,			/* Backbone is exactly vertical */
	CROSS,			/* The ultimate degeneracy -- a cross */
	FAKE			/* Internal use only -- no geometric meaning */
};

/*
 * This structure is used to represent a particular full-set and/or
 * full-topology.
 */

struct full_set {
	struct full_set *	next;
	enum backbone		style;
	int16u			tree_num;
	dist_t			tree_len;
	struct pset *		terminals;
	struct pset *		steiners;
	bitmap_t		tmap [N_TERM_BMAP];
};


/*
 * This type defines the relative compatibility of two full-sets.
 */

enum compat {
	COMPATIBLE,
	INCOMPATIBLE,
	DISJOINT
};


/*
 * A collection of various compatibility data bases...
 */

struct cinfo {
	/* static parameters. */
	int			full_tree_count;
	struct full_set **	full_trees;
	int			num_terms;
	int			num_term_masks;
	int			num_fset_masks;
	int			first_fake_term;
	struct pset *		pts;
	/* pre-initialized tables. */
	int8u *			carray;
	bitmap_t *		cmasks;
	bitmap_t *		incmasks;
	int **			term_trees;
	bitmap_t *		term_tree_masks;
	/* stack of search sub-problems... */
	struct sinfo *		sinfo_ptr;
};


/*
 * The stack-oriented data structures for backtrack search...
 */

struct sinfo {
	/* the best solution so far... */
	dist_t			best_length;
	int			best_count;
	bitmap_t *		best_solution;
	bool			found;
	/* Current partial solution... */
	bitmap_t *		solution;
	/* arrays accessed in stack fashion during backtrack... */
	bitmap_t *		terms_left;
	bitmap_t *		compat;
	bitmap_t *		incompat;
	bitmap_t *		try;
	bitmap_t *		omit;
};

/*
 * Useful macros.
 */

#define	DIFF(x,y)	(((x) > (y)) ? ((x)-(y)) : ((y)-(x)))
#define	DISTANCE(p1, p2)	(DIFF ((p1) -> x, (p2) -> x) + \
				 DIFF ((p1) -> y, (p2) -> y))

#define	BETWEEN_X(p1, p2, p3)					\
		(((p1) -> x < (p3) -> x) ?			\
		 (((p1) -> x <= (p2) -> x) AND ((p2) -> x <= (p3) -> x)) \
			:					\
		 (((p3) -> x <= (p2) -> x) AND ((p2) -> x <= (p1) -> x)))

#define	BETWEEN_Y(p1, p2, p3)					\
		(((p1) -> y < (p3) -> y) ?			\
		 (((p1) -> y <= (p2) -> y) AND ((p2) -> y <= (p3) -> y)) \
			:					\
		 (((p3) -> y <= (p2) -> y) AND ((p2) -> y <= (p1) -> y)))

#define	STRICTLY_BETWEEN_X(p1, p2, p3)				\
		(((p1) -> x < (p3) -> x) ?			\
		 (((p1) -> x < (p2) -> x) AND ((p2) -> x < (p3) -> x)) \
			:					\
		 (((p3) -> x < (p2) -> x) AND ((p2) -> x < (p1) -> x)))

#define	STRICTLY_BETWEEN_Y(p1, p2, p3)				\
		(((p1) -> y < (p3) -> y) ?			\
		 (((p1) -> y < (p2) -> y) AND ((p2) -> y < (p3) -> y)) \
			:					\
		 (((p3) -> y < (p2) -> y) AND ((p2) -> y < (p1) -> y)))

#define	min(a, b)	(((a) < (b)) ? (a) : (b))
#define	max(a, b)	(((a) > (b)) ? (a) : (b))

#define	NEW(type)	((type *) new (sizeof (type)))
#define	NEWA(n, type)	((type *) new ((size_t) ((n) * sizeof (type))))


#define	NULL_PSET	((struct pset *) 0)
#define	PSET_SIZE(n)	(sizeof (struct pset) -			\
			 sizeof (NULL_PSET -> a) +		\
			 (n)*sizeof (NULL_PSET -> a [0]))
#define	COPY_PSET(dp, sp)	(void) memcpy ((dp), (sp), PSET_SIZE ((sp) -> n))

#define	SETBIT(bm, n)	((bm) [(n) / BPW] |= (1ul << ((n) % BPW)))
#define	CLRBIT(bm, n)	((bm) [(n) / BPW] &= ~(1ul << ((n) % BPW)))
#define	BITON(bm, n)	(((bm) [(n) / BPW] & (1ul << ((n) % BPW))) NE 0)
#define	NBITSON(m)	(  nbits [ (m)        & 0xFFlu]		\
			 + nbits [((m) >>  8) & 0xFFlu]		\
			 + nbits [((m) >> 16) & 0xFFlu]		\
			 + nbits [((m) >> 24) & 0xFFlu])

/*
 * Stuff for measuring CPU time usage.
 */

typedef int32u		cpu_time_t;

#define	TICKS_PER_SEC	100		/* This is the units WE use! */

/*
 * Global Variables
 */

		/* command line options... */

extern bool		Disable_Recursive_Decomposition;
extern int		Piece_Split_Threshold;
extern bool		Print_Bounds;
extern bool		Print_Compatibility_Matrix;
extern bool		Print_Full_Sets;
extern bool		Print_Overlaid_Full_Sets;
extern bool		Print_Relief_Map;
extern bool		Print_Split_Full_Sets;
extern bool		Skip_Kahng_Robins;
extern bool		Skip_Minimum_Spanning_Tree;
extern bool		Stop_After_Stage_1;
extern bool		Use_Thorough_Pruning;

extern int8u		nbits [];
extern int8u		one_bits_in_byte_array [];
extern int8u *		one_bits_in_byte [];


/*
 * Function Prototypes.
 */

extern dist_t		backbone_distance (struct point *,
					   struct point *,
					   struct point *);
extern bool		backtrack_search (bitmap_t *,
					  bitmap_t *,
					  struct cinfo *,
					  bitmap_t *);
extern void		build_compatibility_data (struct pset *,
						  bitmap_t *,
						  struct full_set *,
						  bitmap_t *,
						  struct cinfo *);
extern void		coarse_grained_tests (struct pset *,
					      bitmap_t *,
					      int *);
extern void		compute_mst_longest_edges (struct pset *);
extern void		convert_cpu_time (cpu_time_t, char *);
extern struct full_set * copy_full_set (struct full_set *);
extern void		copy_terminal_subset (struct pset *,
					      struct pset *,
					      bitmap_t *);
extern void		define_Plot_Terminals (struct pset *);
extern void		fatal (char *);
extern void		free_compatibility_data (struct cinfo *);
extern bool		full_topology_edge_too_long (int, struct pset *,
						     struct edge *, int);
extern struct full_set * generate_all_full_sets (struct pset *,
						 bitmap_t *,
						 int *);
extern struct full_set * generate_duplicate_full_sets (struct pset *,
						       bitmap_t *);
extern cpu_time_t	get_cpu_time (void);
extern void		init_tables (void);
extern void		initial_pruning (bitmap_t *,
					 bitmap_t *,
					 struct cinfo *);
extern bool		is_point_close_to (struct pset *,
					   struct point *, struct point *);
extern int		kahng_robins (struct pset *, dist_t, struct edge *);
extern dist_t		kahng_robins_length (struct pset *, dist_t);
extern void		list_valid_points (struct pset *,
					   struct point *,
					   struct point *,
					   bitmap_t *,
					   struct pset *);
extern void		make_second_topology (struct full_set *,
					      struct full_set *);
extern int		mst (struct pset *, struct edge *);
extern void *		new (size_t);
extern void		overlay_plot_innards (bitmap_t *, struct cinfo *);
extern void		overlay_plot_subset (char *,
					     bitmap_t *,
					     struct cinfo *);
extern bool		passes_fine_tests (struct pset *, struct full_set *);
extern void		plot_full_set_split_pair (int,
						  int,
						  bitmap_t *,
						  struct cinfo *);
extern void		plot_full_sets (bitmap_t *, struct cinfo *);
extern void		plot_segment (struct point *, struct point *);
extern void		plot_split_full_sets (bitmap_t *, struct cinfo *);
extern void		print_mask (char *, bitmap_t *, int);
extern void		print_tree (char *, bitmap_t *, struct cinfo *);
extern struct full_set ** put_trees_in_array (struct full_set *, int *);
extern void		relief_plot_full_sets (char *,
					       bitmap_t *,
					       struct cinfo *);
extern struct pset *	save_pset (struct pset *);
extern struct full_set * sort_full_trees (struct full_set *);
extern void		sort_x_dec (struct pset *);
extern void		sort_x_inc (struct pset *);
extern void		sort_xy_inc (struct pset *);
extern void		sort_y_dec (struct pset *);
extern void		sort_y_inc (struct pset *);
extern void		sort_yx_inc (struct pset *);
extern bool		split_and_solve (bitmap_t *,
					 bitmap_t *,
					 struct cinfo *,
					 bitmap_t *);
extern dist_t		total_tree_length (bitmap_t *, struct cinfo *);

#endif
