/***********************************************************************

	File:	utils.c
	Rev:	a-1
	Date:	04/18/93

	Copyright (c) 1993, by The University of Virginia

************************************************************************

	Various utility routines.

************************************************************************

	Modification Log:

	a-1:	04/18/93	dmw
		: Created.  Collected these routines into this file
		:  from other places.

************************************************************************/

#include "steiner.h"


/*
 * Global Routines
 */

dist_t		backbone_distance (struct point *,
				   struct point *,
				   struct point *);
void		copy_terminal_subset (struct pset *,
				      struct pset *,
				      bitmap_t *);
void		fatal (char *);
bool		is_point_close_to (struct pset *,
				   struct point *,
				   struct point *);
void		list_valid_points (struct pset *,
				   struct point *,
				   struct point *,
				   bitmap_t *,
				   struct pset *);
void *		new (size_t);
void		sort_x_dec (struct pset *);
void		sort_x_inc (struct pset *);
void		sort_xy_inc (struct pset *);
void		sort_y_dec (struct pset *);
void		sort_y_inc (struct pset *);
void		sort_yx_inc (struct pset *);


/*
 * External References
 */

	/* none */


/*
 * Local Routines
 */

static bool		valid_point (struct point *,
				     struct point *,
				     struct point *,
				     struct pset *);

/*
 * This routine displays a fatal message and then dies!
 */

	void
fatal (

char *		msg		/* IN - message to display. */
)
{
	(void) fprintf (stderr, "%s\n", msg);
	(void) fflush (stderr);
	abort ();
}

/*
 * This routine performs all dynamic memory allocation for the program.
 * We test for out of memory condition here.
 */

	void *
new (

size_t		size		/* IN - size of chunk in bytes. */
)
{
void *		p;

	p = malloc (size);
	if (p EQ NULL) {
		(void) fprintf (stderr, "Out of memory!\n");
		exit (1);
	}

	return (p);
}

/*
 * This routine returns a list of all terminals that could possibly be part
 * of a full topology with backbone from A to B, where the vertical
 * line of the backbone is incident to A.  The terminals A and B are not
 * included in the list.
 */

	void
list_valid_points (

struct pset *		pts,		/* IN - the point set. */
struct point *		ap,		/* IN - point A of backbone. */
struct point *		bp,		/* IN - point B of backbone. */
bitmap_t *		tmap,		/* IN - terminals to consider. */
struct pset *		valid_pts	/* OUT - set of valid terminals. */
)
{
int			i;
int			n;
struct point *		p1;
struct point *		outp;
int			count;

	n = pts -> n;

	count = 0;
	outp = &(valid_pts -> a [0]);

	p1 = &(pts -> a [0]);
	for (i = 0; i < n; p1++, i++) {
		if (p1 EQ ap) continue;
		if (p1 EQ bp) continue;
		if (NOT BITON (tmap, i)) continue;
		if (valid_point (p1, ap, bp, pts)) {
			*outp++ = *p1;
			++count;
		}
	}

	valid_pts -> n = count;
}

/*
 * This routine returns TRUE if terminal X satisfies the following
 * property:
 *
 *	No terminal is simultaneously closer to X and the backbone
 *	than X's distance from the backbone.
 */

	static
	bool
valid_point (

struct point *		xp,		/* IN - point X to test. */
struct point *		ap,		/* IN - point A of backbone. */
struct point *		bp,		/* IN - point B of backbone. */
struct pset *		pts		/* IN - original point set. */
)
{
int			i;
int			n;
struct point *		p1;
dist_t			bound;

	if (NOT (STRICTLY_BETWEEN_X (ap, xp, bp) OR
		 STRICTLY_BETWEEN_Y (ap, xp, bp))) {
		/* Totally impossible to hook this point to backbone! */
		return (FALSE);
	}

	bound = backbone_distance (ap, bp, xp);

	n = pts -> n;

	p1 = &(pts -> a [0]);
	for (i = 0; i < n; p1++, i++) {
		if ((DISTANCE (xp, p1) < bound) AND
		    (backbone_distance (ap, bp, p1) < bound)) {
			return (FALSE);
		}
	}

	return (TRUE);
}

/*
 * This routine determines the length of the shortest straight (horizontal
 * or vertical) from point X to the backbone from A to B.  The vertical leg
 * of the backbone is incident to A.
 */

	dist_t
backbone_distance (

struct point *		ap,		/* IN - point A */
struct point *		bp,		/* IN - point B */
struct point *		xp		/* IN - point X */
)
{
dist_t			delta_x;
dist_t			delta_y;
dist_t			bd;
dist_t			d;

	delta_x = DIFF (ap -> x, xp -> x);
	delta_y = DIFF (bp -> y, xp -> y);

	if (BETWEEN_X (ap, xp, bp)) {
		if (BETWEEN_Y (ap, xp, bp)) {
			bd = min (delta_x, delta_y);
		}
		else {
			d = DISTANCE (xp, ap);
			bd = min (d, delta_y);
		}
	}
	else if (BETWEEN_Y (ap, xp, bp)) {
		d = DISTANCE (xp, bp);
		bd = min (d, delta_x);
	}
	else {
		bd = DISTANCE (xp, ap);
		d  = DISTANCE (xp, bp);
		if (d < bd) {
			bd = d;
		}
		d = delta_x + delta_y;
		if (d < bd) {
			bd = d;
		}
	}

	return (bd);
}

/*
 * This routine returns TRUE if there is a point that is closer to both
 * points A and B than points A and B are to each other.
 */

	bool
is_point_close_to (

struct pset *		pts,		/* IN - the point set. */
struct point *		ap,		/* IN - point A. */
struct point *		bp		/* IN - point B. */
)
{
int			i;
int			n;
struct point *		p1;
dist_t			d;

	d = DISTANCE (ap, bp);

	n = pts -> n;

	p1 = &(pts -> a [0]);
	for (i = 0; i < n; p1++, i++) {
		if ((DISTANCE (p1, ap) < d) AND
		    (DISTANCE (p1, bp) < d)) {
			return (TRUE);
		}
	}
	return (FALSE);
}

/*
 * This routine copies a SUBSET of the given terminal set.  The subset is
 * specified by means of a bit-map.
 */

	void
copy_terminal_subset (

struct pset *		dset,	/* OUT - destination set */
struct pset *		sset,	/* IN - source set */
bitmap_t *		tmap	/* IN - bit-map indicating subset to copy */
)
{
int		i;
int		num_in;
int		num_out;
struct point *	src;
struct point *	dst;

	num_in	= sset -> n;
	src	= &(sset -> a [0]);
	dst	= &(dset -> a [0]);
	num_out	= 0;

	for (i = 0; i < num_in; i++, src++) {
		if (BITON (tmap, i)) {
			*dst++ = *src;
			++num_out;
		}
	}

	dset -> n = num_out;
}

/*
 * This routine sorts the given point set in order by INCREASING X coord.
 */

	void
sort_x_inc (

struct pset *		pts	/* IN/OUT - point set to sort. */
)
{
int		n;
int		h;
struct point	tmp;
coord_t		key;
struct point *	p1;
struct point *	p2;
struct point *	p3;
struct point *	p4;
struct point *	endp;

	n = pts -> n;

	endp = &(pts -> a [n]);

	for (h = 1; h <= n; h = 3*h+1) {
	}

	do {
		h = h / 3;
		p4 = &(pts -> a [h]);
		p1 = p4;
		while (p1 < endp) {
			tmp = *p1;
			key = tmp.x;
			p2 = p1;
			while (TRUE) {
				p3 = (p2 - h);
				if (p3 -> x <= key) break;
				*p2 = *p3;
				p2 = p3;
				if (p2 < p4) break;
			}
			*p2 = tmp;
			++p1;
		}
	} while (h > 1);
}

/*
 * This routine sorts the given point set in order by DECREASING X coord.
 */

	void
sort_x_dec (

struct pset *		pts	/* IN/OUT - point set to sort. */
)
{
int		n;
int		h;
struct point	tmp;
coord_t		key;
struct point *	p1;
struct point *	p2;
struct point *	p3;
struct point *	p4;
struct point *	endp;

	n = pts -> n;

	endp = &(pts -> a [n]);

	for (h = 1; h <= n; h = 3*h+1) {
	}

	do {
		h = h / 3;
		p4 = &(pts -> a [h]);
		p1 = p4;
		while (p1 < endp) {
			tmp = *p1;
			key = tmp.x;
			p2 = p1;
			while (TRUE) {
				p3 = (p2 - h);
				if (p3 -> x >= key) break;
				*p2 = *p3;
				p2 = p3;
				if (p2 < p4) break;
			}
			*p2 = tmp;
			++p1;
		}
	} while (h > 1);
}

/*
 * This routine sorts the given point set on two keys.  First by
 * INCREASING X coord.  If equal, then by INCREASING Y coord.
 */

	void
sort_xy_inc (

struct pset *		pts	/* IN/OUT - point set to sort. */
)
{
int		n;
int		h;
struct point	tmp;
struct point *	p1;
struct point *	p2;
struct point *	p3;
struct point *	p4;
struct point *	endp;

	n = pts -> n;

	endp = &(pts -> a [n]);

	for (h = 1; h <= n; h = 3*h+1) {
	}

	do {
		h = h / 3;
		p4 = &(pts -> a [h]);
		p1 = p4;
		while (p1 < endp) {
			tmp = *p1;
			p2 = p1;
			while (TRUE) {
				p3 = (p2 - h);
				if (p3 -> x < tmp.x) break;
				if ((p3 -> x EQ tmp.x) AND
				    (p3 -> y <= tmp.y)) break;
				*p2 = *p3;
				p2 = p3;
				if (p2 < p4) break;
			}
			*p2 = tmp;
			++p1;
		}
	} while (h > 1);
}

/*
 * This routine sorts the given point set in order by INCREASING Y coord.
 */

	void
sort_y_inc (

struct pset *		pts	/* IN/OUT - point set to sort. */
)
{
int		n;
int		h;
struct point	tmp;
coord_t		key;
struct point *	p1;
struct point *	p2;
struct point *	p3;
struct point *	p4;
struct point *	endp;

	n = pts -> n;

	endp = &(pts -> a [n]);

	for (h = 1; h <= n; h = 3*h+1) {
	}

	do {
		h = h / 3;
		p4 = &(pts -> a [h]);
		p1 = p4;
		while (p1 < endp) {
			tmp = *p1;
			key = tmp.y;
			p2 = p1;
			while (TRUE) {
				p3 = (p2 - h);
				if (p3 -> y <= key) break;
				*p2 = *p3;
				p2 = p3;
				if (p2 < p4) break;
			}
			*p2 = tmp;
			++p1;
		}
	} while (h > 1);
}

/*
 * This routine sorts the given point set in order by DECREASING Y coord.
 */

	void
sort_y_dec (

struct pset *		pts	/* IN/OUT - point set to sort. */
)
{
int		n;
int		h;
struct point	tmp;
coord_t		key;
struct point *	p1;
struct point *	p2;
struct point *	p3;
struct point *	p4;
struct point *	endp;

	n = pts -> n;

	endp = &(pts -> a [n]);

	for (h = 1; h <= n; h = 3*h+1) {
	}

	do {
		h = h / 3;
		p4 = &(pts -> a [h]);
		p1 = p4;
		while (p1 < endp) {
			tmp = *p1;
			key = tmp.y;
			p2 = p1;
			while (TRUE) {
				p3 = (p2 - h);
				if (p3 -> y >= key) break;
				*p2 = *p3;
				p2 = p3;
				if (p2 < p4) break;
			}
			*p2 = tmp;
			++p1;
		}
	} while (h > 1);
}

/*
 * This routine sorts the given point set on two keys.  First by
 * INCREASING Y coord.  If equal, then by INCREASING X coord.
 */

	void
sort_yx_inc (

struct pset *		pts	/* IN/OUT - point set to sort. */
)
{
int		n;
int		h;
struct point	tmp;
struct point *	p1;
struct point *	p2;
struct point *	p3;
struct point *	p4;
struct point *	endp;

	n = pts -> n;

	endp = &(pts -> a [n]);

	for (h = 1; h <= n; h = 3*h+1) {
	}

	do {
		h = h / 3;
		p4 = &(pts -> a [h]);
		p1 = p4;
		while (p1 < endp) {
			tmp = *p1;
			p2 = p1;
			while (TRUE) {
				p3 = (p2 - h);
				if (p3 -> y < tmp.y) break;
				if ((p3 -> y EQ tmp.y) AND
				    (p3 -> x <= tmp.x)) break;
				*p2 = *p3;
				p2 = p3;
				if (p2 < p4) break;
			}
			*p2 = tmp;
			++p1;
		}
	} while (h > 1);
}
